/*
*  SNOM low level audio driver for AC 97 codec
*  2002 snom technology AG
*
* Description:
*
* This file is one of files for the low level driver for the 
* SNOM Rev D. audio hardware.
*
* It implements the HAL for the AC97 codec.
*
*/


#ifdef CONFIG_SNOM_AC97
#ifndef AC97_H
#define AC_97 1

#define AC97_RST (1<<10)

#define COMMAND_ONLY 0x0007
#define SOUND_ONLY 0x0019
#define SOUND&COMMAND 0x001F

/* AC'97 registers */
#define  AC97_RESET              0x0000 /* read only reg*/    
#define  AC97_MASTER_VOL_STEREO  0x0002     
#define  AC97_HEADPHONE_VOL      0x0004      // 
#define  AC97_MASTER_VOL_MONO    0x0006      // TAD Output
#define  AC97_MASTER_TONE        0x0008      //
#define  AC97_PCBEEP_VOL         0x000a      // none
#define  AC97_PHONE_VOL          0x000c      // TAD Input (mono)
#define  AC97_MIC_VOL            0x000e      // MIC Input (mono)
#define  AC97_LINEIN_VOL         0x0010      // Line Input (stereo)
#define  AC97_CD_VOL             0x0012      // CD Input (stereo)
#define  AC97_VIDEO_VOL          0x0014      // none
#define  AC97_AUX_VOL            0x0016      // Aux Input (stereo)
#define  AC97_PCMOUT_VOL         0x0018      // Wave Output (stereo)
#define  AC97_RECORD_SELECT      0x001a      //
#define  AC97_RECORD_GAIN        0x001c
#define  AC97_RECORD_GAIN_MIC    0x001e
#define  AC97_GENERAL_PURPOSE    0x0020
#define  AC97_3D_CONTROL         0x0022
#define  AC97_MODEM_RATE         0x0024
#define  AC97_POWER_CONTROL      0x0026

/* registers 0x0028 - 0x0058 are reserved */

/* AC'97 2.0 */
#define AC97_EXTENDED_ID	0x0028	/* Extended Audio ID */
#define AC97_EXTENDED_STATUS	0x002A	/* Extended Audio Status */
#define AC97_PCM_FRONT_DAC_RATE 0x002C  /* PCM Front DAC Rate */
#define AC97_PCM_SURR_DAC_RATE  0x002E  /* PCM Surround DAC Rate */
#define AC97_PCM_LFE_DAC_RATE   0x0030  /* PCM LFE DAC Rate */
#define AC97_PCM_LR_ADC_RATE	0x0032	/* PCM LR DAC Rate */
#define AC97_PCM_MIC_ADC_RATE   0x0034  /* PCM MIC ADC Rate */
#define AC97_CENTER_LFE_MASTER  0x0036  /* Center + LFE Master Volume */
#define AC97_SURROUND_MASTER    0x0038  /* Surround (Rear) Master Volume */
#define AC97_RESERVED_3A	0x003A	/* Reserved */
/* range 0x3c-0x58 - MODEM */

/* registers 0x005a - 0x007a are vendor reserved */

#define  AC97_VENDOR_ID1         0x007c
#define  AC97_VENDOR_ID2         0x007e

/* volume control bit defines */

#define AC97_MUTE                0x8000
#define MIC_MUTE                 0x8000
#define ADC_MUTE                 0x8000
#define AC97_MICBOOST            0x0040
#define AC97_MICATTEN            0x001f
#define AC97_LEFTVOL             0x3f00
#define AC97_RIGHTVOL            0x003f
#define AC97_MIC_SEL             0x0100 /* Mic Select 0=Mic1, 1=Mic2 */
#define AC97_LPBK                0x0080 /* Loopback mode */
#define AC97_REC_GAIN_LEFT       0X0f00
#define AC97_REC_GAIN_RIGHT      0X000f
#define AC97_MONOVOL             0x003f
#define ON                       0x0001
#define AC97_DAC_ATT_R           0x001f
#define AC97_DAC_ATT_L           0x1f00
/* record mux defines */

#define AC97_RECMUX_MIC          0x0000
#define AC97_RECMUX_CD           0x0101
#define AC97_RECMUX_VIDEO        0x0202      /* not used */
#define AC97_RECMUX_AUX          0x0303      
#define AC97_RECMUX_LINE         0x0404      
#define AC97_RECMUX_STEREO_MIX   0x0505
#define AC97_RECMUX_MONO_MIX     0x0606
#define AC97_RECMUX_PHONE        0x0707


/* general purpose register bit defines */


#define AC97_GP_MIX              0x0200      /* Mono output select 0=Mix, 1=Mic */
#define AC97_GP_RLBK             0x0400      /* Remote Loopback - Modem line codec */
#define AC97_GP_LLBK             0x0800      /* Local Loopback - Modem Line codec */
#define AC97_GP_LD               0x1000      /* Loudness 1=on */
#define AC97_GP_3D               0x2000      /* 3D Enhancement 1=on */
#define AC97_GP_ST               0x4000      /* Stereo Enhancement 1=on */
#define AC97_GP_POP              0x8000      /* Pcm Out Path, 0=pre 3D, 1=post 3D */


/* powerdown control and status bit defines */

/* status */
#define AC97_PWR_MDM             0x0010      /* Modem section ready */
#define AC97_PWR_REF             0x0008      /* Vref nominal */
#define AC97_PWR_ANL             0x0004      /* Analog section ready */
#define AC97_PWR_DAC             0x0002      /* DAC section ready */
#define AC97_PWR_ADC             0x0001      /* ADC section ready */

/* control */
#define AC97_PWR_PR0             0x0100      /* ADC and Mux powerdown */
#define AC97_PWR_PR1             0x0200      /* DAC powerdown */
#define AC97_PWR_PR2             0x0400      /* Output mixer powerdown (Vref on) */
#define AC97_PWR_PR3             0x0800      /* Output mixer powerdown (Vref off) */
#define AC97_PWR_PR4             0x1000      /* AC-link powerdown */
#define AC97_PWR_PR5             0x2000      /* Internal Clk disable */
#define AC97_PWR_PR6             0x4000      /* HP amp powerdown */
#define AC97_PWR_PR7             0x8000      /* Modem off - if supported */

/* useful power states */
#define AC97_PWR_D0              0x0000      /* everything on */
#define AC97_PWR_D1              AC97_PWR_PR0|AC97_PWR_PR1|AC97_PWR_PR4
#define AC97_PWR_D2              AC97_PWR_PR0|AC97_PWR_PR1|AC97_PWR_PR2|AC97_PWR_PR3|AC97_PWR_PR4
#define AC97_PWR_D3              AC97_PWR_PR0|AC97_PWR_PR1|AC97_PWR_PR2|AC97_PWR_PR3|AC97_PWR_PR4
#define AC97_PWR_ANLOFF          AC97_PWR_PR2|AC97_PWR_PR3  /* analog section off */

/* Total number of defined registers.  */
#define AC97_REG_CNT 64


/*** Initialization ***/

/*
* This function is used to configure the second time multiplexed channel 
* (TDMb) of the time slot assigner (TSA) of the serial interface (SI) to
* route out data to the second serial management controller (SMC2). It
* also configures the port pins to do their special function.
*/

static void tsa_init(void)
{
  volatile	cpm8xx_t	*cp;
  volatile	immap_t		*immap;
  volatile      unsigned long   *si_ram;
	
  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* and to internal registers */
	
  immap = (immap_t *)IMAP_ADDR;	
	
  /* Program SI RAM to route first 96 bits (12 bytes) to SMC2 and to
     ignore the rest.
  */1;0c
  1;0c
  /* get 32 bit pointer */
	
  si_ram = (unsigned long *) &(cp->cp_siram);
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = SMC2, 12 byte
     CNT = 11 and BYT = 1 and NOT last entry for both RXb and TX000000b entry.
  */
	
  si_ram[32] = 0x01ae0000;
  si_ram[96] = 0x01ae0000;
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 16 byte CNT = 15 and BYT = 1 and NOT last entry for both 
vvvv1     RXb and TXb entry.
  */
	
  si_ram[33] = 0x01fe0000;
  si_ram[97] = 0x01fe0000;
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 4 byte CNT = 3 and BYT = 1 and is last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[34] = 0x018f0000;
  si_ram[98] = 0x018f0000;
	
  /* Set the SIMODE register to SMC2 connected to SI, normal mode,
     RFSDx = 01 1 bit delay, DSCx = 0 normal speed, common receive and
     transmit clocks CRTb = 1, STZb = 0 normal, CE = 0 data sent on
     rising and received on falling clock edge, FEb = 1 frame sync edge
     rising, GMb = 0, TFSDb = 01 1 bit delay.
	
     Leave setting for SMC1 as is.
  */
	
  cp->cp_simode = 0x81490000 | (cp->cp_simode & 0xffff);
	
  /* Configure PA2 as L1RCLKB, PA10 as L1RXBD, PA11 as L1TXDB and
     PC6 as L1RSYNCB 
  */
	
  immap->im_ioport.iop_papar |= 0x2030;
  immap->im_ioport.iop_padir |= 0x0030;
  immap->im_ioport.iop_padir &= 0xdfff;
  immap->im_ioport.iop_pcpar |= 0x0200;
	
  /* Set SIGMR to enable TDMb, static TDMa and TDMb RAM division */
	
  cp->cp_sigmr = 0x0a;
	
}

static init_ac97(void)
{
}

cbd_t *send_command(enum eAudioCmd, void *argval, cbd_t *bd)
{ 
  short a, b;
  short *ac97_handset, *ac97_headset, *ac97_mic, *ac97_speaker, *ac97_gpr, *ac97_pwrdn, *ac97_ac_mode, *ac97_in_mux, *ac97_rec_gain;
  ac97_slot->slot_0 = COMMAND_ONLY; 
  switch(eAudioCmd) 
   {
   case TSA_INIT_AUD_CMD:
	  tsa_init();
	  break; 
   case CHIP_INIT_AUD_CMD:
	  ac_97_init()
	  break; 
	  /* REINIT_AUD_CMD:*/

	  /* SHUTDOWN_AUD_CMD:*/

	  /* configuration commands */ 
  case LPBK_AUD_CMD: /* new for eAudCmd */
          ac97_slot->slot_1 =AC97_GENERAL_PURPOSE ;
	  *ac97_gpr &= ~AC97_LPBK;
	  if (*argv <> 0 ) *ac97_gpr| =AC97_LPBK; /* 1 = loopback is on */
	  ac97_slot->slot_2 =*ac97_gpr;
	  break;
  case RATE_ANABLE_AUD_CMD: /* new for eAudCmd */
          ac97_slot->slot_1 = AC97_EXTENDED_STATUS;
          ac97_slot->slot_2 = ~ON;
           if (*argval <> 0)  ac97_slot->slot_2 =ON; /* must be set to change sample rate  */
	  break; 
  case AD_RATE_AUD_CMD: /* new for eAudCmd */
          ac97_slot->slot_1 =AC97_PCM_LR_ADC_RATE;
	  ac97_slot->slot_2 =*argv; /* no protection for value so far */
	  break;
  case DA_RATE_AUD_CMD: /* new for eAudCmd */
          ac97_slot->slot_1 =AC97_PCM_FRONT_DAC_RATE;
	  ac97_slot->slot_2 =*argv; /* no protection for the value so far */
	  break;
  case LINE_IN_SHUTDOWN_AUD_CMD:
	  ac97_slot->slot_1 = AC97_LINE_IN_VOL;
	  *ac97_in_mux &= ~AC97_MUTE;
	  if (*argval <> 0) *ac97_in_mux |= AC97_MUTE; /* 1 = mute on*/
	  ac97_slot->slot_2 = *ac97_mux_in; /* main gate for dac */ 
	  break;
   case SET_LINE_IN_ATTEN_AUD_CMD:/* new for eAudCmd */
          ac97_slot->slot_1 = AC97_LINE_IN_VOL;
	  *ac97_in_mux &= ~( AC97_DAC_ATT_L |  AC97_DAC_ATT_R);
	  if (*argv >= 0 & *argv < 0x1f)
	    { 
	      *ac97_in_mux |= *argv;
	      b = *argv;
              b >>= 8; /* mul. with 2**8, BE?? */
	      ac97_slot->slot_2 = *ac97_in_mux |=b; 	   
            } /* wrong value gives max. gain */
	      break;
  
	  /* handset commands */
   case SHUTDOWN_SPEAKER_AUD_CMD:
	  ac97_slot->slot_1 = AC97_MASTER_VOL_STEREO;
	  *ac97_handset &= ~AC97_MUTE;
	  if (*argval <> 0) *ac97_handset |= AC97_MUTE; /* 1 = mute on*/
	  ac97_slot->slot_2 = *ac97_handset; 
	  break;
   case SET_HANDSET_ATTEN_AUD_CMD:/* new for eAudCmd */
          ac97_slot->slot_1 = AC97_MASTER_VOL_STEREO;
	  *ac97_handset &= ~(AC97_LEFTVOL | AC97_RIGHTVOL);
	  if (*argv >= 0 & *argv < 0x7f)
	    { 
	      *ac97_handset |= *argv;
	      b = *argv;
              b >>= 8; /* mul. with 2**8, BE?? */
	      ac97_slot->slot_2 = *ac97_handset |=b; 	   
            } /* wrong value gives max. gain */
	      break; 

	      /* Mic commands */
   case MIC_MUTE_AUD_CMD:/* new for eAudioCmd */
	  ac97_slot->slot_1 = AC97_MIC_VOL;
	  *ac97_mic &= ~MIC_MUTE;
	  if (*argval <> 0) *ac97_mic |= MIC_MUTE; /* 1 = mute on*/
	  ac97_slot->slot_2 =*ac97_mic; 
	  break; 
   case MIC_ATTEN_AUD_CMD: /* new for eAudioCmd */
	  ac97_slot->slot_1 = AC97_MIC_VOL;
	  *ac97_mic &= ~AC97_MICATTEN;
	  if (*argv >= 0 & *argv < 0x20) *ac97_mic+ = *argv; /* if not max. atten. */
	  ac97_slot->slot_2 =*ac97_mic;
	  break; 
   case MIC_BOOST_AUD_CMD: /* new for eAudCmd */
          ac97_slot->slot_1 = AC97_MIC_VOL;
	  *ac97_mic &= ~AC97_MICBOOST;
	  if (*argv <> 0 ) *ac97_mic| =AC97_MICBOOST;
	  ac97_slot->slot_2 =*ac97_mic;
	  break; 
   case MIC_SEL_AUD_CMD: /* new for eAudCmd */
          ac97_slot->slot_1 =AC97_GENERAL_PURPOSE;
	  *ac97_gpr &= ~AC97_MIC_SEL;
	  if (*argv <> 0 ) *ac97_gpr| =AC97_MIC_SEL; /* 0 = handset/hf,1 = headset */
	  ac97_slot->slot_2 =*ac97_gpr;
	  break;

 	  /* headset commands */
   case SHUTDOWN_HEADSET_AUD_CMD:/* new for eAudCmd */
	  ac97_slot->slot_1 = AC97_HEADPHONE_VOL;
	  *ac97_headset &= ~AC97_MUTE;
	  if (*argval <> 0)   *ac97_headset |= AC97_MUTE; /* 1 = mute on*/
	  ac97_slot->slot_2 =*ac97_headset; 
	  break;
   case SET_HEADSET_ATTEN_AUD_CMD:/* new for eAudCmd */
          ac97_slot->slot_1 = AC97_HEADPHONE_VOL;
	  *ac97_headset &= ~(AC97_LEFTVOL | AC97_RIGHTVOL);
	  if (*argv >= 0 & *argv < 0x7f)
	    { 
              *ac97_headset |= *argv;
	      b = *argv;
              b >>= 8; /* mul. with 2**8, BE?? */
	      ac97_slot->slot_2 = *ac97_haedset |=b; 	   
            } /* wrong value gives max. gain */
	  break;

	  /* speaker commands */
   case SHUTDOWN_SPEAKER_AUD_CMD:/* new for eAudCmd */
	  ac97_slot->slot_1 = AC97_MASTER_VOL_MONO;
	  *ac97_speaker &= ~AC97_MUTE;
	  if (*argval <> 0)   *ac97_speaker |= AC97_MUTE; /* 1 = mute on*/
	  ac97_slot->slot_2 =*ac97_speaker; 
v	  break;
   case SET_SPEAKER_ATTEN_AUD_CMD:/* new for eAudCmd */
          ac97_slot->slot_1 = AC97_MASTER_VOL_MONO;
	  *ac97_speaker &= ~AC97_MONOVOL;
	  if (*argv >= 0 & *argv < 0x3f)  *ac97_speaker |= *argv;
	  ac97_slot->slot_2 = *ac97_speaker; 	   
            /* wrong value gives max. gain */
	  break;

	  /* adc settings */
   case ADC_ENABLE_AUD_CMD: /* new for eAudCmd */       
          ac97_slot->slot_1 = AC97_RECORD_GAIN;
	  *ac97_rec_gain &= ~ADC_MUTE;
	  if (*argval <> 0) *ac97_rec_gain |= ADC_MUTE; /* 1 = mute on*/
	  ac97_slot->slot_2 =*ac97_rec_gain; 
	  break;
   case ADC_GAIN_AUD_CMD: /* new for eAudCmd */       
          ac97_slot->slot_1 = AC97_RECORD_GAIN;
	  *ac97_rec_gain &= ~(AC97_REC_GAIN_LEFT | AC97_REC_GAIN_RIGHT);
	  if (*argv >= 0 & *argv < 0x10)
	    { 
	      *ac97_rec_gain |= *argv;
	      b = *argv;
              b >>= 8; /* mul. with 2**8, BE?? */
	      ac97_slot->slot_2 = *ac97_rec_gain |=b; 	   
            } /* wrong value gives min. gain */
	      break;
   case ADC_MUX_SELECT_AUD_CMD: /* new for eAudCmd */       
          ac97_slot->slot_1 = AC97_RECORD_SELECT;
	  *ac97_rec_gain &= ~(AC97_REC_MUX_MIC);
	  if (*argv >= 0 & *argv < 0x7)
	    { 
	      a |= *argv; /* ne need to keep old value */
	      b = *argv;
              b >>= 8; /* mul. with 2**8, BE?? */
	      ac97_slot->slot_2 = a |=b; 	   
            } /* wrong value selects mic */
	      break;	   
   
   case SET_OUTPUT_SRC_AUD_CMD:
		break;
   case GET_FRAME_SIZE_AUD_CMD:
		break;
   case FILL_BUFFER_AUD_CMD:
		break;
   case EMPTY_BUFFER_AUD_CMD:
		break;
		/* case SET_DATA_RATE_AUD_CMD: */
   default:
	
   }
  ac97_slot->slot1 >> 1; /* first bit is r/w = 0 we do only write;
 
}
static void ac_97_init(void)
{
     /* one pulse to enable audio_chip*/
set_port_dir(PORT_C, OUTPUT_PINS, AC97_RST);
set_port(PORT_C, 0 ,AC97_RST);
set_port(PORT_C, AC97_RST, 0);

}
static structure 
{
   short slot0;
   short slot_1;
   short slot_2;
   short slot3_audio_right;
   short slot4_audio_left;
}ac97_slot;
/* this routine makes first 12 bytes for ac97_frame */
static void assembly_ac97_frame []
{
     short i, ac97_tx_data [];
     for (i=0;i<6;i++){
       switch (i){
         case 0:
	   ac_97_tx_data[i] =ac97_slot->slot0; /*first slot is ready */
	   break;
         case 1:
	   ac_97_tx_data[i] = ac97_slot->slot1; /* 4 lsb bits miss */
	   break;
         case 2:
	   ac_97_tx_data[i] = ac97_slot2 >>4; /* 4lsb bits + first 12 bits of slot2 (reg. data) */
	   break;
         case 3:
	   ac_97_tx_data[i] = ac97_slot->slot2 << 12; /* slot2 is fiinished, some space is left in this vector */
	   ac_97_tx_data[i] = ac97_slot->slot3 >>8;
	   break;
         case 4:
	   ac_97_tx_data[i] = ac97_slot->slot3 << 8;
	   ac_97_tx_data[i] = ac97_slot->slot4 >> 12; /* 4 last 0 from slot3 */
	   break;
         case 5:
	   ac_97_tx_data[i] = ac97_slot->slot4 << 4;
	   break;
       default:
       }
     }
}
static disassembly_ac97_frame ()
{
  /* byte 8 and byte 9 left adc */ 
  /* last four bits of byte 10, byte 11 and first four bits of byte 12 right adc (offset from "phy.rx_bd_base data pointer/12") */
}  

#endif
#endif










