/*
*  SNOM low level audio driver Hardware Layer Abstraction (HAL)
*  2002 snom technology AG
*
* Description:
*
* This file is one of files for the low level driver for the 
* SNOM Rev D. audio hardware.
*
* It generally defines an interface that abstracts from the underlying
* hardware (TP3070, AC97, etc). 
*
*/

typedef enum OutputSrc {
	eHandSetOutput,
	eHeadSetOutput,
	eSpeakerOutput
} eOutputSrc;

typedef enum AudioCmd {
	TSA_INIT_AUD_CMD, 
	CHIP_INIT_AUD_CMD, 
	REINIT_AUD_CMD, 
	SHUTDOWN_AUD_CMD, 
	SHUTDOWN_SPEAKER_AUD_CMD, 
	SET_MIC_GAIN_AUD_CMD, 
	SET_HANDSET_GAIN_AUD_CMD, 
	SET_HEADSET_GAIN_AUD_CMD, 
	SET_SPEAKER_GAIN_AUD_CMD, 
	SET_OUTPUT_SRC_AUD_CMD,
	GET_FRAME_SIZE_AUD_CMD,
	FILL_BUFFER_AUD_CMD,
	EMPTY_BUFFER_AUD_CMD,
	SET_DATA_RATE
} eAudioCmd;

/*
  TSA_INIT_AUD_CMD          -  initialize the TSA. No argument.
  INIT_AUD_CMD              -  initialize the HW. No argument.
  REINIT_AUD_CMD            -  reinitialize the HW. No argument. 
  MUTE_AUD_CMD              -  mute all outputs. No argument.
  SET_MIC_GAIN_AUD_CMD      -  set microphone gain. Integer value 0-100
  SET_HANDSET_GAIN_AUD_CMD  -  set handset gain. Integer value 0-100
  SET_HEADSET_GAIN_AUD_CMD  -  set headset gain. Integer value 0-100
  SET_SPEAKER_GAIN_AUD_CMD  -  set speaker gain. Integer value 0-100
  SET_OUTPUT_SRC_AUD_CMD    -  determine output. eOutputSrc is the value.
  GET_FRAME_SIZE_AUD_CMD    -  returns the number of bytes needed for a sample.
  FILL_BUFFER_AUD_CMD       -  fills CPM buffer. Generic source buffer.
  EMPTY_BUFFER_AUD_CMD      -  empties CPM buffer. Generic destination buffer.
  SET_DATA_RATE             - set rate per second. 8000 or 16000 only.

cbd_t *send_command(enum eAudioCmd, void *argval, cbd_t *bd);

/*
  Description:
	Sends a command to HW and if a CPM buffer needs to be sent on PCM
	highway then the return value will be the pointer to the buffer.

  Parameters:
	eAudioCmd specifies the command - see above
	argval - is normally y

*/


