/***********************************************************************

 This module is used to access the ports of the MPC860/855 and the 
 control and status register of the SNOM1 board. It provides
 an abstraction layer above the different physical layers. 

************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <asm/mpc8xx.h>
#include <asm/port.h>

#include "commproc.h"

/**** Flags */

#define PARANOID 1

/*** Procedures */

static volatile immap_t	  *immap = (immap_t *)IMAP_ADDR;

static unsigned char snom_csr_val = (unsigned char) 0; 

int read_port(t_port port)
{
  volatile unsigned short s;
  volatile unsigned int   n;

  switch(port)
    {
    case PORT_A:
      s = immap->im_ioport.iop_padat;
      return s;
    case PORT_B:
      n = immap->im_cpm.cp_pbdat; 
      return n;
    case PORT_C:
      s = immap->im_ioport.iop_pcdat; 
      return s;
    case PORT_SNOM_CSR:
      return (int) snom_csr_val;
    default:
      return -1;
    }
}

void set_port(t_port port, int set_bit_mask, int clear_bit_mask)
{
  volatile unsigned short *s;
  volatile unsigned int   *n;

  switch(port)
    {
    case PORT_A:
      s = &(immap->im_ioport.iop_padat);
      *s = (*s | set_bit_mask) & ~clear_bit_mask;
      break;
    case PORT_B:
      n = &(immap->im_cpm.cp_pbdat); 
      *n = (*n | set_bit_mask) & ~clear_bit_mask;
      break;
    case PORT_C:
      s = &(immap->im_ioport.iop_pcdat); 
      *s = (*s | set_bit_mask) & ~clear_bit_mask;
      break;
    case PORT_SNOM_CSR:
      snom_csr_val = (snom_csr_val | set_bit_mask) & ~clear_bit_mask;
      *((unsigned char *) SNOM_CSR_BASE_ADDR) = snom_csr_val;
      break;
    }
} 

void set_port_ex(t_port port, int value, int mask)
{
  volatile unsigned short *s;
  volatile unsigned int   *n;

  switch(port)
    {
    case PORT_A:
      s = &(immap->im_ioport.iop_padat);
      *s = (*s & ~mask) | (value & mask);
      break;
    case PORT_B:
      n = &(immap->im_cpm.cp_pbdat); 
      *n = (*n & ~mask) | (value & mask);
      break;
    case PORT_C:
      s = &(immap->im_ioport.iop_pcdat); 
      *s = (*s & ~mask) | (value & mask);
      break;
    case PORT_SNOM_CSR:
      snom_csr_val = (snom_csr_val & ~mask) | (value & mask);
      *((unsigned char *) SNOM_CSR_BASE_ADDR) = snom_csr_val;
      break;
    }
} 

/* Sets the port pins to GPIO and to the given direction. */
void set_port_dir(t_port port, t_port_dir port_dir, int bit_mask)
{
  volatile unsigned short *sdir, *spar;
  volatile unsigned int   *ndir, *npar;

  switch(port)
    {
    case PORT_A:
      {
	spar = &(immap->im_ioport.iop_papar); 
	sdir = &(immap->im_ioport.iop_padir); 
	*spar &= ~bit_mask;
	if(port_dir == OUTPUT_PINS)
	  *sdir |= bit_mask; 
	else
	  *sdir &= ~bit_mask;
	break;
      }
    case PORT_B:
      {
	npar = &(immap->im_cpm.cp_pbpar); 
	ndir = &(immap->im_cpm.cp_pbdir); 
	*npar &= ~bit_mask;
	if(port_dir == OUTPUT_PINS)
	  *ndir |= bit_mask; 
	else
	  *ndir &= ~bit_mask;
	break;
      }
    case PORT_C:
      {
	spar = &(immap->im_ioport.iop_pcpar); 
	sdir = &(immap->im_ioport.iop_pcdir); 
	*spar &= ~bit_mask;
	if(port_dir == OUTPUT_PINS)
	  *sdir |= bit_mask; 
	else
	  *sdir &= ~bit_mask;
	immap->im_ioport.iop_pcso &= ~bit_mask; 
	immap->im_ioport.iop_pcint &= ~bit_mask; 
	break;
      }
    case PORT_SNOM_CSR:  /* not possible to influence this */
    default:
      break;
    }
}

