/*
 * SNOM low level audio driver for MPC860 SMC2 1999-2003 snom
 * technology AG
 *
 * I used the uart.c driver from Dan Malek who in turn used the
 * serial.c driver. Theiver was originally written for TP3070, and
 * then was changed to AC97.
 *
 * This module exports the following snom_aud io functions:
 *
 *	int snom_aud_init(void);
 * 
 * Description:
 *
 * This file is the low level driver for the SNOM audio hardware. Its
 * core element is a AC97 compliant codec that is addressed with a
 * SCC1.
 * */

#include <linux/config.h>
#include <linux/types.h>

#if defined (CONFIG_SOUND_SNOM) || defined (CONFIG_SOUND_SNOM2)

#include <asm/mpc8xx.h>

/* Have to check out what we need here ... */
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/soundcard.h>
#include <linux/delay.h>
#include <linux/in.h>
#include <linux/udp.h>
#include <linux/pkt_sched.h>
#include <net/sock.h>
#include <net/ip.h>
#include <net/udp.h>
#include <asm/uaccess.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>
#include <asm/checksum.h>
#include "commproc.h"
#include "snomaud.h"
#include "mec.h"

#include <asm/port.h>

#define _INLINE_ inline

#define MAX_PACKET_SIZE 320

/* forward declarations for snomcodec.c: */

#ifndef CONFIG_SNOM100
static void set_ac97_reg(int, unsigned long);
#endif
static void write_to_hardware(unsigned char *ulaw, short *linear);

extern int snom_aud_encode(int channel, int type, short *src, int length, unsigned char *dst, int maxlen);
extern int snom_aud_decode(int channel, int type, unsigned char *src, int length, short *dst, int maxlen);
extern int snom_aud_fill(int channel, short *dst, int maxlen);
extern void snom_aud_get_encoded_length(int type, int length, int *encoded_length, int *unencoded_length);
extern void snom_aud_reset_codec(int channel);

extern const unsigned char snom_aud_linear2ulaw[16384];
extern const short snom_aud_ulaw2linear[256];
#ifndef CONFIG_SNOM100
static void init_ac97(void);
static void reset_ac97 (void);
static void hardware_platform(void);
#endif

extern void snom_aud_set_audio (unsigned int audio_settings); 
void set_handset (unsigned int aValue);
void set_ringer(unsigned int aValue); 
void set_headset(unsigned int aValue);
void set_loopback(unsigned int aValue);
void set_handset_mic(unsigned int aValue);
void set_headset_mic(unsigned int aValue);
void set_speaker_phone(unsigned int aValue);
unsigned int reverse_volume(unsigned int aValue);
unsigned short adjust_adc_gain(unsigned int volume);
void set_external_io(unsigned int aValue);
extern int global_serial_num;
extern unsigned int hardware_audio_device;

static void jit_recv_reset(int c);
static void jit_recv_rtp(int c, unsigned char *content, int len);

struct dtmf_state_t {
  int key;
  int pos0;
  int pos1;
};

static void set_dtmf(int key, struct dtmf_state_t *dtmf);

#ifdef CONFIG_SNOM100
#define CPM_BUF_SIZE 80
#else
#define CPM_BUF_SIZE (CONFIG_SOUND_SNOM_BYTES_PER_SAMPLE * CONFIG_SOUND_SNOM_SAMPLES_PER_8KHZ_SAMPLE * \
	CONFIG_SOUND_SNOM_AUD_BLOCK_SIZE)
#endif //conif_snom100

/* Shutdown control pins on port C. */
 
#ifndef CONFIG_SNOM100

#define AC97_RST (1<<10)
#define SHUTDOWN_SPEAKER_PIN (1<<2)
#define HEADSET_LED (1<<3)

#define GPIO0  (1<<7) 
#define GPIO1  (1<<6) 
#define GPIO2  (1<<5) 
#define GPIO3  (1<<4) 
#define GPIO4  (1<<3) 
#define GPIO5  (1<<2)
#define GPIO6  (1<<1)
#define GPIO7  (1<<0)

#ifdef CONFIG_SNOM105

#define WARM_RESET  	GPIO0 // warm reset to codec  (for cpld)
#define CLK_EN_48KHZ	GPIO1 // clk enable for 48 Khz counter, altera
#define MIC_CONTROL 	GPIO3 // handset/headset mic control
#define VCC		GPIO2 // always high output from altera, wired to vcc

#else // CONFIG_SNOM105

#define WARM_RESET  	GPIO4 // warm reset to codec  (for cpld)
#define CLK_EN_48KHZ	GPIO5 // clk enable for 48 Khz counter, altera
#define MIC_CONTROL 	GPIO6 // handsfree mic control
#define VCC		GPIO7 // always high output from altera, wired to vcc

#endif // CONFIG_SNOM105


#ifdef CONFIG_SNOM105
#define ADC_GAIN 0x800d
#define HANDSET_DAC_GAIN 0x0505
#define DAC_GAIN_SPHONE 0x000c
#define DAC_RINGER 0x000b
#else
#define ADC_GAIN 0x8006
#define HANDSET_DAC_GAIN 0x0606
#define DAC_GAIN_SPHONE 0x000b
#define DAC_RINGER 0x000b
#define DAC_GAIN_SPHONE_A 0x000d
#endif

#define HANDSET_MIC 0x0000
#define HEADSET_MIC 0x0100
#define LPBK 0x0080
#define SIDE_TONE 0x000a
#define MIC_BOOST 0x0040
#define SOUND_3D 0x2000
#define HEADSET_RJ14 0x8000
#define HEADSET_PC 0x0000

#define HEADSET_IN (1<<3) // pc12 
static unsigned short ac97_speaker = 0x8000;
static unsigned short ac97_handset = 0x8000;
static unsigned short ac97_headset = 0x8000;
static unsigned short ac97_adc = ADC_GAIN;
static unsigned short ac97_dac = DAC_GAIN_SPHONE | 0x8000;// closed at start up
static unsigned short ac97_gp = HANDSET_MIC;
static unsigned short ac97_headset_mic = HEADSET_PC;
static unsigned short ac97_adc_handset;
static unsigned short ac97_adc_sphone;
static unsigned short ac97_dac_sphone;
static unsigned short ac97_mic = SIDE_TONE;// only for side tone
#endif

#ifdef CONFIG_SNOM100
#define SHUTDOWN_PIN (1<<3)
#define SHUTDOWN_SPEAKER_PIN (1<<2)
#define SNOM100_TX_GAIN 230
#define SNOM100_HANDSET_GAIN 160
#define SNOM100_SPEAKER_GAIN 100
#define SNOM100_TX_GAIN_SPEKAER_PHONE 210
#endif

#define MODE_WR 2 /* I dont understand this, anyway. */

#define HW_NUM_BLOCKS CONFIG_SOUND_SNOM_HW_NUM_BLOCKS /* should be 2 */
#define SW_NUM_BLOCKS 32

/* buffers for the audio streams: */
static short au_buffer[SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE];
static unsigned int local_time = 0; /* timestamp on the local machine */
static int tos = IPTOS_PREC_CRITIC_ECP;
// static unsigned int send_time = 0; /* time of last packet sent */
static unsigned int au_rd_pos = 0;
static unsigned int au_wr_pos = 0;

/* statistics for the jitter buffer */
#define WRITE_ADVANCE 160
#define JITTER_SIZE 32

static struct socket *rtp_socket[MAX_NUM_RTP] = { 0 };
static struct socket *rtcp_socket[MAX_NUM_RTP] = { 0 };
static int local_port[MAX_NUM_RTP] = { 0 };
static unsigned int packet_size[MAX_NUM_RTP] = { 160 };
static unsigned int packet_type[MAX_NUM_RTP] = { 0 }; /* ulaw */
static unsigned long local_ssrc[MAX_NUM_RTP]; /* ssrc identifier */
static int need_marker[MAX_NUM_RTP] = { 2 }; /* bool */
static unsigned int packet_seqnr[MAX_NUM_RTP] = { 0 };
static struct sockaddr_in remote_adr[MAX_NUM_RTP];

static _INLINE_ void check_c(char *s, int c) {
  //   if (c < 0 || c >= MAX_NUM_RTP) printk("%s: %d\n", s, c);
}

// check if we should do AGC (which is the case if 
static int do_agc = 1;
static void check_for_agc(void) {
  int i;
  do_agc = 1;
  for(i = 0; i < MAX_NUM_RTP; i++) {
    if (packet_type[i] == 18) {
      do_agc = 0;
      break;
    }
  }
}

// check if we are ready for sending and receiving:
static _INLINE_ int jit_rtp_active(int c) {
  // say if a channel is sending or not:
  check_c("jit_rtp_active", c);
  return rtp_socket[c] && remote_adr[c].sin_addr.s_addr != 0;
}

// we might not be ready for sending, but for receiving:
static _INLINE_ int jit_rtp_listen(int c) {
  // say if a channel is sending or not:
  check_c("jit_rtp_listen", c);
  return rtp_socket[c] != 0;
}

static int mixer_mode = SNOM_AUD_MIXER_RTP_AUDIO;
static int half_duplex = 0;

/* The bloody half duplex mode is now implemented as state
   machine. That makes the transition from the speaker roles more
   smooth and easier to understand in software. */
#define HD_SEND 0 // Microphone on, speaker off
#define HD_RECV 1 // Speaker turned on, Microphone off
static int duplex_state = HD_SEND;


#define AUDIO_STARTING 0
#define AUDIO_RUNNING 1
static int audio_state = AUDIO_STARTING;

#define CLIP(s) ((short)((s) > 32767 ? 32767 : ((s) < -32767 ? -32767 : (s))))

// intermediate buffer for decoded data:
static short jit_decode_buffer[MAX_NUM_RTP][320]; // max GSM frame = 320 samples
static int jit_decode_pos[MAX_NUM_RTP];
static int jit_decode_len[MAX_NUM_RTP];
static int jit_decode_start[MAX_NUM_RTP];

// jitter buffer for incoming packets:
static unsigned char jit_recv_buffer[MAX_NUM_RTP][JITTER_SIZE][320];
static int jit_recv_len[MAX_NUM_RTP][JITTER_SIZE];
static int jit_recv_type[MAX_NUM_RTP][JITTER_SIZE]; // (0..255, includes marker)
static int jit_recv_timestamp[MAX_NUM_RTP][JITTER_SIZE];

// intermedia buffer for unencoded outgoing data:
static short jit_send_buffer[MAX_NUM_RTP][640];
static int jit_send_len[MAX_NUM_RTP];

// buffer for encoded outgoing data:
static unsigned char jit_encoded_buffer[MAX_NUM_RTP][640 + 12]; // 12 bytes for rtp header
static int jit_encoded_len[MAX_NUM_RTP]; // in encoded form
static int jit_unencoded_len[MAX_NUM_RTP]; // in encoded form
static int jit_encoded_ts[MAX_NUM_RTP];

// statistics on the energy levels presented to speaker and from mic:
static unsigned long energy_lc = 0; // local current
static unsigned long energy_lf = 0; // local filtered
static unsigned long energy_rc = 0; // remote current
static unsigned long energy_rf = 0; // remote filtered

/* Now come the devices ... */

typedef struct {
	
  /* CPM Buffer Descriptor pointers.*/
  cbd_t		     *rx_bd_base;
  cbd_t		     *rx_cur;
  cbd_t		     *tx_bd_base;
  cbd_t		     *tx_cur;
	
  /* Underruns */
  int                is_underrun;
	
} snom_aud_physical_dev;

typedef struct {
  int is_open;
} snom_aud_info_t;

static snom_aud_info_t info;
static char *snom_aud_name = "CPM SNOM low level audio driver";
static snom_aud_physical_dev phy;
static int snom_aud_minor_num = 0;
static int snom_aud_mode = 0;

static void (*snom_aud_call_back)(void);

/* Forward decl. */

void snom_aud_reset (snom_aud_info_t *info);
void snom_aud_set_tx_gain2(int level);

#define CPM_RX_ONLY 11
#define CPM_TX_ONLY 22
#define CPM_RXTX 33

#ifdef CONFIG_SNOM100
static _INLINE_ void restart_cpm(int eMode)
{
  /* Get pointer to Communication Processor */
	
  volatile cpm8xx_t *cp = cpmp;
  volatile smc_t    *sp = &cp->cp_smc[1];
	
#ifdef SNOM_AUD_DEBUG_BUFFERS
  printk("\nr");
#endif
	
  /* Send the CPM a stop transmitter command. They say it is not
	really necessary but who knows. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SMC2, CPM_CR_STOP_TX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* reset transmitter and receiver by disabling it. */
	
  sp->smc_smcmr &= 0xfffc;

  /* Enable all interrupts and clear all pending events. */
	
  sp->smc_smcm |= SMCM_TXE | SMCM_BSY | SMCM_TX | SMCM_RX;
  sp->smc_smce = 0xff;
	
  /* Send the CPM an initialize transmitter command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SMC2, CPM_CR_INIT_TRX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  phy.is_underrun = 0;
	
  /* Clear pending TX interrupts. */
	
  sp->smc_smce = 0x12;
	
  /* Send the CPM a restart transmitter command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SMC2, CPM_CR_RESTART_TX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* Enable it ... */
	
  switch(eMode) {
  case CPM_RX_ONLY :
    sp->smc_smcmr |= 0001;
    break;
  case CPM_TX_ONLY :
    sp->smc_smcmr |= 0002;
    break;
  case CPM_RXTX :
    sp->smc_smcmr |= 0003;
    break;
  }
}
#else //snom105, snom200
static _INLINE_ void restart_cpm(int eMode)
{
  /* Get pointer to Communication Processor */

  volatile cpm8xx_t *cp = cpmp;
  volatile scc_t    *sp = &cp->cp_scc[0];
	
  // printk("\ns");
       
  /* Send the CPM a stop transmitter command. They say it is not
     really necessary but who knows. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_STOP_TX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* reset transmitter and receiver by disabling it. */
	
  sp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);

  /* Enable all interrupts and clear all pending events. */
	
  sp->scc_sccm |= SCCM_GLR | SCCM_GLT | SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX;
  sp->scc_scce = 0xffff;
	
  /* Send the CPM an initialize transmitter command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_INIT_TRX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  phy.is_underrun = 0;
	
  /* Clear pending TX interrupts. */
	
  sp->scc_scce = 0xffff;
	
  /* Send the CPM a restart transmitter command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_RESTART_TX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* Enable it ... */
	
  switch(eMode) {
  case CPM_RX_ONLY :
    sp->scc_gsmrl |= SCC_GSMRL_ENR;
    break;
  case CPM_TX_ONLY :
    sp->scc_gsmrl |= SCC_GSMRL_ENT;
    break;
  case CPM_RXTX :
    sp->scc_gsmrl |= SCC_GSMRL_ENR | SCC_GSMRL_ENT;
    break;
  }
}
#endif  //snom105, snom200

/*
 * Automatic Gain Control (AGC). The algorithm we use here reduces the
 * gain immediately when we are reaching the threshold value
 * (MAX_SHORT). Otherwise, the gain is increased every 100 ms if one
 * value was above a threshold (speech detection).
 */ 

static long agc_gain = 32768 / 4; // 32768 means times 4, the maximum amplification

static void init_agc(void) {
  agc_gain = 32768 / 4;
}

// AGC for 80 samples:
static _INLINE_ void agc_data(short *sample) {
  int idx;
  const int old_gain = agc_gain;

  // This factor is used to make the step from one gain factor to
  // another more smooth:
  static const long factor1[80] = {
    0, 410, 819, 1229, 1638, 2048, 2458, 2867, 3277, 3686,
    4096, 4506, 4915, 5325, 5734, 6144, 6554, 6963, 7373, 7782,
    8192, 8602, 9011, 9421, 9830, 10240, 10650, 11059, 11469, 11878,
    12288, 12698, 13107, 13517, 13926, 14336, 14746, 15155, 15565, 15974,
    16384, 16794, 17203, 17613, 18022, 18432, 18842, 19251, 19661, 20070,
    20480, 20890, 21299, 21709, 22118, 22528, 22938, 23347, 23757, 24166,
    24576, 24986, 25395, 25805, 26214, 26624, 27034, 27443, 27853, 28262,
    28672, 29082, 29491, 29901, 30310, 30720, 31130, 31539, 31949, 32358
  };

  // gain table: transform raw energy (divided by 0x400000):

  static const long gain_table[] = { 
    
    32768, 32768, 32768, 32768, 32768, 32768, 31769, 30992, 30366,
    29848, 29409, 29031, 28702, 28412, 28153, 27921, 27710, 27274,
    26875, 26508, 26169, 25854, 25561, 25287, 25031, 24790, 24563,
    24349, 24146, 23954, 23772, 23598, 23433, 23276, 23125, 22981,
    22843, 22710, 22583, 22461, 22343, 22230, 22121, 22015, 21913,
    21815, 21720, 21628, 21538, 21452, 21368, 21287, 21208, 21131,
    21056, 20984, 20913, 20844, 20777, 20712, 20648, 20586, 20526,
    20467, 20409, 20352, 20297, 20243, 20191, 20139, 20089, 20040,
    19991, 19944, 19898, 19852, 19808, 19764, 19721, 19679, 19638,
    19598, 19558, 19519, 19481, 19443, 19406, 19370, 19334, 19299,
    19264, 19231, 19197, 19164, 19132, 19100, 19069, 19038, 19008,
    18978, 18949, 18920, 18891, 18863, 18835, 18808, 18781, 18754,
    18728, 18702, 18677, 18652, 18627, 18602, 18578, 18554, 18531,
    18507, 18485, 18462, 18440, 18417, 18396, 18374, 18353, 18332,
    18311, 18291, 18270, 18250, 18230, 18211, 18191, 18172, 18153,
    18135, 18116, 18098, 18080, 18062, 18044, 18026, 18009, 17992,
    17975, 17958, 17941, 17925, 17909, 17893, 17877, 17861, 17845,
    17830, 17814, 17799, 17784, 17769, 17754, 17740, 17725, 17711,
    17697, 17682, 17668, 17655, 17641, 17627, 17614, 17601, 17587,
    17574, 17561, 17548, 17535, 17523, 17510, 17498, 17485, 17473,
    17461, 17449, 17437, 17425, 17413, 17402, 17390, 17379, 17367,
    17356, 17345, 17334, 17323, 17312, 17301, 17290, 17280, 17269,
    17258, 17248, 17238, 17227, 17217, 17207, 17197, 17187, 17177,
    17167, 17157, 17148, 17138, 17129, 17119, 17110, 17100, 17091,
    17082, 17073, 17063, 17054, 17045, 17036, 17028, 17019, 17010,
    17001, 16993, 16984, 16976, 16967, 16959, 16950, 16942, 16934,
    16926, 16918, 16910, 16902, 16894, 16886, 16878, 16870, 16862,
    16854, 16847, 16839, 16831, 16824, 16816, 16809, 16802, 16794,
    16787, 16780, 16772, 16765, 16758, 16751, 16744, 16737, 16730,
    16723, 16716, 16709, 16702, 16695, 16689, 16682, 16675, 16669,
    16662, 16656, 16649, 16643, 16636, 16630, 16623, 16617, 16611,
    16604, 16598, 16592, 16586, 16579, 16573, 16567, 16561, 16555,
    16549, 16543, 16537, 16531, 16526, 16520, 16514, 16508, 16502,
    16497, 16491, 16485, 16480, 16474, 16468, 16463, 16457, 16452,
    16446, 16441, 16436, 16430, 16425, 16419, 16414, 16409, 16404,
    16398, 16393, 16388, 16384 
  };

#define XSCALE 1048576

  // update the energy if the current energy is at least 25 % of the filtered energy:
  if (energy_lc > energy_lf / 4) {
    int index = energy_lf / XSCALE;
    int fraction = energy_lf - (index * XSCALE);
    if (index < sizeof(gain_table) / sizeof(gain_table[0]) - 1) {
      agc_gain = gain_table[index] + ((gain_table[index + 1] - gain_table[index]) * (fraction / 16)) / (XSCALE / 16);
    }
    else {
      // lowest amplification (saturation):
      agc_gain = gain_table[sizeof(gain_table) / sizeof(gain_table[0]) - 1];
    }
  }

  // amplify the samples according to the gain:
  for(idx = 0; idx < 80; idx++) {
    const long fac = factor1[idx];
    const int mix = (fac * agc_gain + (32768 - fac) * old_gain) / 32768;
    const long result = (mix * (long)(sample[idx])) / (32768 / 2);
    sample[idx] = CLIP(result);
  }  
}


/******************************************************************************/

/*
 * Here are the routines that are called from ioctl().
 */

int snom_aud_get_blocksize     (snom_aud_info_t *info){
  return SNOM_AUD_BLOCK_SIZE;
}

int snom_aud_get_caps          (snom_aud_info_t *info) {
	
  /* TBD: is real time correct ? I do not know the exact associated semantics ... */
	
  return DSP_CAP_DUPLEX | DSP_CAP_BATCH | DSP_CAP_REALTIME; 
}

int  snom_aud_get_formats       (snom_aud_info_t *info) {
  return AFMT_S16_BE;
}

void snom_aud_set_blocking_mode (snom_aud_info_t *info, int f_flags) {
  /* the device never blocks */
}

int  snom_aud_set_format        (snom_aud_info_t *info, int format) {
  if(format == AFMT_S16_BE) {
    return 0;
  } else {
    return -EINVAL;
  }
}
 
#ifdef CONFIG_SNOM100
// old functions
void snom_aud_shutdown(int bTrue)
{}
void snom_aud_set_amp_gain(int level)
{}
static void pipe_out_byte(unsigned char ucVal);
void snom_aud_set_rx_gain(int level)
{}
void snom_aud_set_tx_gain2(int level)
{}
void snom_aud_set_tx_gain(int level)
{}
void snom_aud_shutdown_speaker(int bTrue)
{}
// end of old functions

#define AUDIO_VOL 0x0000000f // volume of left speaker is ignored at the moment
#define AUDIO_DEV 0x00000f00
void snom_aud_set_audio (unsigned int audio_settings)
{
  unsigned int  volume;
  hardware_audio_device = (audio_settings & AUDIO_DEV) >> 8;
  volume = audio_settings & AUDIO_VOL;

  // printk("snom_aud_set_audio: %0xd dev:%d vol:%d\n", audio_settings, device, volume); // use carefully, because its so slow

  switch(hardware_audio_device){
  case 0: set_handset (volume); break; // handset
  case 1: set_ringer(volume);break; // casing speaker
  case 2: set_headset(volume); break; // headset
  case 3: set_loopback(volume); break; // loopback
  case 4: set_handset_mic(volume);break; // mic selection
  case 5: set_headset_mic(volume);break;
  case 6: set_speaker_phone(volume);break;//handsfree mic
  case 7: set_external_io(volume);break;//external audio input
  default:
  }
}
void set_handset (unsigned int volume)
{
  if(!volume) {
    set_port(PORT_C,SHUTDOWN_PIN,0); 
  }
  else {
    if(volume < 0) volume = 0;
    if(volume > 0xf) volume = 0xf;
    switch(volume) {
      case 0:  volume = 0xf0; break;
      case 1:  volume = 0x70; break;
      case 2:  volume = 0xb0; break;
      case 3:  volume = 0x30; break;
      case 4:  volume = 0xd0; break;
      case 5:  volume = 0x50; break;
      case 6:  volume = 0x90; break;
      case 7:  volume = 0x10; break;
      case 8:  volume = 0xe0; break;
      case 9:  volume = 0x60; break;
      case 10: volume = 0xa0; break;
      case 11: volume = 0x20; break;
      case 12: volume = 0xc0; break;
      case 13: volume = 0x40; break;
      case 14: volume = 0x80; break;
      case 15: volume = 0x00; break;
    }
    set_port (PORT_SNOM_CSR, volume, ~volume & 0xf0);//change the value on lm4830
    pipe_out_byte(0x22);
    pipe_out_byte(SNOM100_HANDSET_GAIN);//fixed gain for TP3070
    set_port(PORT_C,0,SHUTDOWN_PIN);
  }
}
void set_ringer(unsigned int volume){
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);
  }
  else {
    switch(volume) {
      case 0:  volume = 30; break;
      case 1:  volume = 40; break;
      case 2:  volume = 50; break;
      case 3:  volume = 60; break;
      case 4:  volume = 70; break;
      case 5:  volume = 80; break;
      case 6:  volume = 90; break;
      case 7:  volume = 100; break;
      case 8:  volume = 110; break;
      case 9:  volume = 120; break;
      case 10: volume = 130; break;
      case 11: volume = 140; break;
      case 12: volume = 150; break;
      case 13: volume = 160; break;
      case 14: volume = 170; break;
      case 15: volume = 180; break;
    }
    pipe_out_byte(0x22);
    pipe_out_byte(volume); //change gain only on TP3070
    set_port(PORT_C,0,SHUTDOWN_SPEAKER_PIN);
 }
}
void set_handset_mic(unsigned int volume){
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_PIN, 0);
    }
  else {
    pipe_out_byte(0x2A);
    pipe_out_byte(SNOM100_TX_GAIN);
    set_port(PORT_C,0,SHUTDOWN_PIN);
  }
}
void set_speaker_phone(unsigned int volume){
  half_duplex = volume > 0;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0); //TPA721 OFF
    set_port(PORT_C, SHUTDOWN_PIN, 0);// LM4830 OFF
  }
  else {
    switch(volume) {
      case 0:  volume = 30; break;
      case 1:  volume = 40; break;
      case 2:  volume = 50; break;
      case 3:  volume = 60; break;
      case 4:  volume = 70; break;
      case 5:  volume = 80; break;
      case 6:  volume = 90; break;
      case 7:  volume = 100; break;
      case 8:  volume = 110; break;
      case 9:  volume = 120; break;
      case 10: volume = 130; break;
      case 11: volume = 140; break;
      case 12: volume = 150; break;
      case 13: volume = 160; break;
      case 14: volume = 170; break;
      case 15: volume = 180; break;
    }
    //rx, casing speaker
    pipe_out_byte(0x22);
    pipe_out_byte(volume);
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);
    //tx part hf mic
    pipe_out_byte(0x2A);
    pipe_out_byte(SNOM100_TX_GAIN_SPEKAER_PHONE);
    set_port(PORT_C,0,SHUTDOWN_PIN);
  }
}
void set_external_io(unsigned int volume){};
void set_headset_mic(unsigned int volume){};
void set_headset(unsigned int volume){};
void set_loopback(unsigned int volume){};

#else //if def SNOM_100

#define AUDIO_VOL 0x0000000f // volume of left speaker is ignored at the moment
#define AUDIO_DEV 0x00000f00

void snom_aud_set_audio (unsigned int audio_settings)
{
  unsigned int  volume;
  hardware_audio_device = (audio_settings & AUDIO_DEV) >> 8;
  volume = audio_settings & AUDIO_VOL;

  // printk("snom_aud_set_audio: %0xd dev:%d vol:%d\n", audio_settings, device, volume); // use carefully, because its so slow

  switch(hardware_audio_device){
  case 0: set_handset (volume); break; // handset
  case 1: set_ringer(volume);break; // casing speaker
  case 2: set_headset(volume); break; // headset
  case 3: set_loopback(volume); break; // loopback
  case 4: set_handset_mic(volume);break; // mic selection
  case 5: set_headset_mic(volume);break;
  case 6: set_speaker_phone(volume);break;//handsfree mic
  case 7: set_external_io(volume);break;//external audio input
  default:
  }
}

unsigned int reverse_volume(unsigned int volume)
{
  unsigned int ac97_value;

  if (volume < 0) volume = 1;
  if (volume > 15) volume = 15;
 
  ac97_value = 15 - volume;

  return ac97_value; 
}

unsigned short adjust_adc_gain(unsigned int volume)
{
  unsigned short ac97_value;
  switch(volume){
  case 1: ac97_value = 5-4;break; 
  case 2: ac97_value = 6-4;break; //default 
  case 3: ac97_value = 7-4;break;
  case 4: ac97_value = 8-4;break; 
  default:ac97_value = 6-4;break;
  }
  return ac97_value;
}


void set_handset(unsigned int volume)
{
  ac97_handset &= ~0x8000; 
  if(!volume) {
    ac97_handset |= 0x8000;//speaker  mute bit set
    ac97_dac |= 0x8000;
  }
  else {
    ac97_handset = reverse_volume(volume) | (reverse_volume(volume)) << 8;
    ac97_dac = HANDSET_DAC_GAIN;       
  }
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x02,ac97_handset); // mute on/off
}

void hardware_platform(void)
{
#ifdef CONFIG_SNOM105
  //nothing to do
#else
if(global_serial_num >= SNOM200_B){
      ac97_adc_handset = 0x6;
      ac97_adc_sphone = 0xa;
      ac97_dac_sphone = DAC_GAIN_SPHONE;
    }
    else{
      ac97_adc_handset = 0x1;
      ac97_adc_sphone = 0x6;
      ac97_dac_sphone = DAC_GAIN_SPHONE_A;
    }
#endif //CONFIG_SNOM105
}

#ifdef CONFIG_SNOM105
void set_ringer(unsigned int volume)
{
  ac97_speaker &= ~0x8000;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;
    set_ac97_reg(0x18,ac97_dac);// dac preamp closed
    set_ac97_reg(0x06, ac97_speaker); //mute on sent
    ac97_mic &= ~0x8000; //sidetone enabled
  }
  else {  
    ac97_dac = DAC_RINGER;
    set_ac97_reg(0x18,ac97_dac);// new value for the ringer
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);    
    set_ac97_reg(0x06,ac97_speaker); //mute off, new value 
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is on
 }
  set_ac97_reg(0xe, ac97_mic);
}

void set_handset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set     
  }
  else {
    //if (volume < 0 || volume > 4) return;
    // ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= ac97_adc_handset; 
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic selected
    set_port(PORT_SNOM_CSR, 0, MIC_CONTROL);//handset on
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
}

void set_headset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    set_port(PORT_SNOM_CSR, 0,MIC_CONTROL);//handset 
  }
  else {
  if (volume < 0 || volume > 4) return;
   //ac97_adc |= adjust_adc_gain(volume);
   ac97_adc |= 4; // cs: lowest possible amplification for wide linear area
   set_port(PORT_SNOM_CSR, MIC_CONTROL, 0);// Headset is on     
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) dis/en
}

void set_external_io(unsigned int volume)
{}// only for snom200, snom220 

void set_speaker_phone(unsigned int volume)
{
 half_duplex = volume > 0;
 ac97_speaker &= ~0x8000;
 ac97_adc &= ~0x800f;
  if(!volume){
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;
    set_ac97_reg(0x18,ac97_dac);// returning old value for speakerphone
    set_ac97_reg(0x06, ac97_speaker); //mute on sent
    ac97_mic &= ~0x8000; //sidetone enabled
    ac97_adc |= 0x8000; // adc (mic) mute bit set      
  }
  else {
    if (volume < 0 || volume > 4) return;
    // ac97_adc |= adjust_adc_gain(volume);
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);    
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is on
    set_ac97_reg(0x06,ac97_speaker); //mute off, new value
    ac97_adc |= 0x7; // cs: lowest possible amplification for wide linear area
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic selected
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
  set_ac97_reg(0xe, ac97_mic);
}
void set_headset(unsigned int volume)
{
  ac97_headset &= ~0x8000; 
  if(!volume) {
    ac97_headset |= 0x8000;//speaker  mute bit set
    ac97_dac = 0x8000;//preamp dac mute bit set
  }
  else {
    unsigned int vol_left, vol_right;
    volume = reverse_volume(volume);
    vol_left  = volume;
    vol_right = volume;
    volume = (vol_left << 8) | vol_right;
    ac97_headset = volume;    
    ac97_dac = HANDSET_DAC_GAIN;       
  }    
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x04,ac97_headset); // mute on/off
}
#else //ifdef snom105
void set_ringer(unsigned int volume)
{
  ac97_speaker &= ~0x8000;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac = 0x8000;// mute on preamp
    ac97_mic &= ~0x8000; //sidetone enabled
    set_ac97_reg(0x06, ac97_speaker);// mute on
    set_ac97_reg(0x18,ac97_dac);// sent
  }
  else {  
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);
    ac97_dac = DAC_RINGER;
    set_ac97_reg(0x18,ac97_dac);// sent
    set_ac97_reg(0x06, ac97_speaker); //speaker opened
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is open
  }
  set_ac97_reg(0xe, ac97_mic);//side tone on/off sent
}

void set_handset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set     
  }
  else {
    //if (volume < 0 || volume > 4) return;
    // ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= ac97_adc_handset; //see hardware plaform for the value
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic selected
    set_port(PORT_SNOM_CSR, 0, MIC_CONTROL);//handset on
    //ac97_mic &= ~MIC_BOOST; //mic boost disabled
    //set_ac97_reg(0xe, ac97_mic); 
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
}
void set_headset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  ac97_headset_mic &= ~0x8000;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    ac97_gp &= ~HEADSET_MIC; //headset_mic dis
    set_port(PORT_B, 0,HEADSET_LED);//disconnected 
  }
  else {
    if (volume < 0 || volume > 4) return;
    //ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= 0x6;
    set_port(PORT_B, HEADSET_LED, 0);// connected
    ac97_gp |= HEADSET_MIC;
    ac97_headset_mic |= HEADSET_RJ14;
  }
  set_ac97_reg(0x20, ac97_gp);
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) dis/en
  set_ac97_reg(0x26,ac97_headset_mic); 
}

void set_external_io(unsigned int volume)
{
  unsigned int vol_left, vol_right;
  if(global_serial_num <  SNOM200_B) return; // doesn't work for snom200a
  ac97_headset &= ~0x8000;
  ac97_adc &= ~0x800f;
  ac97_headset_mic &= ~0x8000;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    ac97_gp &= ~HEADSET_MIC; //headset_mic dis
    ac97_headset |= 0x8000;//speaker  mute bit set
    ac97_dac = 0x8000;
  }
  else {
    // if (volume < 0 || volume > 4) return;
    //ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= 0x6;
    ac97_gp |= HEADSET_MIC;
    ac97_headset_mic |= HEADSET_PC;
    volume = reverse_volume(volume);
    vol_left  = volume;
    vol_right = volume;
    volume = (vol_left << 8) | vol_right;
    ac97_headset = volume;    
    ac97_dac = HANDSET_DAC_GAIN;
  }
  set_ac97_reg(0x20, ac97_gp);
  set_ac97_reg(0x26,ac97_headset_mic); //external audio from pc audio connector
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) dis/en
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x04,ac97_headset); // mute on/off
}

void set_speaker_phone(unsigned int volume)
{
  half_duplex = volume > 0;
  ac97_adc &= ~0x800f;
  ac97_speaker &= ~0x8000;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac = 0x8000;// mute on
    ac97_mic &= ~0x8000; //sidetone enabled
    set_ac97_reg(0x06, ac97_speaker);// mute on
  }
  else {
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);     
    set_ac97_reg(0x06, ac97_speaker); //speaker opened
    ac97_adc |= ac97_adc_sphone;
    ac97_dac = ac97_dac_sphone;
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic/hfmic  selected
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is open
    //set_port(PORT_SNOM_CSR, MIC_CONTROL,0);//handsfree mic on
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
  set_ac97_reg(0xe, ac97_mic);//side tone level sent
  set_ac97_reg(0x18,ac97_dac);// new value for dac gain
}

void set_headset(unsigned int volume)
{
  if(global_serial_num <  SNOM200_B) set_handset(volume);
   ac97_headset &= ~0x8000; 
  if(!volume) {
    ac97_headset |= 0x8000;//speaker  mute bit set
    ac97_dac = 0x8000;
  }
  else {
    unsigned int vol_left, vol_right;
    volume = reverse_volume(volume);
    vol_left  = volume;
    vol_right = volume;
    volume = (vol_left << 8) | vol_right;
    ac97_headset = volume;    
    ac97_dac = HANDSET_DAC_GAIN;       
  }    
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x04,ac97_headset); // mute on/off
}
#endif //ifdef snom105


void set_loopback(unsigned int volume) // volume means device in that case
{
  set_handset_mic (4);

  switch(volume){
    case 1: set_handset (15); break; // handset
    case 2: set_ringer (10); break; // casing speaker
    case 3: set_headset (15); break; // headset
    default: {
      set_handset_mic (0);
      set_handset (0);
      set_ringer (0);
      set_headset (0);
    }
  }

  if(!volume) 
    ac97_gp &=~LPBK;       
  else  
    ac97_gp |= LPBK;    
  set_ac97_reg(0x20, ac97_gp);
}

void snom_aud_set_amp_gain(int level)
{}

void snom_aud_set_tx_gain(int level)
{}

void snom_aud_set_tx_gain2(int level)
{}

void snom_aud_set_rx_gain(int level)
{}

void snom_aud_shutdown_speaker(int bTrue)
{}

void snom_aud_shutdown(int bTrue)
{}

#endif

// start the ice process:
int snom_aud_set_ice_parameters(struct snom_aud_ice_parameter_s *parm) {
  return -EINVAL;
}

void snom_aud_set_mixer_mode(int mode) {
  unsigned long flags;
	
  save_flags(flags); cli();
  mixer_mode = mode;

  // reset the audio buffer (exception: "knocking"):
  au_rd_pos = au_wr_pos;

  restore_flags(flags);
}

/* table for DTMF tones */
/* gawk -v s=23 -v r=2 -v vol=8000 'BEGIN{for(i=0;i<s;i++) printf("%d, ", sin(3.1415927*2*r*i/s)*vol ); exit;}' */
static const short freq_697[]  = { /* 23 samples, 2 reps */
  0, 4156, 7103, 7981, 6535, 3187, -1089, -5048, -7538, -7832, -5846, -2158, 2158, 5846, 7832, 7538, 5048, 1089, -3187, -6535, -7981, -7103, -4156
};

static const short freq_770[]  = { /* 52 samples, 5 reps */
  0, 4544, 7480, 7767, 5304, 964, -3717, -7083, -7941, -5988, -1914, 2836, 6583, 7999, 6583, 2836, -1914, -5988, -7941, -7083, -3717, 964, 5304, 7767, 7480, 4544, -0, -4544, -7480, -7767, -5304, -964, 3717, 7083, 7941, 5988, 1914, -2836, -6583, -7999, -6583, -2836, 1914, 5988, 7941, 7083, 3717, -964, -5304, -7767, -7480, -4544
};

static const short freq_852[]  = { /* 47 samples, 5 reps */
  0, 4957, 7782, 7256, 3608, -1593, -6109, -7995, -6440, -2113, 3123, 7015, 7888, 5366, 534, -4527, -7640, -7465, -4076, 1066, 5750, 7959, 6743, 2624, -2624, -6743, -7959, -5750, -1066, 4076, 7465, 7640, 4527, -534, -5366, -7888, -7015, -3123, 2113, 6440, 7995, 6109, 1593, -3608, -7256, -7782, -4957
};

static const short freq_941[]  = { /* 17 samples, 2 reps */
  0, 5389, 7965, 6384, 1469, -4211, -7694, -7161, -2889, 2889, 7161, 7694, 4211, -1469, -6384, -7965, -5389
};

static const short freq_1209[] = { /* 53 samples, 8 reps */
  0, 6499, 7578, 2336, -4854, -7996, -4469, 2785, 7717, 6212, -473, -6764, -7413, -1879, 5222, 7968, 4068, -3224, -7828, -5902, 946, 7005, 7222, 1415, -5572, -7912, -3653, 3653, 7912, 5572, -1415, -7222, -7005, -946, 5902, 7828, 3224, -4068, -7968, -5222, 1879, 7413, 6764, 473, -6212, -7717, -2785, 4469, 7996, 4854, -2336, -7578, -6499
};

static const short freq_1336[] = { /* 6 samples, 1 rep */
  0, 6928, 6928, -0, -6928, -6928 
};

static const short freq_1477[] = { /* 65 samples, 12 rep */
  0, 7334, 5858, -2655, -7978, -3717, 5009, 7719, 1155, -6795, -6583, 1537, 7811, 4702, -4055, -7941, -2287, 6114, 7171, -386, -7480, -5588, 3016, 7997, 3371, -5304, -7608, -772, 6991, 6356, -1914, -7885, -4384, 4384, 7885, 1914, -6356, -6991, 772, 7608, 5304, -3371, -7997, -3016, 5588, 7480, 386, -7171, -6114, 2287, 7941, 4055, -4702, -7811, -1537, 6583, 6795, -1155, -7719, -5009, 3717, 7978, 2655, -5858, -7334  
};

static const short *dtmf_freq[16][2] = {
  { freq_941, freq_1336 }, // 0
  { freq_697, freq_1209 }, // 1
  { freq_697, freq_1336 }, // 2
  { freq_697, freq_1477 }, // 3
  { freq_770, freq_1209 }, // 4
  { freq_770, freq_1336 }, // 5
  { freq_770, freq_1477 }, // 6
  { freq_852, freq_1209 }, // 7
  { freq_852, freq_1336 }, // 8
  { freq_852, freq_1477 }, // 9
  { freq_941, freq_1209 }, // #
  { freq_941, freq_1477 }, // *
  { freq_941, freq_941 }, // fixme
  { freq_941, freq_941 }, // fixme
  { freq_941, freq_941 }, // fixme
  { freq_941, freq_941 }, // fixme
};

static const int dtmf_len[16][2] = {
  { sizeof(freq_941) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 0
  { sizeof(freq_697) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // 1
  { sizeof(freq_697) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 2
  { sizeof(freq_697) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // 3
  { sizeof(freq_770) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // 4
  { sizeof(freq_770) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 5
  { sizeof(freq_770) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // 6
  { sizeof(freq_852) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // 7
  { sizeof(freq_852) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 8
  { sizeof(freq_852) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // 9
  { sizeof(freq_941) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // #
  { sizeof(freq_941) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // *
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
};

struct dtmf_state_t internal_dtmf; // outgoing
struct dtmf_state_t external_dtmf[MAX_NUM_RTP]; // incoming from RTP
int play_dtmf = 0; // bool for enabling internal dtmf

#define DTMF_STATE_NO_DTMF 0
#define DTMF_STATE_SEND_DTMF 1
#define DTMF_STATE_SEND_LAST 2

static int dtmf_key[MAX_NUM_RTP] = { 0 };
static int dtmf_state[MAX_NUM_RTP] = { DTMF_STATE_NO_DTMF };
static int dtmf_shutoff[MAX_NUM_RTP] = { 0 }; // when we are playing dtmf of the other side
// static int dtmf_in_duration = 0; // duration of current incoming dtmf sound
static int dtmf_recv_type[MAX_NUM_RTP] = { 101 }; // hardwired?
static unsigned long dtmf_recv_ssrc_id[MAX_NUM_RTP];
static int dtmf_recv_ssrc_valid[MAX_NUM_RTP] = { 0 };
static int dtmf_duration[MAX_NUM_RTP] = { 0 }; // duration of the outgoing dtmf so far
static int dtmf_recv_pos[MAX_NUM_RTP] = { 0 }; // last received position

// set dtmf key: bits:
// 0..3: key
// 4: key valid
// 5: play sound (on local speaker)
// 6: send dtmf rtp (rfc 2833)
// 7: send tone inband (on rtp packets)
// 8..15: payload type
// 15..17: channel
void snom_aud_set_dtmf(int key) {
  unsigned long flags;

  int channel = key >> 16;
  if (channel < 0 || channel >= MAX_NUM_RTP) return;
  key &= 0xffff; 

  // make sure we dont send both in- and out-of-band:
  if (key & 0x40) {
    key &= ~0x80;
  }

  save_flags(flags); cli();

  switch (dtmf_state[channel]) {
  case DTMF_STATE_NO_DTMF:
    if (key & 16) {
      dtmf_state[channel] = DTMF_STATE_SEND_DTMF;
      dtmf_duration[channel] = 0;
      need_marker[channel] = 1;
    }
    dtmf_key[channel] = key;
    break;
  case DTMF_STATE_SEND_DTMF:
    if (!(key & 16)) {
      dtmf_state[channel] = DTMF_STATE_SEND_LAST;
    }
    else
      // set the new key: fixme: if we did not send the last key
      // termination code, its probably too early to send the next
      // one. However, what can we do?
      dtmf_key[channel] = key;
    break;
  case DTMF_STATE_SEND_LAST:
    // fixme: if we did not send the last key termination code, its
    // probably too early to send the next one. However, what can we
    // do?
    if (key & 16) {
      dtmf_duration[channel] = 0;
      dtmf_state[channel] = DTMF_STATE_SEND_DTMF;
    }
    dtmf_key[channel] = key;
    break;
  }

  if (key & 32) {
    set_dtmf(dtmf_key[channel] & 15, &internal_dtmf);
    play_dtmf = 1;
  }
  else {
    play_dtmf = 0;
  }

  restore_flags(flags);
}

static void clear_buffers(int restart) {
  int j;
  volatile cbd_t  *bdp;
  // printk("clear_buffers\n");
  bdp = phy.tx_cur = phy.tx_bd_base;

  for (j = 0; j < HW_NUM_BLOCKS - 1; j++) {
    bdp->cbd_sc = BD_SC_READY; // ints only on rx side
    bdp->cbd_datlen = CPM_BUF_SIZE;
    write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 
    bdp++;
  }
  bdp->cbd_sc = BD_SC_READY | BD_SC_WRAP; // ints only on rx side
  bdp->cbd_datlen = CPM_BUF_SIZE;
  write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 
	
  bdp = phy.rx_cur = phy.rx_bd_base;
	
  for (j=0; j<(HW_NUM_BLOCKS-1); j++) {
    bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_INTRPT);
    bdp->cbd_datlen = CPM_BUF_SIZE;
    write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 
    bdp++;
  }
  bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT);
  bdp->cbd_datlen = CPM_BUF_SIZE;
  write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 

#if 0
  /* mark all frames as clean */
  for(j = 0; j < SW_NUM_BLOCKS; j++) {
    // tx_buffer.avail[j] = 0;
    // rx_buffer.avail[j] = 0;
  }
#endif

  if (restart) {
    restart_cpm(CPM_RXTX);
    
    /* init automatic gain control */
    init_agc();
  }
}

void snom_aud_set_tos(int t) {
  tos = t;
}

// set the parameters of a specific audio rtp channel:
int snom_aud_set_rtp_parameters2(int c, struct snom_aud_rtp_parameter_s *parm) {
  int rc;
  struct sockaddr_in sock;
  struct rtable *rt = 0;
  unsigned long flags;
  save_flags(flags); cli();

  /* set the type & length: */
#if 0
  printk("rtp:");
  printk("%d ", c);
  printk("%08x", parm->raddr);
  printk(":%d", parm->remote_port);
  printk(" %d\n", parm->local_port);
#endif
  check_c("snom_aud_set_rtp_parameters2", c);
  packet_type[c] = parm->type;
  check_for_agc();
     {
	/* ulaw or alaw */
	if      (parm->length == 80)  packet_size[c] = 80;
	else if (parm->length == 240) packet_size[c] = 240;
	else if (parm->length == 320) packet_size[c] = 320;
	else                          packet_size[c] = 160;
      }

  // printk("rtp: a=%08x p=%d t=%d l=%d\n", (int)parm->raddr, (int)parm->remote_port, (int)packet_type, (int)packet_size);

  /* set the local ssrc */
  local_ssrc[c] = parm->ssrc;

  /* A change in the local socket? */
  if (local_port[c] != parm->local_port) {
    /* close the local port */
    if (rtp_socket[c] != 0) {
      sock_release(rtp_socket[c]);
      rtp_socket[c] = 0;
    }
    if (rtcp_socket[c] != 0) {
      sock_release(rtcp_socket[c]);
      rtcp_socket[c] = 0;
    }
    local_port[c] = 0;
    play_dtmf = 0;
  
    /* open a new port? */
    if (parm->local_port > 0) {
      jit_recv_reset(c);
      jit_decode_start[c] = 1;
  
      /* create a socket: */
      // printk("rtp: open socket %d\n", parm->local_port);
      rc = sock_create(PF_INET, SOCK_DGRAM, 0, &rtp_socket[c]);
      if (rc < 0) {
	rtp_socket[c] = 0; /* this socket cannot be used! */
	// printk("snom_aud_set_rtp_parameters: Could not create socket\n");
	goto ret_release;
      }
      
      rc = sock_create(PF_INET, SOCK_DGRAM, 0, &rtcp_socket[c]);
      if (rc < 0) {
	rtcp_socket[c] = 0; /* this socket cannot be used! */
	goto ret_release;
      }
      
      /* set up the local port: */
      memset(&sock, 0, sizeof(sock));
      sock.sin_family = AF_INET;
      sock.sin_addr.s_addr = htonl(INADDR_ANY);
      sock.sin_port = htons(parm->local_port);
      rc = rtp_socket[c]->ops->bind(rtp_socket[c], (struct sockaddr *)&sock, sizeof(sock));
      if (rc != 0) {
	// printk("snom_aud_set_rtp_parameters: Could not bind\n");
	goto ret_release;
      }
      
      memset(&sock, 0, sizeof(sock));
      sock.sin_family = AF_INET;
      sock.sin_addr.s_addr = htonl(INADDR_ANY);
      sock.sin_port = htons(parm->local_port + 1);
      rc = rtcp_socket[c]->ops->bind(rtcp_socket[c], (struct sockaddr *)&sock, sizeof(sock));
      if (rc != 0) {
	// printk("snom_aud_set_rtp_parameters: Could not bind\n");
	goto ret_release;
      }
      
      /* set the socket type of service: IP_TOS = 5 */
      rtp_socket[c]->sk->protinfo.af_inet.tos = tos;
      rtp_socket[c]->sk->priority = TC_PRIO_INTERACTIVE;
      rtp_socket[c]->sk->protinfo.af_inet.pmtudisc = IP_PMTUDISC_DONT; // always increase the identity
      
      // clear_buffers(1);
      local_port[c] = parm->local_port;
    } /* local socket required */
  } /* change in local socket */

  /* prepare the destination */
  if ((remote_adr[c].sin_addr.s_addr != parm->raddr || remote_adr[c].sin_port != parm->remote_port) &&
      rtp_socket[c] != 0 && rtcp_socket[c] != 0) {
    need_marker[c] = 2;
    dtmf_key[c] = 0;
    dtmf_state[c] = DTMF_STATE_NO_DTMF;
    remote_adr[c].sin_addr.s_addr = parm->raddr;
    remote_adr[c].sin_port = parm->remote_port;
    remote_adr[c].sin_family = AF_INET;
    
    if (remote_adr[c].sin_addr.s_addr != 0) {
      if (ip_route_output(&rt, remote_adr[c].sin_addr.s_addr, rtp_socket[c]->sk->saddr,
 			  RT_TOS(rtp_socket[c]->sk->protinfo.af_inet.tos) | RTO_CONN | rtp_socket[c]->sk->localroute,
 			  rtp_socket[c]->sk->bound_dev_if)) {
 	// printk("rtp: ip_route_output failed!\n");
 	goto ret_release;
      }
      __sk_dst_set(rtp_socket[c]->sk, &rt->u.dst);
      
      if (ip_route_output(&rt, remote_adr[c].sin_addr.s_addr, rtcp_socket[c]->sk->saddr,
 			  RT_TOS(rtcp_socket[c]->sk->protinfo.af_inet.tos) | RTO_CONN | rtcp_socket[c]->sk->localroute,
 			  rtcp_socket[c]->sk->bound_dev_if)) {
 	// printk("rtcp: ip_route_output failed!\n");	
 	goto ret_release;
      }
      __sk_dst_set(rtcp_socket[c]->sk, &rt->u.dst);
    } // remote_adr available
  }

  restore_flags(flags);
  return 0;

 ret_release:
  if (rtp_socket[c] != 0) {
    sock_release(rtp_socket[c]);
  }
  if (rtcp_socket[c] != 0) {
    sock_release(rtcp_socket[c]);
  }
  rtp_socket[c] = 0;
  rtcp_socket[c] = 0;
  clear_buffers(0);
  restore_flags(flags);
  return rc;
}

// this function is kept for backward compatibility. It sets the
// parameters for the RTP channel 0 and disables all other channels.
int snom_aud_set_rtp_parameters(struct snom_aud_rtp_parameter_s *parm) {
  return snom_aud_set_rtp_parameters2(0, parm);
}

/******************************************************************************/

/*
 * Call back mechanism
 */

void snom_aud_set_call_back(void (*cb)(void))
{
  unsigned long flags;
	
  save_flags(flags); cli();
	
  snom_aud_call_back = cb;
	
  restore_flags(flags);
}

#ifndef CONFIG_SNOM100
static _INLINE_ void shutdown_cpm(void)
{	
  volatile cpm8xx_t *cp = cpmp;
  volatile scc_t    *sp = &cp->cp_scc[0];
  int i;

  /* Disable reception and transmission. */
	
  sp->scc_gsmrl &= ~(SCC_GSMRL_ENT | SCC_GSMRL_ENR); 
	
  /* Enable all interrupts and clear all pending events. */
	
  sp->scc_sccm |= SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX; // SCCM_GLR | SCCM_GLT
  sp->scc_scce = 0xffff;
	
  // printk("<sd>");
  clear_buffers(0);

  local_time = 0;
  for(i = 0; i < MAX_NUM_RTP; i++) jit_recv_reset(i);

  phy.is_underrun = 1;
}
#else
static _INLINE_ void shutdown_cpm(void)
{	
  volatile cpm8xx_t *cp = cpmp;
  volatile smc_t    *sp = &cp->cp_smc[1];
  int i;

  /* Disable reception and transmission. */
	
  sp->smc_smcmr &= 0xfffc; 
	
  /* Enable all interrupts and clear all pending events. */
	
  sp->smc_smcm |= SMCM_TXE | SMCM_BSY | SMCM_TX | SMCM_RX;
  sp->smc_smce = 0xff;
	
  // printk("<sd>");
  clear_buffers(0);

  local_time = 0;
  for(i = 0; i < MAX_NUM_RTP; i++) jit_recv_reset(i);

  phy.is_underrun = 1;
}
#endif

void snom_aud_reset(snom_aud_info_t *info) {
  unsigned long flags;
  int i;
  save_flags(flags); cli();
  
  for(i = 0; i < MAX_NUM_RTP; i++) {
    if (rtp_socket[i] != 0)
      sock_release(rtp_socket[i]);
    if (rtcp_socket[i] != 0)
      sock_release(rtcp_socket[i]);
    rtp_socket[i] = 0;
    rtcp_socket[i] = 0;
    local_port[i] = 0;
  }
  clear_buffers(0);

  restore_flags(flags);
}

/*
 * real time rtp part
 */

static void snom_aud_sendmsg(struct sock *sk, char *data, int len, int dport) {
  struct sk_buff *skb;
  struct udphdr *uh;
  unsigned char *msg;
  int rc;

  /* fixme: get hand on device and read of physical header size */
  skb = alloc_skb(len + sizeof(struct udphdr) + sizeof(struct iphdr)
		  + 14 /* ethernet header size */ + 15, GFP_ATOMIC);
  if (skb == NULL) {
    // printk("rtp: no skb\n");
    return;
  }

  skb_reserve(skb, (14 /* ethernet header size */ + sizeof(struct udphdr) + sizeof(struct iphdr) + 15) & ~15);
  skb->sk  = sk;
  
  /* put the data on the packet */
  msg = (unsigned char *) skb_put(skb, len);
  memcpy(msg, data, len);

  /* add udp header */
  uh = (struct udphdr *) skb_push(skb, sizeof(struct udphdr));
  uh->source = sk->sport;
  uh->dest = dport;
  uh->len = len + 8; // length with! udp header
  uh->check = 0; // no checksum
  skb->h.uh = uh;

  // put it on the cable:
  rc = ip_queue_xmit(skb);
  if (rc != 0) {
    // printk("Failed to send msg: %d\n", rc);
  }
  return;
}

/* check if something is available for sending: */
static _INLINE_ void snom_aud_check_outgoing(void) {
  unsigned long flags;
  int c;

  /* make sure we have a socket: */
  save_flags(flags); cli();

  // check if we can encode something:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (jit_rtp_active(c)) {
      const int type = packet_type[c];
      int *length = &jit_send_len[c];
      int *el = &jit_encoded_len[c];
      int *ul = &jit_unencoded_len[c];
      unsigned char *buffer = jit_encoded_buffer[c];

      if (type == PT_GSM) {
	if (*length >= 160) {
	  short *src = jit_send_buffer[c];
	  int rem = *length % 160;
	  int l = *length - rem;
	  unsigned char *dst = &buffer[*el + 12];
	  int maxlen = sizeof(jit_encoded_buffer[0]) - (*el + 12);
	  int len;
	  restore_flags(flags);
	  len = snom_aud_encode(c, type, src, l, dst, maxlen);
	  // len = 0;
	  save_flags(flags); cli();
	  *el += len;
	  if (*ul == 0) jit_encoded_ts[c] = local_time;
	  *ul += *length - rem;
	  if (rem > 0)
	    memmove(&jit_encoded_buffer[c][*el + 12], 
		    &jit_encoded_buffer[c][*el + 12 + *length], 
		    rem * sizeof(short));
	  *length = rem;
	}
      } // GSM
      else {
	if (*length >= 80) {
	  // all other codecs work on 80 samples:
	  short *src = jit_send_buffer[c];
	  int l = *length;
	  unsigned char *dst = &buffer[*el + 12];
	  int maxlen = sizeof(jit_encoded_buffer[0]) - (*el + 12);
	  int len;	
	  restore_flags(flags);
	  len = snom_aud_encode(c, type, src, l, dst, maxlen);
	  save_flags(flags); cli();
	  *el += len;
	  if (*ul == 0) jit_encoded_ts[c] = local_time;
	  *ul += *length;
	  *length = 0;
	}
      } // no GSM
    }
  } // for c
  restore_flags(flags);

  // now check if we can send something:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (jit_rtp_active(c) && jit_unencoded_len[c] >= packet_size[c] && jit_encoded_len[c] > 0) {
      if (dtmf_state[c] == DTMF_STATE_NO_DTMF || (dtmf_key[c] & 0x80) != 0) {
	unsigned char *buffer = jit_encoded_buffer[c];
	unsigned long ssrc = local_ssrc[c];
	int ts = jit_encoded_ts[c];

	buffer[0] = 0x80;
	buffer[1] = packet_type[c];
	if (need_marker[c] > 0) {
	  need_marker[c]--;
	  if (need_marker[c] > 0) return; // skip first packet so that the jitter is minimized
	  buffer[1] |= 0x80;
	}
  
	/* sequence number: */
	buffer[2] = (unsigned char)(packet_seqnr[c] >> 8);
	buffer[3] = (unsigned char)(packet_seqnr[c] & 0xff);
    
	/* timestamp */
	buffer[4] = (unsigned char)((ts >> 24) & 0xff);
	buffer[5] = (unsigned char)((ts >> 16) & 0xff);
	buffer[6] = (unsigned char)((ts >> 8) & 0xff);
	buffer[7] = (unsigned char)(ts & 0xff);

	/* ssrc */
	buffer[8]  = (unsigned char)((ssrc >> 24) & 0xff);
	buffer[9]  = (unsigned char)((ssrc >> 16) & 0xff);
	buffer[10] = (unsigned char)((ssrc >> 8) & 0xff);
	buffer[11] = (unsigned char)(ssrc & 0xff);

	// printk("t=%d\n", jit_encoded_len[c]);
	// if (c == 0) printk("%d", packet_type[c]);
	snom_aud_sendmsg(rtp_socket[c]->sk, buffer, jit_encoded_len[c] + 12, remote_adr[c].sin_port);
	packet_seqnr[c]++;
	// if (c > 0) printk("q");
	jit_unencoded_len[c] = 0;
	jit_encoded_len[c] = 0;
      } // if no dtmf or dtmf inband

      /* RTP out of band signalling: */
      if (dtmf_state[c] != DTMF_STATE_NO_DTMF && (dtmf_key[c] & 0x40) != 0) {
	unsigned char *buffer = jit_encoded_buffer[c];
	unsigned long ssrc = local_ssrc[c];
	int ts = jit_encoded_ts[c];

	buffer[0] = 0x80; // RTP
	buffer[1] = (unsigned char)((dtmf_key[c] >> 8) & 0x7f); // media-type
	if (need_marker[c]) {
	  buffer[1] |= 0x80;
	  need_marker[c] = 0;
	}

	/* sequence number: */
	buffer[2] = (unsigned char)(packet_seqnr[c] >> 8);
	buffer[3] = (unsigned char)(packet_seqnr[c] & 0xff);
  
	/* timestamp */
	buffer[4] = (unsigned char)((ts >> 24) & 0xff);
	buffer[5] = (unsigned char)((ts >> 16) & 0xff);
	buffer[6] = (unsigned char)((ts >> 8) & 0xff);
	buffer[7] = (unsigned char)(ts & 0xff);
    
	/* ssrc */
	buffer[8]  = (unsigned char)((ssrc >> 24) & 0xff);
	buffer[9]  = (unsigned char)((ssrc >> 16) & 0xff);
	buffer[10] = (unsigned char)((ssrc >> 8) & 0xff);
	buffer[11] = (unsigned char)(ssrc & 0xff);

	buffer[12] = (unsigned char)(dtmf_key[c] & 15);
	if (dtmf_state[c] == DTMF_STATE_SEND_LAST) {
	  buffer[13] = 0x8a; // volume, 0x80 = end of digit
	  dtmf_state[c] = DTMF_STATE_NO_DTMF;
	  dtmf_key[c] = 0; // reset the key
	}
	else 
	  buffer[13] = 0x0a; // volume
	dtmf_duration[c] += packet_size[c];
	buffer[14] = (unsigned char)((dtmf_duration[c] >> 8) & 0xff); // duration
	buffer[15] = (unsigned char)(dtmf_duration[c] & 0xff);
	snom_aud_sendmsg(rtp_socket[c]->sk, buffer, 16, remote_adr[c].sin_port);
	packet_seqnr[c]++;
      } // out-of-band
    } // packet available
  } // for c
}

/* check if the data contains a rtp packet and process it: */
#if 0
static _INLINE_ void snom_aud_process_rtp(unsigned char *data, int length, unsigned int remote_time) {
  static short decode_buffer[640];
  int plen; /* payload length, audio length */
  int start_pos, i;
  unsigned long flags;

  if (data != 0) {
    /* determine the length of the payload: */
    plen = (data[0] & 0xf) * 4 + 12;
    if (data[0] & 0x10) plen += ((int)(data[plen + 2]) << 8) | (int)(data[plen + 3]);
    plen = length - plen;
    if (data[0] & 0x20) plen -= data[length - 1]; /* padding */
    if (plen <= 0 || plen > sizeof(decode_buffer) / sizeof(decode_buffer[0])) return;
    
    /* determine the type and decode the payload: */
    remote_packet_type = data[1] & 0x7f;
    remote_packet_size = snom_aud_decode(remote_packet_type, &data[length - plen], plen, decode_buffer, sizeof(decode_buffer) / sizeof(decode_buffer[0]));
    if (remote_packet_size == 0) return; /* unknown decode type */
  }
  else {
    /* fill the gap: */
    snom_aud_fill(remote_packet_type, remote_packet_size, decode_buffer, sizeof(decode_buffer) / sizeof(decode_buffer[0]));
  }
  save_flags(flags);
  cli();

  /* copy the data: */
  start_pos = remote_time % (sizeof(rx_buffer.buf) / sizeof(rx_buffer.buf[0]));
  if (start_pos + remote_packet_size > (sizeof(rx_buffer.buf) / sizeof(rx_buffer.buf[0]))) {
    /* split up into two memcpy: */
    const int length = (sizeof(rx_buffer.buf) / sizeof(rx_buffer.buf[0])) - start_pos;
    memcpy(&rx_buffer.buf[start_pos], decode_buffer, length * sizeof(rx_buffer.buf[0]));
    for(i = 0; i < length; i += SNOM_AUD_BLOCK_SIZE)
      rx_buffer.avail[(start_pos + i) / SNOM_AUD_BLOCK_SIZE] = 1;
    memcpy(&rx_buffer.buf[0], &decode_buffer[length], (remote_packet_size - length) * sizeof(rx_buffer.buf[0]));
    for(i = 0; i < remote_packet_size - length; i += SNOM_AUD_BLOCK_SIZE)
      rx_buffer.avail[i / SNOM_AUD_BLOCK_SIZE] = 1;
  }
  else {
    memcpy(&rx_buffer.buf[start_pos], decode_buffer, remote_packet_size * sizeof(rx_buffer.buf[0]));
    for(i = 0; i < remote_packet_size; i += SNOM_AUD_BLOCK_SIZE)
      rx_buffer.avail[(start_pos + i) / SNOM_AUD_BLOCK_SIZE] = 1;
  }
  restore_flags(flags);
}
#endif

/* check if we have incoming packets: */
static _INLINE_ void snom_aud_check_incoming(void) {
  unsigned long flags;
  int c;
  for(c = 0; c < MAX_NUM_RTP; c++) {
    /* rtp handling: */
    if (rtp_socket[c]) {
      /* get all packets that are available: */
      while(1) {
	struct sk_buff *skb;
	int len, err;
      
	/* check if there is a packet ready: (code from udp.c:udp_recvmsg()) */
	skb = skb_recv_datagram(rtp_socket[c]->sk, 0, MSG_DONTWAIT, &err);
	if (!skb) break;
      
	/* process the packet: */
	len = skb->len - sizeof(struct udphdr);

	/* check if we need to fill up buffers */
	save_flags(flags); cli();
	// printk("r=%d\n", len);
	jit_recv_rtp(c, skb->h.raw + sizeof(struct udphdr), len);
	restore_flags(flags);
	skb_free_datagram(rtp_socket[c]->sk, skb);
      }
    } /* rtp */

    /* rtcp handling: */
    if (rtcp_socket[c]) {
      while(1) {
	struct sk_buff *skb;
	int err;

	/* just eat all packets from the buffer: */
	skb = skb_recv_datagram(rtcp_socket[c]->sk, 0, MSG_DONTWAIT, &err);
	if (!skb) break;
	skb_free_datagram(rtcp_socket[c]->sk, skb);
      } /* while */
    } // rtcp
  }
} /* snom_check_incoming */

#ifndef CONFIG_SNOM100
#define MAX_NUM_PENDING_AC97_CMDS 128
static int ac97_fifo_rx = 0;
static int ac97_fifo_tx = 0;
static unsigned long ac97_regs[128]; 
// static int delay_sending_commands = 100000;
#endif

// Update the filtered energy: Quick for high energy, slow for low energy
static long update_energy(long old_e, long new_e) {
  if (new_e > 100000000) 
    return old_e - (old_e / 2) + new_e / 2;
  else if (new_e > 10000000) 
    return old_e - (old_e / 4) + new_e / 4;
  else if (new_e > 1000000) 
    return old_e - (old_e / 8) + new_e / 8;
  else if (new_e > 100000) 
    return old_e - (old_e / 16) + new_e / 16;
  else 
    return old_e - (old_e / 32) + new_e / 32;
}

#ifndef CONFIG_SNOM100
static struct pending_command {
  unsigned long reg;
  unsigned long value;
} ac97_fifo_buffer[MAX_NUM_PENDING_AC97_CMDS];
#endif

// write a block of short samples (16 bit) to the hardware: This
// includes the calculation of the current energy level presented to
// the speaker.
#ifdef CONFIG_SNOM100
static void write_to_hardware(unsigned char *raw, short *linear) {
  int i;
  if (linear) {
    // printk("%d", (((int)linear) / 80) % 10);
    for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
      *raw++ = snom_aud_linear2ulaw[*linear++ / 4 + 8192];
    }
  }
  else {
    for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
      *raw++ = 0xff;
    }
  }
}

static void read_from_hardware(short *linear, unsigned char *raw) {
  int i;
  energy_lc = 0;
  /* it makes sense here to do a little loop optimization: */
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    int val = snom_aud_ulaw2linear[*raw++];
    *linear++ = val;
    energy_lc += (val * val) / 256;
  }
  // update the filtered energy:
  energy_lf = update_energy(energy_lf, energy_lc);
}
/* convert from ulaw to linear: */

#else //config_snom100
static void write_to_hardware(unsigned char *raw, short *linear) {
  static short last_sf = 0;
  short sf[6];
  short diff1, diff2;
  int i;
  unsigned long *p = (unsigned long*)raw;

  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    int sub;
    register unsigned long int0, int1, int2;
    // short sm = ((i % 20) - 10) * 1000;
    short sm = (linear == NULL ? 0 : ((*linear++) & 0xffff));
    sf[5] = sm;

    diff1 = (sf[5] - last_sf) >> 1;
    diff2 = ((int)(sf[5] - last_sf) * 171) / 1024; // divide by 6
    sf[2] = last_sf + diff1;
    sf[0] = last_sf + diff2;
    sf[1] = sf[0] + diff2;
    sf[3] = sf[2] + diff2;
    sf[4] = sf[3] + diff2;
    last_sf = sf[5]; // remember for next time
    
    for(sub = 0; sub < 6; sub++) {
      unsigned long sample = (unsigned short)(sf[sub]);
      int0 = 0x98000000; // pure audio
      int1 = sample >> 8;
      int2 = (sample << 24) | (sample << 4);

      // Need to fill in a command?
      if (ac97_fifo_rx != ac97_fifo_tx) {
	// writing a sample and a command
	const register unsigned long reg = ac97_fifo_buffer[ac97_fifo_rx].reg & 0xff;
	const register unsigned long val = ac97_fifo_buffer[ac97_fifo_rx].value & 0xffff;
	
	ac97_regs[reg] = val;
	
	int0 |= 0x60000000 | (reg << 8);
	int1 |= val << 12;	
	
	// increment the rx pointer
	ac97_fifo_rx = ((ac97_fifo_rx + 1) % MAX_NUM_PENDING_AC97_CMDS);
      }
      
      // write the three registers to memory:
      p[0] = int0;
      p[1] = int1;
      p[2] = int2;    
      p += 3; // go to the next frame
    } // sub 
  }
}

// read from the hardware and update the current energy level.
static void read_from_hardware(short *linear, unsigned char *raw) {
  int i, sub;
  unsigned long *p = (unsigned long*)raw;
  energy_lc = 0;

  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    int val = 0;

    val += (short)((p[2     ] >> 4) & 0xffff);
    val += (short)((p[2 +  3] >> 4) & 0xffff);
    val += (short)((p[2 +  6] >> 4) & 0xffff);
    val += (short)((p[2 +  9] >> 4) & 0xffff);
    val += (short)((p[2 + 12] >> 4) & 0xffff);
    val += (short)((p[2 + 15] >> 4) & 0xffff);
    p += 18;
    val = (val * 2730) / 16384;
    energy_lc += (val * val) / 256;
    *linear++ = val;
  } /* i */

  // update the filtered energy:
  energy_lf = update_energy(energy_lf, energy_lc);
}
#endif //config_snom100

#ifndef CONFIG_SNOM100
typedef struct _ac97_slot {
  unsigned short slot_1; //slot0 is defined in int0 
  unsigned short slot_2;
} t_ac97_slot;

static void set_ac97_reg(int reg, unsigned long val) {
  ac97_fifo_buffer[ac97_fifo_tx].reg = reg;
  ac97_fifo_buffer[ac97_fifo_tx].value = val;
  ac97_fifo_tx = ((ac97_fifo_tx + 1) % MAX_NUM_PENDING_AC97_CMDS);

  // read back: (debug)
#if 0
  ac97_fifo_buffer[ac97_fifo_tx].reg = reg | 0x80;
  ac97_fifo_buffer[ac97_fifo_tx].value = 0;
  ac97_fifo_tx = ((ac97_fifo_tx + 1) % MAX_NUM_PENDING_AC97_CMDS);
#endif
}
#endif //config_snom100

// get 80 samples of dtmf:
static void get_dtmf(struct dtmf_state_t *dtmf, short *buffer) {
  int i;
  const int k = dtmf->key;
  const int l0 = dtmf_len[k][0];
  const int l1 = dtmf_len[k][1];
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    const short *p0 = dtmf_freq[k][0];
    const short *p1 = dtmf_freq[k][1];
    *buffer++ = p0[dtmf->pos0++] + p1[dtmf->pos1++];
    if (dtmf->pos0 >= l0) dtmf->pos0 = 0;
    if (dtmf->pos1 >= l1) dtmf->pos1 = 0;
  }
}

// set a dtmf key:
static void set_dtmf(int key, struct dtmf_state_t *dtmf) {
  key &= 15;
  if (dtmf->key != key) {
    dtmf->key = key;
    dtmf->pos0 = 0;
    dtmf->pos1 = 0;
  }
}

// get audio from an audio channel:
short *jit_get_audio(const int c) {
  int timestamp;
  int i;
  int packet;
  int type;
  unsigned char *buffer;

  check_c("jit_get_audio", c);
  // do we have the data from a previous decode step?
  if (jit_decode_len[c] >= SNOM_AUD_BLOCK_SIZE) {
    short *result = &jit_decode_buffer[c][jit_decode_pos[c]];
    // if (jit_decode_pos[c] < 0 || jit_decode_pos[c] + SNOM_AUD_BLOCK_SIZE > 320) printk("jit_get_audio");
    jit_decode_pos[c] += SNOM_AUD_BLOCK_SIZE;
    jit_decode_len[c] -= SNOM_AUD_BLOCK_SIZE;
    return result;
  }

  // get some more data: search for the lowest timestamp
  packet = -1;
  if (jit_decode_start[c]) {
    // in the beginning, search for the biggest! timestamp:
    for(i = 0; i < JITTER_SIZE; i++) {
      if (jit_recv_len[c][i] > 0) {
	const int ts = jit_recv_timestamp[c][i];
	if (ts - timestamp > 0 || packet < 0) {
	  timestamp = ts;
	  packet = i;
	}
      }
    } // for i
    if (packet >= 0) {
      jit_decode_start[c] = 0; // no more starting
      for(i = 0; i < JITTER_SIZE; i++) {
	if (i != packet) jit_recv_len[c][i] = 0;
      } // for i      
    } // packet >= 0
  } // decode start
  else {
    for(i = 0; i < JITTER_SIZE; i++) {
      if (jit_recv_len[c][i] > 0) {
	const int ts = jit_recv_timestamp[c][i];
	if (ts - timestamp < 0 || packet < 0) {
	  timestamp = ts;
	  packet = i;
	}
      }
    } // for i
  }

  if (packet < 0) {
    dtmf_recv_ssrc_valid[c] = 0; // underrun: accept any ssrc from now on!
    check_c("snom_aud_fill", c);
    jit_decode_len[c] = snom_aud_fill(c, jit_decode_buffer[c],
				      sizeof(jit_decode_buffer[0]) / sizeof(jit_decode_buffer[0][0]));
  }
  else {
    // decode the first part of the found packet:
    type = jit_recv_type[c][packet] & 0x7f;
    buffer = jit_recv_buffer[c][packet];
    // if (c > 0) printk("%d", type);
    jit_decode_pos[c] = 0;
    if (type == dtmf_recv_type[c]) {
      set_dtmf(buffer[0], &external_dtmf[c]);
      get_dtmf(&external_dtmf[c], jit_decode_buffer[c]);
      jit_decode_len[c] = SNOM_AUD_BLOCK_SIZE;
      jit_recv_len[c][packet] = 0; // this packet is used!
      dtmf_shutoff[c] = SNOM_AUD_BLOCK_SIZE * 3;
    }
    else {
      // no DTMF:
      int *length = &jit_recv_len[c][packet];
      int src_size, dec_size;
#if 0
      if (c > 0) {
	int x;
	printk("recv: t=%d", type);
	for(x = 0; x < *length; x++) printk(" %02x", (int)(buffer[x]));
	printk("\n");
      }
#endif
      snom_aud_get_encoded_length(type, *length, &src_size, &dec_size);
      // if (src_size == 0) printk("s=%d", type);
      jit_decode_len[c] = snom_aud_decode(c, 
					  type,
					  buffer,
					  src_size,
					  jit_decode_buffer[c],
					  sizeof(jit_decode_buffer[0]) / sizeof(jit_decode_buffer[0][0]));
      // if (jit_decode_len[c] == 0) printk("0 t=%d s=%d d=%d", type, src_size, dec_size);
      *length = 0; // this packet is used.
    }
  } // packet available
  
  if (jit_decode_len[c] >= SNOM_AUD_BLOCK_SIZE) {
    short *result = jit_decode_buffer[c];
    jit_decode_pos[c] += SNOM_AUD_BLOCK_SIZE;
    jit_decode_len[c] -= SNOM_AUD_BLOCK_SIZE;
    return result;
  }
  else {
    dtmf_recv_ssrc_valid[c] = 0; // underrun: accept any ssrc from now on!
    return 0;
  }
}

// reset a jitter buffer:
static void jit_recv_reset(int c) {
  int i;
  unsigned long flags;	
  save_flags(flags); cli();

  check_c("jit_recv_reset", c);
  jit_decode_len[c] = 0;
  for(i = 0; i < JITTER_SIZE; i++) {
    jit_recv_len[c][i] = 0;
  }
  dtmf_recv_ssrc_valid[c] = 0;
  snom_aud_reset_codec(c);
  restore_flags(flags);
}

// find a free parking place and put a RTP fragment there:
static void jit_place_content(int c, int type, int len, int ts, unsigned char *buffer) {
  int i;
  int found = 0;
  unsigned long flags;	
  save_flags(flags); cli();
  check_c("jit_place_content", c);
  // if (len == 0) printk("jit_place_content: len == 0\n");
  for(i = 0; i < JITTER_SIZE; i++) {
    if (jit_recv_len[c][i] == 0) {
      if (!found) {
	jit_recv_len[c][i] = len;
	jit_recv_type[c][i] = type;
	jit_recv_timestamp[c][i] = ts;
	// if (len > sizeof(jit_recv_buffer[0][0])) printk("jit_place_content: len = %d\n", len);
	memcpy(jit_recv_buffer[c][i], buffer, len);
	found = 1;
      }
    }
    else {
      if (jit_recv_timestamp[c][i] == ts) {
	// printk("d");
	jit_recv_len[c][i] = 0; // double packet!
      }
    }
  } // for i

#if 0
  for(i = 0; i < JITTER_SIZE; i++) {
    if (jit_recv_len[c][i] > 0) used++;
  }
  if (used > 0) printk("%d", used);
#endif
  restore_flags(flags);
}

// process an incoming rtp packet and put the media data into the
// jitter buffer. This is done according to the media type in smallest
// possible portions.
static void jit_recv_rtp(int c, unsigned char *content, int len) {
  int i;

  check_c("jit_recv_rtp", c);
  if (len >= 12) {
    len -= 12;
    if ((content[0] & 0xc0) == 0x80) {
      const int type = content[1] & 0x7f;
      int ts = ((int)(content[4]) << 24) | 
	((int)(content[5]) << 16) | 
	((int)(content[6]) << 8) | 
	(int)(content[7]);

      unsigned long ssrc = ((int)(content[8]) << 24) | 
	((int)(content[9]) << 16) | 
	((int)(content[10]) << 8) | 
	(int)(content[11]);
      
      // check if the SSRC if is correct:
      if (dtmf_recv_ssrc_valid[c]) {
	if (ssrc != dtmf_recv_ssrc_id[c]) {
	  return; // wrong id
	}    
      }
      else {
	jit_recv_reset(c);
	dtmf_recv_ssrc_id[c] = ssrc;
	dtmf_recv_ssrc_valid[c] = 1;
      }

      if (type == dtmf_recv_type[c] && len == 4) {
	// get the duration of the current digit:
	int duration = content[14] * 256 + content[15];

	// check if we have a new digit:
	if (dtmf_recv_pos[c] >= duration) {
	  // new digit:
	  dtmf_recv_pos[c] = 0;
	}

	// put pieces into the receive buffer:
	while(dtmf_recv_pos[c] < duration) {
	  jit_place_content(c, content[1], 4, ts + dtmf_recv_pos[c], &content[12]);
	  dtmf_recv_pos[c] += SNOM_AUD_BLOCK_SIZE;
	}
      }
      else {
	int frag_length, dec_size;
	snom_aud_get_encoded_length(type, len, &frag_length, &dec_size);
	// if(c > 0) printk("%d", c);
	for(i = 0; i < len; i += frag_length) {
	  jit_place_content(c, content[1], dec_size, ts, &content[12 + i]);
	  ts += dec_size;
	}
      }
    } /* rtp */
  } /* possible length */
}

// put 80 samples into the output buffer. This action may start the
// lower half processing, but the encoding is not done here.
static inline void set_rtp_audio(int c, short *clipped_buffer) {
  int *len = &jit_send_len[c];
  check_c("set_rtp_audio", c);
  if (*len + SNOM_AUD_BLOCK_SIZE <= sizeof(jit_send_buffer[0]) / sizeof(jit_send_buffer[0][0])) {
    memcpy(&(jit_send_buffer[c][*len]), clipped_buffer, SNOM_AUD_BLOCK_SIZE * sizeof(short));
    *len += SNOM_AUD_BLOCK_SIZE;
  }
}

// add 80 samples
static _INLINE_ void clr_signal(int *dest) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE / 8; i++) {
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
  }
}

static _INLINE_ void add_signal(int *dest, const short *src) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE / 8; i++) {
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
  }
}

static _INLINE_ void set_signal(int *dest, const short *src) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE / 8; i++) {
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
  }
}

static _INLINE_ void clip_signal(short *dest, const int *src) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    const int c = *src++;
    *dest++ = CLIP(c);
  }
}

// generic routing to handle all the audio stuff for one packet:
//
// out = 80 shorts output buffer
// in = 80 shorts input from microphone/dtmf (may be zero if turned off)
// rtp = bool if rtp should be used
// synth = 80 synthesize samples (if available)
// echo = bool if in should be copied to out (dtmf loopback)
//
static inline void generate_output(short *out, short *in, int rtp, const short *synth, int echo) {
  // get the audio from all channels:
  static int buffer[SNOM_AUD_BLOCK_SIZE];
  static int buffer2[SNOM_AUD_BLOCK_SIZE];
  static short clipped_buffer[SNOM_AUD_BLOCK_SIZE];
  short *rtp_in[MAX_NUM_RTP];
  int c; // channel index
  int i;

  // white noise for CNG:
  static const short rand_data[SNOM_AUD_BLOCK_SIZE * 4] = {
    27, -5, -29, 43, 22, 7, 23, 16, -36, -26, 29, 40, -4, 48, 40, 39,
    -7, -13, -21, -37, 10, -49, 11, -31, 36, -13, 22, -17, -14, 43,
    21, 44, -4, -27, -5, -32, -46, 21, 35, 29, 25, 34, -11, 46, 46,
    35, 30, -40, -1, 4, 31, 49, -7, -44, -25, 39, -37, -12, 12, 32,
    43, -31, -3, 27, -50, 30, -37, 1, 33, 1, -29, 6, 8, 15, -3, -50,
    32, 28, 8, 8, 46, -48, 19, -2, 42, 0, -31, 26, -44, -16, 2, 20, 8,
    48, -25, 2, -15, -8, 10, -21, -26, 5, 16, 14, 28, 17, 36, 31, -38,
    -22, 22, 27, -35, 31, -2, -39, 44, 40, 32, -20, -35, -31, -27,
    -38, -2, -33, 43, -42, 28, -7, -17, 26, 49, -11, 7, -11, -1, -23,
    30, 6, -7, -39, 19, 6, -9, 38, -2, 28, 12, 26, 30, 0, 6, -1, 48,
    44, 6, -14, 36, -14, -23, 27, 30, 16, -29, -24, -47, 28, -16, -26,
    41, 47, 47, -38, 44, -25, -14, 7, 24, 2, -29, 10, -5, -11, 20,
    -17, 5, -41, 30, -22, 3, -11, 20, -28, -22, -2, -28, 11, -47, 34,
    39, 31, 46, -16, 30, 25, 16, -47, -9, -5, 10, 21, -28, 33, -24,
    -15, 9, 8, -3, -35, 12, 0, 44, 48, 18, 42, -2, 0, 39, 10, -27,
    -20, 31, 11, -2, 26, 0, -40, 7, 36, -43, 44, 49, -32, -46, 25, 33,
    -16, 34, 2, -26, -31, 33, -46, -31, -24, -33, 29, -17, 46, -50,
    24, -10, 7, -41, 33, 31, -18, -35, 36, -11, 14, -29, 39, -23, 21,
    20, 15, 34, 14, -17, -21, -34, 14, 8, 9, -49, 44, 43, 28, 7, 39,
    -42, -43, -17, 42, -37, -43, -9, -14, 39, -37, 0, -6, 20, 18, -47,
    -43, 19, -38, 40, -23, 30, 2, -31, 32, -44, -32, 12, -49
  };
  static short rand_pos = 0;

  // check from where RTP is available:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (rtp && jit_rtp_listen(c)) {
      rtp_in[c] = jit_get_audio(c);
    }
    else
      rtp_in[c] = 0;	
  }

  // update the filtered energy and calculate incoming RTP energy:
  clr_signal(buffer2);
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (rtp_in[c]) add_signal(buffer2, rtp_in[c]);
  }
  energy_rc = 0; // local current energy
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    const int sm = buffer2[i];
    energy_rc += (sm * sm) / 256;
  }

  energy_rf = energy_rf - (energy_rf / 4) + energy_rc / 4;
 
  // if half duplex mode, we sometimes have to use comfort noise to replace the audio.
  if (half_duplex && in) {
    if(energy_rf > 100000){
      duplex_state = HD_RECV;
    }else{
      duplex_state = HD_SEND;
    }
    
    // do the state machine:
    /*if (duplex_state == HD_RECV) {
      if (energy_lf / 2 > energy_rf) {
	duplex_state = HD_SEND;
      }
    }
    else { // duplex_state == HD_SEND
      if (energy_lf < energy_rf / 2) {
	duplex_state = HD_RECV;
      }
    }*/
    
    if(energy_lf < 30000 && energy_rf < 30000){
      duplex_state = HD_SEND;
    }

    // check if we should turn the mic off:
    if (duplex_state == HD_RECV) {
      for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
	in[i] = rand_data[rand_pos++];
      }
      if (rand_pos >= sizeof(rand_data) / (sizeof(rand_data[0])))
	rand_pos = 0;
    }
  }

  // generate speaker output:
  if (synth) set_signal(buffer, synth);
  else clr_signal(buffer);
  if (!half_duplex || duplex_state == HD_RECV) {
    // normal output mode:    
    for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
      buffer[i] += buffer2[i];
    }
  }
  if (echo && in) add_signal(buffer, in);
  clip_signal(out, buffer);
  
  // generate rtp streams:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    int d;
    if (!jit_rtp_active(c)) continue; // this channel is not active

    // initialize the buffer with the microphone input:
    if (in) set_signal(buffer, in);
    else clr_signal(buffer);
    
    // mix the other channels:
    for(d = 0; d < MAX_NUM_RTP; d++) {
      if (c == d) continue; // not mixing to itself
      if (!rtp_in[d]) continue;
      add_signal(buffer, rtp_in[d]);
    } // for d
    
    clip_signal(clipped_buffer, buffer);
    set_rtp_audio(c, clipped_buffer);
  } // for c
}

// process the next block. This procedure is called when the input is
// available and the output needs to be refilled.
static void next_block(short *output, short *input, int echo) {
  int audio_size;
  if (au_wr_pos >= au_rd_pos) 
    audio_size = au_wr_pos - au_rd_pos;
  else
    audio_size = sizeof(au_buffer) / sizeof(au_buffer[0]) - (au_rd_pos - au_wr_pos);
  
  // buffer size hysteris to avoid fast starting:
  if (audio_state == AUDIO_STARTING) {
    if (audio_size < (3 * SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) / 4) audio_size = 0;
    else audio_state = AUDIO_RUNNING;
  }
  else {
    if (audio_size < SNOM_AUD_BLOCK_SIZE) audio_state = AUDIO_STARTING;
  }

  // generate the output:
  switch (mixer_mode) {
  case SNOM_AUD_MIXER_ONLY_RTP:
    generate_output(output, input, 1, 0, echo);
    break;
  case SNOM_AUD_MIXER_ONLY_AUDIO:
    if (audio_size >= SNOM_AUD_BLOCK_SIZE) {
      // if (au_rd_pos < 0 || au_rd_pos + SNOM_AUD_BLOCK_SIZE > SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) printk("au_rd_pos out of range\n");
      generate_output(output, 0, 0, &au_buffer[au_rd_pos], echo);
    }
    else
      generate_output(output, 0, 0, 0, echo);
    break;
  case SNOM_AUD_MIXER_RTP_AUDIO: 
    /* if rtp is available, take it; otherwise audio. This mode has
       been obsoleted! */
  case SNOM_AUD_MIXER_AUDIO_RTP:
    if (audio_size >= SNOM_AUD_BLOCK_SIZE) {
      // if (au_rd_pos < 0 || au_rd_pos + SNOM_AUD_BLOCK_SIZE > SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) printk("au_rd_pos out of range\n");
      generate_output(output, input, 1, &au_buffer[au_rd_pos], echo);
    }
    else
      generate_output(output, input, 1, 0, echo);
    break;
  default: /* SNOM_AUD_MIXER_NOTHING */
    generate_output(output, 0, 0, 0, echo);
    break;
  }

  /* advance the audio read position ANYWAY */
  if (audio_size > 0) {
    au_rd_pos += SNOM_AUD_BLOCK_SIZE;
    if (au_rd_pos >= sizeof(au_buffer) / sizeof(au_buffer[0])) au_rd_pos = 0;
  }
}

/* try to refill a rx and tx buffer */
static int refill_buffer(void) {
  static short tx_location[SNOM_AUD_BLOCK_SIZE];
  static short rx_location[SNOM_AUD_BLOCK_SIZE];
#ifdef CONFIG_GIPS
  ulong flags;
#endif
  volatile cbd_t *tx_bdp = phy.tx_bd_base;
  volatile cbd_t *rx_bdp = phy.rx_bd_base;
  int i;
  int retval = 0;
  // printk("r");
  /* refill both rx and tx buffer if both buffers are "empty" and "ready" */
  for (i = 0; i < HW_NUM_BLOCKS; i++) {
    int rx_avail = !(rx_bdp->cbd_sc & BD_SC_EMPTY);

    if (rx_avail) {
      unsigned char *rx = __va(rx_bdp->cbd_bufaddr);

      rx_bdp->cbd_sc = (rx_bdp->cbd_sc | BD_SC_EMPTY) & (~BD_SC_OV);

      local_time += SNOM_AUD_BLOCK_SIZE;

      retval = 1;

      if (play_dtmf) { // play synthesis sound
	// overwrite the microphone samples:
	get_dtmf(&internal_dtmf, rx_location);
	next_block(tx_location, rx_location, 1);
      }
      else {
	// read the microphone samples:
	read_from_hardware(rx_location, rx);
	if (do_agc)
	  agc_data(rx_location);
	next_block(tx_location, rx_location, 0);
      }
    }
    rx_bdp++;
  }

  for (i = 0; i < HW_NUM_BLOCKS; i++) {
    int tx_avail = !(tx_bdp->cbd_sc & BD_SC_READY);
    if (tx_avail) {
      unsigned char *tx = __va(tx_bdp->cbd_bufaddr);
      tx_bdp->cbd_sc = (tx_bdp->cbd_sc | BD_SC_READY) & (~BD_SC_OV);
      write_to_hardware(tx, tx_location);
    } /* if buffer available */
    tx_bdp++;
  } /* for */

  return retval;
}

void snom_aud_do_tasklet(unsigned long unused) {
  /* check if we have a incoming rtp packet ready: */
  snom_aud_check_incoming();
  
  /* check if we have enough data to send a packet: */
  snom_aud_check_outgoing();
}

DECLARE_TASKLET(snom_aud_tasklet, snom_aud_do_tasklet, 0);

/*
 * This is the snom_aud driver's interrupt routine for a single port
 */
#ifndef CONFIG_SNOM100
static void snom_aud_interrupt(void) {
  u_char events;

  volatile scc_t *sccp;
  int run_tasklet = 0;

  sccp = &cpmp->cp_scc[0];
  events = sccp->scc_scce;

  if (events & (SCCM_GLR | SCCM_GLT | SCCM_TXE | SCCM_BSY)) { /* optimize, since this seldomly happens */
    if (events & SCCM_TXE) {
      //printk("<t>");
      clear_buffers(1);
      // delay_sending_commands = 1000;
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SCCM_BSY) {
      //printk("<b>");
      clear_buffers(1);
      // delay_sending_commands = 1000;
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SCCM_GLR) {
      //printk("<g>");
      clear_buffers(1);
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SCCM_GLT) {
      //printk("<l>");
      clear_buffers(1);
      if (refill_buffer()) run_tasklet = 1;
    }    
  }

  if (events & SCCM_RX) {
    if (refill_buffer()) run_tasklet = 1;
  }
  else if (events & SCCM_TX) {
    if (refill_buffer()) run_tasklet = 1;
  }

  /* check if we have to wake up interruptable processes */
  if (snom_aud_call_back) {
    snom_aud_call_back();
  }
    
  if (run_tasklet) {
    tasklet_schedule(&snom_aud_tasklet);
  }

  sccp->scc_scce = events; 
  // printk("<i>");
}
#else //ifndef snom100
static void snom_aud_interrupt(void) {
  u_char events;

  volatile smc_t *smcp;
  int run_tasklet = 0;

  smcp = &cpmp->cp_smc[1];
  events = smcp->smc_smce;
  if (events & (SMCM_TXE | SMCM_BSY)) { /* optimize, since this seldomly happens */
    if (events & SMCM_TXE) {
/*       printk("<t>"); */
      clear_buffers(1);
      // delay_sending_commands = 1000;
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SMCM_BSY) {
/*       printk("<b>"); */
      clear_buffers(1);
      // delay_sending_commands = 1000;
      if (refill_buffer()) run_tasklet = 1;
    }
  }

  if (events & SMCM_RX) {
/*     printk("r"); */
    if (refill_buffer()) run_tasklet = 1;
  }
  else if (events & SMCM_TX) {
/*     printk("t"); */
    if (refill_buffer()) run_tasklet = 1;
  }

  /* check if we have to wake up interruptable processes */
  if (snom_aud_call_back) {
    snom_aud_call_back();
  }
    
  if (run_tasklet) {
    tasklet_schedule(&snom_aud_tasklet);
  }

  smcp->smc_smce = events;
  /* printk("<i>");*/
}
#endif //ifndef snom_100


/*
 * -------------------------------------------------------------------
 * Here ends the snom_aud interrupt routines.
 * -------------------------------------------------------------------
 */

// calculate how many *bytes* can be written:
int snom_aud_write_avail(snom_aud_info_t *info) {
  unsigned long flags;
  int rem;
  save_flags(flags); cli();
  
  /* determine how many words are available in the buffer (keep one buffer spare) */
  if (au_wr_pos >= au_rd_pos)
    rem = sizeof(au_buffer) - (au_wr_pos - au_rd_pos) * sizeof(short) - SNOM_AUD_BLOCK_SIZE * sizeof(short);
  else
    rem = (au_rd_pos - au_wr_pos) * sizeof(short) - SNOM_AUD_BLOCK_SIZE * sizeof(short);  

  restore_flags(flags);
  // if (rem < 500) printk("<%d>", rem);
  return rem;
}

/*
 * This routine is used to handle the "bottom half" processing for the
 * snom_aud driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * snom_aud_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using snom_aud_sched_event(), and they get done here.
 */

/* Only short words are written to the driver! (pcm16) */
int snom_aud_write(snom_aud_info_t *info, const unsigned char *buf, int count) {
  unsigned long flags;
  int rem = snom_aud_write_avail(info);
  int written = 0;

  if (rem < SNOM_AUD_BLOCK_SIZE) return -EAGAIN;
  if (rem < count) count = rem;

  save_flags(flags); cli();

  while(count >= SNOM_AUD_BLOCK_SIZE * 2) {
    // printk("w");
    // if (au_wr_pos < 0 || au_wr_pos + SNOM_AUD_BLOCK_SIZE > SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) printk("au_rd_pos out of range\n");
    memcpy(&au_buffer[au_wr_pos], buf, SNOM_AUD_BLOCK_SIZE * 2);
    count -= SNOM_AUD_BLOCK_SIZE * 2;
    buf += SNOM_AUD_BLOCK_SIZE * 2;
    written += SNOM_AUD_BLOCK_SIZE * 2;
    au_wr_pos = (au_wr_pos + SNOM_AUD_BLOCK_SIZE) % (sizeof(au_buffer) / sizeof(au_buffer[0]));
  }

  restore_flags(flags);

  if (written > 0)
    return written;
  else
    return -EAGAIN;
}


int snom_aud_read(snom_aud_info_t *info, unsigned char *buf, int count) {
  /* the device does not support reading any more. */
  return -EAGAIN;
}

/*
 * ------------------------------------------------------------
 * snom_aud_close()
 * 
 * This routine is called when the snom_aud port gets closed.
 * ------------------------------------------------------------
 */
void snom_aud_close(snom_aud_info_t *p_info) {
  unsigned long flags;
  int c;
  save_flags(flags); cli();
	
  if(p_info == &info)
    info.is_open = 0;
	
  shutdown_cpm();

  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (rtp_socket[c] != 0) {
      sock_release(rtp_socket[c]);
      rtp_socket[c] = 0;
    }
    if (rtcp_socket[c] != 0) {
      sock_release(rtcp_socket[c]);
      rtcp_socket[c] = 0;
    }
    packet_size[c] = 160;
    packet_type[c] = 0;
    need_marker[c] = 0;
    local_port[c] = 0;
    dtmf_key[c] = 0;
    dtmf_shutoff[c] = 0;
    dtmf_duration[c] = 0;
    dtmf_state[c] = DTMF_STATE_NO_DTMF;
  }
  mixer_mode = SNOM_AUD_MIXER_RTP_AUDIO;

  restore_flags(flags);
}

/*
 * ------------------------------------------------------------
 * snom_aud_open() and friends
 * ------------------------------------------------------------
 */

/*
 * This routine is called whenever a snom_aud port is opened.
 */
int snom_aud_open(int minor_num, snom_aud_info_t **pp_info, int mode, int f_flags) {
  unsigned long flags;
  int ret_val;
  /*  printk("snom_aud_open\n");	*/
  save_flags(flags); cli();
	
  if(minor_num != 0 && minor_num  != 2) {
    ret_val = -EINVAL;
    goto ret_open;
  }

  if((minor_num == 2) && (mode != MODE_WR)) {
    ret_val = -EINVAL;
    goto ret_open;
  }

  *pp_info = &info;
	
  /* Busy ? */
  if((*pp_info)->is_open) {
    ret_val = -EBUSY;
    goto ret_open;
  }
	
  (*pp_info)->is_open = 1;
  snom_aud_mode	      = mode;
  snom_aud_minor_num  = minor_num;

  shutdown_cpm();
  restart_cpm(CPM_RXTX);
  ret_val = 0;
	
 ret_open:
  restore_flags(flags);
  /*  printk("snom_aud_open(exit)\n");	*/
  return ret_val;
}

/*
 * ---------------------------------------------------------------------
 * snom_aud_init() is called at boot-time to initialize the snom_aud driver.
 * ---------------------------------------------------------------------
 */


/*
 * This routine prints out the appropriate snom_aud driver version
 * number, and identifies which options were configured into this
 * driver.
 */
static _INLINE_ void show_snom_aud_version(void) {
  printk(KERN_INFO "%s $Revision: 1.12 $ $Date: 2004/04/27 11:46:42 $\n", snom_aud_name);
}

/*
 * Initializes the AC97
 */
#ifndef CONFIG_SNOM100
static void reset_ac97 (void) {
  int j;
  // enable power for 24 MHz (VCC) and stop SYNC counter(CLK_EN_48KHZ) and deactivate handsfree mode:
  set_port(PORT_SNOM_CSR, VCC | CLK_EN_48KHZ, 0);
  set_port(PORT_SNOM_CSR, 0, MIC_CONTROL);//low = handset
  // deactivate WARM_RESET warm reset 
  set_port(PORT_SNOM_CSR, 0, WARM_RESET);

  // Set the port directions:
  set_port_dir(PORT_SNOM_CSR, OUTPUT_PINS, VCC | MIC_CONTROL | CLK_EN_48KHZ |WARM_RESET);
 
  // release AC97 reset:
  set_port(PORT_C, AC97_RST, 0);
  set_port_dir(PORT_C, OUTPUT_PINS, AC97_RST);
  set_port(PORT_B, 0,HEADSET_LED); //low, NOT CONNECTED
  set_port_dir(PORT_B, OUTPUT_PINS, HEADSET_LED);

  // Perform hard reset:
  for (j = 0; j < 10000; j++);
  set_port(PORT_C, 0, AC97_RST); // reset
  set_port(PORT_SNOM_CSR, 0, CLK_EN_48KHZ); // reset counter
  for (j = 0; j < 1000; j++);
  set_port(PORT_SNOM_CSR, CLK_EN_48KHZ, 0); // restart counter
  for (j = 0; j < 10000; j++);
  set_port(PORT_C, AC97_RST, 0); // end reset
  for (j = 0; j < 1000000; j++);
}

static void init_ac97(void) {
  //configuration commands:
  set_ac97_reg(0x26, 0x0); // all powered 
  set_ac97_reg(0x2a, 0x1); // enables variable rate audio

  //set_ac97_reg(0x2c, 0x3e80); // 16khz dac
  //set_ac97_reg(0x32,0x1f40);// 8Khz adc
  set_ac97_reg(0x2c, 0x1f40); // 8khz dac
  set_ac97_reg(0x32,0x3e80); // 16Khz adc, better transmision
  //set_ac97_reg(0x26,ac97_headset_mic); //headset mic from RJ14


  //audio commands:
  set_ac97_reg(0x02, ac97_handset); // main amp is muted, max. gain  
  set_ac97_reg(0x0e, ac97_mic);//  no mic boost, (sidetone = on, value 1f)   
  set_ac97_reg(0x1c,ac97_adc); // 0x0006 is starting value
  set_ac97_reg(0x18, ac97_dac); // PCM is an input for mux, gain 1/2
  set_ac97_reg(0x06,ac97_speaker); //mono aut max. mute on  
	       
  //set_ac97_reg(0x04, 0x0000); // alt out enabled
  //set_ac97_reg((0x7c | 0x80),0); //read r 0x7c, vendor id1 
  //set_ac97_reg(0x5E, 0x0100); // AC mode control, da -> main out, no mux
  //set_ac97_reg(0x26, 0x0f00); // adc, pcm analog mix power down
  // set_ac97_reg(0x20, 0x0080); // loop back
 
  set_port_dir(PORT_C, OUTPUT_PINS, SHUTDOWN_SPEAKER_PIN);
}
#endif //ifndef config_snom100

static void port_init(void) {
  // get the serial number
  register int serial_num;
  register unsigned char *puc = (unsigned char *) 0xfe030000; 
  puc += 3; 
  serial_num = *puc; 
  serial_num <<= 8; 
  puc++; 
  serial_num |= *puc; 
  serial_num <<= 8; 
  puc++; 
  serial_num |= *puc;
  global_serial_num = serial_num;

#ifdef CONFIG_SNOM100
#ifdef CONFIG_SOUND_SNOM2
  /* turn headset/handset and speaker off. */
  set_port(PORT_C, SHUTDOWN_SPEAKER_PIN | SHUTDOWN_PIN, 0);
#else
  /* have to stay compatible to the first 5 boards. */
  set_port(PORT_C, 0, SHUTDOWN_PIN | SHUTDOWN_SPEAKER_PIN);
#endif
	
  set_port_dir(PORT_C, OUTPUT_PINS, SHUTDOWN_PIN | SHUTDOWN_SPEAKER_PIN);
#endif
}

#ifdef CONFIG_SNOM100
/*
* The snom_aud driver boot-time initialization code!
*/

#define NO_OF_BOOT_BYTES 17

static unsigned char ucBootSequence[NO_OF_BOOT_BYTES] = {
	
  0x82, 0x89, /* set control register to 2.048 Mhz, u-law, non-delayed data 
		 timing, normal operation and NO power amp enabled in PDN. */
  0xCA, 0x80, /* set receive timeslot to 0 */
  0xD2, 0x80, /* set transmit timeslot to 0 */
  0xA2, 0x00, /* set receiver gain to off. */
  0xAA, 0x00, /* set transmitter gain to off. */
  0xB3, 0xFF, /* write hybal register 0 */
  0xBB, 0xFF, /* write hybal register 1 */
  0xC3, 0xFF, /* write hybal register 2 */
  0x00        /* power up ! */
};

#define MAX_RX_GAIN_LEVEL  0xFF 
#define MAX_TX_GAIN_LEVEL  0xFF

#define CS_H     0x0400      
#define CCLK_H   0x0080  
#define CI_H     0x0040

static void pipe_out_byte(unsigned char ucVal)
{
  int i;
	
  volatile immap_t	*immap;
  immap = (immap_t *)IMAP_ADDR;	/* and to internal registers */
	
#if 0
  printk("Piping out %02x\n", (int) ucVal);
#endif
	
  /* Disable chip select and set data and clock to high. This makes sure 
     that the TP3070 synchronizes. */
	
  immap->im_ioport.iop_pcdat |=  CS_H | CCLK_H | CI_H;
	
  /* Enable chip select */
	
  immap->im_ioport.iop_pcdat &= ~CS_H;
	
  for(i=7;i>=0;i--)
    {
      /* Set the data value */
		
      if(ucVal & (1<<i))
	immap->im_ioport.iop_pcdat |= CI_H; /* set data bit */
      else
	immap->im_ioport.iop_pcdat &= ~CI_H; /* clear data bit */
		
      /* Toggle the clock to latch the data. */
		
      immap->im_ioport.iop_pcdat &= ~CCLK_H;  
		
      immap->im_ioport.iop_pcdat |= CCLK_H;  
		
    }
	
  /* Disable chip select and set data and clock to high. This 
     makes sure that the TP3070 synchronizes. */
	
  immap->im_ioport.iop_pcdat |= CS_H | CCLK_H | CI_H;
	
}

static void pipe_out_boot_sequence(void)
{
  int i;
	
  for(i=0;i<NO_OF_BOOT_BYTES;i++) {
    pipe_out_byte(ucBootSequence[i]);
  }
}

/*
* This routine initializes the National Semiconductor TP3070 chip. 
* Several port pins are used the coinfigure the chip via a 
* primitive serial interface.
*
* See data sheet for details.
*/

static void tp3070_init(void)
{
  volatile	immap_t		*immap;
	
  immap = (immap_t *)IMAP_ADDR;	/* and to internal registers */
	
  /* configure the selected three port C pins as general 
     purpose outputs. */
	
  immap->im_ioport.iop_pcpar &= ~(CS_H | CCLK_H | CI_H);
  immap->im_ioport.iop_pcdir |=  (CS_H | CCLK_H | CI_H);
  immap->im_ioport.iop_pcso  &= ~(CS_H | CCLK_H | CI_H);
  immap->im_ioport.iop_pcint &= ~(CS_H | CCLK_H | CI_H);
	
  /* Set all to high */
	
  immap->im_ioport.iop_pcdat |= (CS_H | CCLK_H | CI_H);
	
  /* Initialize the TP3070 */
	
  pipe_out_boot_sequence();
}
#endif

#ifdef CONFIG_SNOM100
static void tsa_init(void)
{
  volatile	cpm8xx_t	*cp;
  volatile	immap_t		*immap;
  volatile      unsigned long   *si_ram;
	
  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* and to internal registers */
	
  immap = (immap_t *)IMAP_ADDR;	
	
  /* Program SI RAM to route time slot 0 to SMC2 and to ignore the
     other 31 timeslots.
  */
	
  /* get 32 bit pointer */
	
  si_ram = (unsigned long *) &(cp->cp_siram);
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = SMC2, 1 byte
     CNT = 0 and BYT = 1 and NOT last entry for both RXb and TXb entry.
  */
	
  si_ram[32] = 0x01820000;
  si_ram[96] = 0x01820000;
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 16 byte CNT = 15 and BYT = 1 and NOT last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[33] = 0x01fe0000;
  si_ram[97] = 0x01fe0000;
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 15 byte CNT = 14 and BYT = 1 and is last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[34] = 0x01fb0000;
  si_ram[98] = 0x01fb0000;
	
  /* Set the SIMODE register to SMC2 connected to SI, normal mode,
     RFSDx = 00 0 bit delay, DSCx = normal speed, common receive and
     transmit clocks CRTb = 1, STZb = 0 normal, CE = 1 data sent and
     received on falling clock edge, FEb = 1 frame sync edge rising,
     GMb = 0, TFSDb = 00 0 bit delay.
	
     Leave setting for SMC1 as is.
  */
	
  cp->cp_simode = 0x80480000 | (cp->cp_simode & 0xffff);
	
  /* Configure PA2 as L1RCLKB, PA10 as L1RXBD, PA11 as L1TXDB and
     PC6 as L1RSYNCB 
  */
	
  immap->im_ioport.iop_papar |= 0x2030;
  immap->im_ioport.iop_padir |= 0x0030;
  immap->im_ioport.iop_padir &= 0xdfff;
  immap->im_ioport.iop_pcpar |= 0x0200;
	
  /* Set SIGMR to enable TDMb, static TDMa and TDMb RAM division */
	
  cp->cp_sigmr = 0x0a;
}

/* 
* This routine configures the SMC2 to transparent mode, sets up the 
* required buffers and their descriptors, resets the event mask and 
* allocates and activates the interrupts.
*/

static void smc2_init(void)
{
  int                 j;
  uint                mem_addr, dp_addr;
  volatile immap_t    *immap;
  volatile cbd_t      *bdp;
  volatile smc_t      *sp;
  volatile smc_uart_t *up;
  volatile cpm8xx_t   *cp;
	
  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* addr internal registers */
	
  immap = (immap_t *)IMAP_ADDR;
	
  memset((void *) &phy, 0, sizeof(snom_aud_physical_dev)); /* clear buffers */
	
  /* We need to allocate a transmit and receive buffer
   * descriptors from dual port ram, and a character
   * buffer area from host mem.
   */

  dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * HW_NUM_BLOCKS);
	
  /* Allocate space for FIFOs in the host memory. */
	
  mem_addr = m8xx_cpm_hostalloc(HW_NUM_BLOCKS * CPM_BUF_SIZE);
  memset((void *) mem_addr, 0, HW_NUM_BLOCKS * CPM_BUF_SIZE);
	
  /* Set the physical address of the host memory
   * buffers in the buffer descriptors, and the
   * virtual address for us to work with.
   */
	
  bdp = (cbd_t *)&cp->cp_dpmem[dp_addr];
  phy.rx_cur = phy.rx_bd_base = (cbd_t *)bdp;
	
  for (j = 0; j < HW_NUM_BLOCKS - 1; j++) {
    bdp->cbd_bufaddr = __pa(mem_addr);
    bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_INTRPT);
    bdp->cbd_datlen = CPM_BUF_SIZE;
    mem_addr += CPM_BUF_SIZE;
    bdp++;
  }
	
  bdp->cbd_bufaddr = __pa(mem_addr);
  bdp->cbd_sc = (BD_SC_WRAP | BD_SC_EMPTY | BD_SC_INTRPT);
  bdp->cbd_datlen = CPM_BUF_SIZE;
	
  sp = &cp->cp_smc[1];
  up = (smc_uart_t *)&cp->cp_dparam[PROFF_SMC2];
  up->smc_rbase = dp_addr;
	
  dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * HW_NUM_BLOCKS);
	
  /* Allocate space for FIFOs in the host memory.
   */
  mem_addr = m8xx_cpm_hostalloc(HW_NUM_BLOCKS * CPM_BUF_SIZE);
  memset((void *) mem_addr, 0, HW_NUM_BLOCKS * CPM_BUF_SIZE);
	
  /* Set the physical address of the host memory
   * buffers in the buffer descriptors, and the
   * virtual address for us to work with.
   */
  bdp = (cbd_t *)&cp->cp_dpmem[dp_addr];
  phy.tx_cur = phy.tx_bd_base = (cbd_t *)bdp;
	
  for (j = 0; j < HW_NUM_BLOCKS - 1; j++) {
    bdp->cbd_bufaddr = __pa(mem_addr);
    bdp->cbd_sc = BD_SC_READY;
    bdp->cbd_datlen = CPM_BUF_SIZE; 
    mem_addr += CPM_BUF_SIZE;
    bdp++;
  }
  bdp->cbd_bufaddr = __pa(mem_addr);
  bdp->cbd_sc = BD_SC_READY | BD_SC_WRAP;
  bdp->cbd_datlen = CPM_BUF_SIZE;
	
  up->smc_tbase = dp_addr;
	
  /* Set up the uart parameters in the parameter ram. */
	
  up->smc_rfcr = SMC_EB;
  up->smc_tfcr = SMC_EB;
	
  /* Set this to 1 for now, so we get single character interrupts.
     Using idle charater time requires some additional tuning.
  */
	
  up->smc_mrblr = SNOM_AUD_BLOCK_SIZE;
	
  /* Send the CPM an initialize command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SMC2, CPM_CR_INIT_TRX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* Set SDMAs' U-bus arbitration priority level to level 5 which
     is normal operation. */
	
  immap->im_siu_conf.sc_sdcr = 1;
	
  /* Set transparent mode, reversed data, normal operation (i.e.
     no loopback) and do NOT enable interrupts.
  */
	
  sp->smc_smcmr = 0x3930;
	
  /* Enable all interrupts and clear all pending events. */
	
  sp->smc_smcm |= SMCM_TXE | SMCM_BSY | SMCM_TX | SMCM_RX;
  sp->smc_smce = 0xff;
	
  /* Install interrupt handler. */
	
  cpm_install_handler(CPMVEC_SMC2, (void *) snom_aud_interrupt, NULL);
}
#else //ifdef CONFIG_SNOM100
/*
 * This function is used to configure the second time multiplexed channel 
 * (TDMb) of the time slot assigner (TSA) of the serial interface (SI) to
 * route out data to the second serial management controller (SCC1). It
 * also configures the port pins to do their special function.
 */
// tsa is programed to take 12 bytes (96 bits) and ignore rest (up to 256 bits)
// sample is one bit late form sync
static void tsa_init(void)
{
  volatile	cpm8xx_t	*cp;
  volatile	immap_t		*immap;
  volatile      unsigned long   *si_ram;
  int i;

  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* and to internal registers */
	
  immap = (immap_t *)IMAP_ADDR;	
	
  /* Program SI RAM to route first 96 bits (12 bytes) to SCC1 and to
     ignore the rest. */
  
  /* get 32 bit pointer */
	
  si_ram = (unsigned long *) &(cp->cp_siram);

  /* reset the memory */
  for(i = 0; i < 128; i++) si_ram[i] = 0x00010000L;
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = SCC1, 12 byte
     CNT = 11 and BYT = 1 and NOT last entry for both RXb and TX000000b entry.
  */
	
  si_ram[32] = (1L << 22) | (11L << 18) | (1L << 17);
  si_ram[96] = (1L << 22) | (11L << 18) | (1L << 17);

  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 16 byte CNT = 15 and BYT = 1 and NOT last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[33] = (0L << 22) | (15L << 18) | (1L << 17);
  si_ram[97] = (0L << 22) | (15L << 18) | (1L << 17);
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 4 byte CNT = 3 and BYT = 1 and is last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[34] = (0L << 22) | (3L << 18) | (1L << 17) | (1L << 16);
  si_ram[98] = (0L << 22) | (3L << 18) | (1L << 17) | (1L << 16);

  /* Set the SIMODE register to SCC1 connected to SI, normal mode,
     RFSDx = 00 0 bit delay, DSCx = 0 normal speed, common receive and
     transmit clocks CRTb = 1, STZb = 0 normal, CE = 0 data sent on
     rising and received on falling clock edge, FEb = 0 frame sync edge
     falling edge, GMb = 0, TFSDb = 00 0 bit delay.
  */
	
  cp->cp_simode = 0x01410000 | (cp->cp_simode & 0xffff);
  // cp->cp_simode = 0x05400000 | (cp->cp_simode & 0xffff);
  cp->cp_sicr = 0x40 | (cp->cp_sicr & 0xffffff00);

  /* Configure PA2 as L1RCLKB, PA10 as L1RXBD, PA11 as L1TXDB and
     PC6 as L1RSYNCB 
  */
	
  immap->im_ioport.iop_papar |=  0x2030; /* use PA2, PA10 and PA11 as SPR */
  immap->im_ioport.iop_padir |=  0x0030; /* L1TXDB, L1RXDB, L1RCLKB, from table 34-1 */
  immap->im_ioport.iop_padir &= ~0x2000; /* make PA2 and PA10 input */
  immap->im_ioport.iop_pcpar |=  0x0200; /* use PC6 as L1RSZNCB, see table 34-11 */
	
  /* Set SIGMR to enable TDMb, static TDMa and TDMb RAM division */
	
  cp->cp_sigmr = 0x0a;
}

/* 
 * This routine configures the SCC1 to transparent mode, sets up the 
 * required buffers and their descriptors, resets the event mask and 
 * allocates and activates the interrupts.
 */
static void scc1_init(void) {
  int                 j;
  uint                mem_addr, dp_addr;
  volatile immap_t    *immap;
  volatile cbd_t      *bdp;
  volatile scc_t      *sp; // conf. registers for scc
  volatile sccp_t     *up; // parameter ram structure for scc
  volatile cpm8xx_t   *cp; // whole cpm structure
	
  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* addr internal registers */
	
  immap = (immap_t *)IMAP_ADDR;
	
  memset((void *) &phy, 0, sizeof(snom_aud_physical_dev)); /* clear buffers */
	
  /* We need to allocate a transmit and receive buffer
   * descriptors from dual port ram, and a character
   * buffer area from host mem.
   */

  dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * HW_NUM_BLOCKS);
  /* Allocate space for FIFOs in the host memory. */

  mem_addr = m8xx_cpm_hostalloc(HW_NUM_BLOCKS * CPM_BUF_SIZE);
  memset((void *) mem_addr, 0, HW_NUM_BLOCKS * CPM_BUF_SIZE);
	
  /* Set the physical address of the host memory
   * buffers in the buffer descriptors, and the
   * virtual address for us to work with.
   */
	
  bdp = (cbd_t *)&cp->cp_dpmem[dp_addr];
  phy.rx_cur = phy.rx_bd_base = (cbd_t *)bdp;
	
  for (j = 0; j < HW_NUM_BLOCKS - 1; j++) {
    bdp->cbd_bufaddr = __pa(mem_addr);
    bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_INTRPT);
    bdp->cbd_datlen = CPM_BUF_SIZE;
    mem_addr += CPM_BUF_SIZE;
    bdp++;
  }
	
  bdp->cbd_bufaddr = __pa(mem_addr);
  bdp->cbd_sc = (BD_SC_WRAP | BD_SC_EMPTY | BD_SC_INTRPT);
  bdp->cbd_datlen = CPM_BUF_SIZE;
	
  sp = &cp->cp_scc[0];
  up = (sccp_t *)&cp->cp_dparam[PROFF_SCC1];
  up->scc_rbase = dp_addr;
	
  dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * HW_NUM_BLOCKS);
  /* Allocate space for FIFOs in the host memory.
   */
  mem_addr = m8xx_cpm_hostalloc(HW_NUM_BLOCKS * CPM_BUF_SIZE);
  memset((void *) mem_addr, 0, HW_NUM_BLOCKS * CPM_BUF_SIZE);
	
  /* Set the physical address of the host memory
   * buffers in the buffer descriptors, and the
   * virtual address for us to work with.
   */
  bdp = (cbd_t *)&cp->cp_dpmem[dp_addr];
  phy.tx_cur = phy.tx_bd_base = (cbd_t *)bdp;
	
  for (j = 0; j < HW_NUM_BLOCKS - 1; j++) {
    bdp->cbd_bufaddr = __pa(mem_addr);
    bdp->cbd_sc = BD_SC_READY;
    bdp->cbd_datlen = CPM_BUF_SIZE; 
    mem_addr += CPM_BUF_SIZE;
    bdp++;
  }
  bdp->cbd_bufaddr = __pa(mem_addr);
  bdp->cbd_sc = BD_SC_READY | BD_SC_WRAP;
  bdp->cbd_datlen = CPM_BUF_SIZE;
	
  up->scc_tbase = dp_addr;
	
  /* Set up the uart parameters in the parameter ram. */
	
  up->scc_rfcr = SCC_EB;
  up->scc_tfcr = SCC_EB;
	
  /* Set this to 1 for now, so we get single character interrupts.
     Using idle charater time requires some additional tuning.
  */
	
  up->scc_mrblr = CPM_BUF_SIZE;
	
  /* Send the CPM an initialize command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_INIT_TRX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* Set SDMAs' U-bus arbitration priority level to level 5 which
     is normal operation. */
	
  immap->im_siu_conf.sc_sdcr = 1;
	
  /* Set transparent mode, reversed data, normal operation (i.e.
     no loopback) and do NOT enable interrupts.
  */
       
  sp->scc_gsmrh =
    /* total transparent mode */
    SCC_GSMRH_TRX | SCC_GSMRH_TTX | 
    /* inherent synchronization, see 29.4.2.2 */
    SCC_GSMRH_CDP | SCC_GSMRH_CTSP | SCC_GSMRH_CDS | SCC_GSMRH_CTSS |
    /* transmit is synchronous to receiving */
    // SCC_GSMRH_TXSY |
    /* reverse byte order: */
    SCC_GSMRH_REVD;
  sp->scc_gsmrl = 
    /* no adjustments to sample point */
    SCC_GSMRL_EDGE_NONE; 

  /* Enable all interrupts and clear all pending events. */
	
  sp->scc_sccm |=  SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX; //| SCCM_GLT SCCM_GLR  |
  sp->scc_scce = 0xffff;
	
  /* Install interrupt handler. */
	
  cpm_install_handler(CPMVEC_SCC1, (void *) snom_aud_interrupt, NULL);
}
#endif // non-snom 100

static void init_structures(void) {
  int i, j;

  /* init the mem structures */
  for(j = 0; j < MAX_NUM_RTP; j++) {
    dtmf_recv_type[j] = 101;
#if 0
    for(i = 0; i < JITTER_SIZE; i++) {
      memset(jit_recv_buffer[j][i], 0, sizeof(jit_recv_buffer[0][0]));
    }
    
    memset(jit_decode_buffer[j], 0, sizeof(jit_decode_buffer[0]));
#endif

    rtp_socket[j] = 0;
    rtcp_socket[j] = 0;
    local_port[j] = 0;
    packet_size[j] = 0;
    packet_type[j] = 0;
    local_ssrc[j] = 0;
    need_marker[j] = 2;
    packet_seqnr[j] = 0;
    remote_adr[j].sin_addr.s_addr = 0;
    remote_adr[j].sin_port = 0;
    dtmf_recv_ssrc_valid[j] = 0;

    jit_decode_pos[j] = 0;
    jit_decode_len[j] = 0;
    jit_decode_start[j] = 1;

    for(i = 0; i < JITTER_SIZE; i++) {
      jit_recv_len[j][i] = 0;
      jit_recv_type[j][i] = 0;
      jit_recv_timestamp[j][i] = 0;
    }

    jit_send_len[j] = 0;

    jit_encoded_len[j] = 0;
    jit_unencoded_len[j] = 0;

    external_dtmf[j].key = 0;
    external_dtmf[j].pos0 = 0;
    external_dtmf[j].pos1 = 0;
  }

  internal_dtmf.key = 0;
  internal_dtmf.pos0 = 0;
  internal_dtmf.pos1 = 0;
}

int __init snom_aud_init(void) {
  extern void snomcodec_init(void);
  
#if 0
  printk("snom_aud_init:\n");
#endif
	
  show_snom_aud_version();
  init_structures();
 
#ifdef CONFIG_SNOM100
  port_init();
  tp3070_init(); 
  tsa_init();
  smc2_init();
#else
  reset_ac97();
  tsa_init();
  scc1_init();
  init_ac97();
  port_init();
  hardware_platform();
#endif
  snomcodec_init();
  return 0;
}

#endif // CONFIG_SOUND_SNOM || CONFIG_SOUND_SNOM2 
