
/* 
 * Author: Nicolas-Peter Pohland
 *
 * Desription : this file defines the structures needed to 
 *              make the SNOM Audio HW work. 
 *
 */

#ifndef SNOM_AUD_H
#define SNOM_AUD_H

#define MAX_NUM_RTP 5
#define SNOM200_B 0x100a0a // snom200B start with this ser. number

// description of the ICE contact:
struct ice_contact {
  unsigned char id[16]; // the random ID
  unsigned long dest_adr; // destination address
  int dest_port; // destination port
};

#define PT_PCMU      0
#define PT_PCMA      8

#ifdef CONFIG_AC97

#define COMMAND_ONLY 0xE000
#define SOUND_ONLY 0x9800
#define SOUND_COMMAND 0xF800

/* AC'97 registers */
#define  AC97_RESET              0x0000 /* read only reg*/    
#define  AC97_MASTER_VOL_STEREO  0x0002     
#define  AC97_HEADPHONE_VOL      0x0004      // 
#define  AC97_MASTER_VOL_MONO    0x0006      // TAD Output
#define  AC97_MASTER_TONE        0x0008      //
#define  AC97_PCBEEP_VOL         0x000a      // none
#define  AC97_PHONE_VOL          0x000c      // TAD Input (mono)
#define  AC97_MIC_VOL            0x000e      // MIC Input (mono)
#define  AC97_LINE_IN_VOL        0x0010      // Line Input (stereo)
#define  AC97_CD_VOL             0x0012      // CD Input (stereo)
#define  AC97_VIDEO_VOL          0x0014      // none
#define  AC97_AUX_VOL            0x0016      // Aux Input (stereo)
#define  AC97_PCMOUT_VOL         0x0018      // Wave Output (stereo)
#define  AC97_RECORD_SELECT      0x001a      //
#define  AC97_RECORD_GAIN        0x001c
#define  AC97_RECORD_GAIN_MIC    0x001e
#define  AC97_GENERAL_PURPOSE    0x0020
#define  AC97_3D_CONTROL         0x0022
#define  AC97_MODEM_RATE         0x0024
#define  AC97_POWER_CONTROL      0x0026
#define  AC97_AC_MODE            0X005E

/* registers 0x0028 - 0x0058 are reserved */

/* AC'97 2.0 */
#define AC97_EXTENDED_ID	0x0028	/* Extended Audio ID */
#define AC97_EXTENDED_STATUS	0x002A	/* Extended Audio Status */
#define AC97_PCM_FRONT_DAC_RATE 0x002C  /* PCM Front DAC Rate */
#define AC97_PCM_SURR_DAC_RATE  0x002E  /* PCM Surround DAC Rate */
#define AC97_PCM_LFE_DAC_RATE   0x0030  /* PCM LFE DAC Rate */
#define AC97_PCM_LR_ADC_RATE	0x0032	/* PCM LR DAC Rate */
#define AC97_PCM_MIC_ADC_RATE   0x0034  /* PCM MIC ADC Rate */
#define AC97_CENTER_LFE_MASTER  0x0036  /* Center + LFE Master Volume */
#define AC97_SURROUND_MASTER    0x0038  /* Surround (Rear) Master Volume */
#define AC97_RESERVED_3A	0x003A	/* Reserved */
/* range 0x3c-0x58 - MODEM */

/* registers 0x005a - 0x007a are vendor reserved */

#define  AC97_VENDOR_ID1         0x007c
#define  AC97_VENDOR_ID2         0x007e

/* volume control bit defines */

#define AC97_MUTE                0x8000
#define MIC_MUTE                 0x8000
#define ADC_MUTE                 0x8000
#define AC97_MICBOOST            0x0040
#define AC97_MICATTEN            0x001f
#define AC97_LEFTVOL             0x3f00
#define AC97_RIGHTVOL            0x003f
#define AC97_MIC_SEL             0x0100 /* Mic Select 0=Mic1, 1=Mic2 */
#define AC97_LPBK                0x0080 /* Loopback mode */
#define AC97_REC_GAIN_LEFT       0X0f00
#define AC97_REC_GAIN_RIGHT      0X000f
#define AC97_MONOVOL             0x003f
#define ON                       0x0001
#define EAPD_ON                  0x8000
#define AC97_DAC_ATT_R           0x001f
#define AC97_DAC_ATT_L           0x1f00
#define AC97_MIX                 0x0200
#define DAC_DIRR                 0x0100
/* record mux defines */

#define AC97_REC_MUX_MIC          0x0000
#define AC97_REC_MUX_CD           0x0101
#define AC97_REC_MUX_VIDEO        0x0202      /* not used */
#define AC97_REC_MUX_AUX          0x0303      
#define AC97_REC_MUX_LINE         0x0404      
#define AC97_REC_MUX_STEREO_MIX   0x0505
#define AC97_REC_MUX_MONO_MIX     0x0606
#define AC97_REC_MUX_PHONE        0x0707


/* general purpose register bit defines */


#define AC97_GP_MIX              0x0200      /* Mono output select 0=Mix, 1=Mic */
#define AC97_GP_RLBK             0x0400      /* Remote Loopback - Modem line codec */
#define AC97_GP_LLBK             0x0800      /* Local Loopback - Modem Line codec */
#define AC97_GP_LD               0x1000      /* Loudness 1=on */
#define AC97_GP_3D               0x2000      /* 3D Enhancement 1=on */
#define AC97_GP_ST               0x4000      /* Stereo Enhancement 1=on */
#define AC97_GP_POP              0x8000      /* Pcm Out Path, 0=pre 3D, 1=post 3D */


/* powerdown control and status bit defines */

/* status */
#define AC97_PWR_MDM             0x0010      /* Modem section ready */
#define AC97_PWR_REF             0x0008      /* Vref nominal */
#define AC97_PWR_ANL             0x0004      /* Analog section ready */
#define AC97_PWR_DAC             0x0002      /* DAC section ready */
#define AC97_PWR_ADC             0x0001      /* ADC section ready */

/* control */
#define AC97_PWR_PR0             0x0100      /* ADC and Mux powerdown */
#define AC97_PWR_PR1             0x0200      /* DAC powerdown */
#define AC97_PWR_PR2             0x0400      /* Output mixer powerdown (Vref on) */
#define AC97_PWR_PR3             0x0800      /* Output mixer powerdown (Vref off) */
#define AC97_PWR_PR4             0x1000      /* AC-link powerdown */
#define AC97_PWR_PR5             0x2000      /* Internal Clk disable */
#define AC97_PWR_PR6             0x4000      /* HP amp powerdown */
#define AC97_PWR_PR7             0x8000      /* Modem off - if supported */

/* useful power states */
#define AC97_PWR_D0              0x0000      /* everything on */
#define AC97_PWR_D1              AC97_PWR_PR0|AC97_PWR_PR1|AC97_PWR_PR4
#define AC97_PWR_D2              AC97_PWR_PR0|AC97_PWR_PR1|AC97_PWR_PR2|AC97_PWR_PR3|AC97_PWR_PR4
#define AC97_PWR_D3              AC97_PWR_PR0|AC97_PWR_PR1|AC97_PWR_PR2|AC97_PWR_PR3|AC97_PWR_PR4
#define AC97_PWR_ANLOFF          AC97_PWR_PR2|AC97_PWR_PR3  /* analog section off */

/* Total number of defined registers.  */
#define AC97_REG_CNT 64

typedef enum AudioCmd {
	SET_LINE_IN_ATTEN_AUD_CMD,
	LINE_IN_SHUTDOWN_AUD_CMD,
	LINE_IN_ATTEN_AUD_CMD,
	SET_SPEAKER_ATTEN_AUD_CMD,
	SHUTDOWN_SPEAKER_AUD_CMD,
 	SET_HANDSET_ATTEN_AUD_CMD,
	SHUTDOWN_HANDSET_AUD_CMD,
	SET_HEADSET_ATTEN_AUD_CMD,
	SHUTDOWN_HEADSET_AUD_CMD,
	MIC_SEL_AUD_CMD,
	MIC_BOOST_AUD_CMD,
	MIC_MUTE_AUD_CMD,
	SET_MIC_ATTEN_AUD_CMD,
	SET_OUTPUT_SRC_AUD_CMD,
	LPBK_AUD_CMD,
	DA_RATE_AUD_CMD,
	AD_RATE_AUD_CMD,
	RATE_ANABLE_AUD_CMD,
	ADC_MUX_SELECT_AUD_CMD,
	ADC_GAIN_AUD_CMD,
	ADC_MUTE_AUD_CMD,
	TSA_INIT_AUD_CMD, 
	CHIP_INIT_AUD_CMD, 
	SHUTDOWN_AUD_CMD,
	GET_FRAME_SIZE_AUD_CMD,
	FILL_BUFFER_AUD_CMD,
	EMPTY_BUFFER_AUD_CMD,
	MONOUT_SEL_AUD_CMD,
	READ_REG_AUD_CMD,
	EAPD,
	DAC_DIRMODE
} eAudioCmd;

/*
  INIT_AUD_CMD              - initialize the HW. No argument.
  REINIT_AUD_CMD            - reinitialize the HW. No argument. 
  MUTE_AUD_CMD              - mute all outputs. No argument.
  SET_MIC_GAIN_AUD_CMD      - set microphone gain. Integer value 0-100
  SET_HANDSET_GAIN_AUD_CMD  - set handset gain. Integer value 0-100
  SET_HEADSET_GAIN_AUD_CMD  - set headset gain. Integer value 0-100
  SET_SPEAKER_GAIN_AUD_CMD  - set speaker gain. Integer value 0-100
  SET_OUTPUT_SRC_AUD_CMD    - determine output. eOutputSrc is the value.
  GET_FRAME_SIZE_AUD_CMD    - returns the number of bytes needed for a sample.
  FILL_BUFFER_AUD_CMD       - fills CPM buffer. Generic source buffer.
  EMPTY_BUFFER_AUD_CMD      - empties CPM buffer. Generic destination buffer.
  SET_DATA_RATE             - set rate per second. 8000 or 16000 only.
*/
#endif // CONFIG_AC97

#endif /* SNOM_AUD_H */
