/*
 * SNOM low level audio driver for MPC860 SMC2 1999-2003 snom
 * technology AG
 *
 * I used the uart.c driver from Dan Malek who in turn used the
 * serial.c driver. Theiver was originally written for TP3070, and
 * then was changed to AC97.
 *
 * This module exports the following snom_aud io functions:
 *
 *	int snom_aud_init(void);
 * 
 * Description:
 *
 * This file is the low level driver for the SNOM audio hardware. Its
 * core element is a AC97 compliant codec that is addressed with a
 * SCC1.
 * */

#include <linux/config.h>
#include <linux/types.h>

#if defined (CONFIG_SOUND_SNOM) || defined (CONFIG_SOUND_SNOM2)

#include <asm/mpc8xx.h>

/* Have to check out what we need here ... */
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/soundcard.h>
#include <linux/delay.h>
#include <linux/in.h>
#include <linux/udp.h>
#include <linux/pkt_sched.h>
#include <net/sock.h>
#include <net/ip.h>
#include <net/udp.h>
#include <asm/uaccess.h>
#include <asm/8xx_immap.h>
#include <asm/mpc8xx.h>
#include <asm/checksum.h>
#include "commproc.h"
#include "snomaud.h"
#include "mec.h"

#include <asm/port.h>

#define _INLINE_ inline

/* forward declarations for snomcodec.c: */

static void set_ac97_reg(int, unsigned long);
static void write_to_hardware(unsigned char *ulaw, short *linear);

extern int snom_aud_encode(int channel, int type, short *src, int length, unsigned char *dst, int maxlen);
extern int snom_aud_decode(int channel, int type, unsigned char *src, int length, short *dst, int maxlen);
extern int snom_aud_fill(int channel, short *dst, int maxlen);
extern void snom_aud_get_encoded_length(int type, int length, int *encoded_length, int *unencoded_length);
extern void snom_aud_reset_codec(int channel);
extern void snom_on_stun_available(int c,
				   struct ice_contact *ice_contacts, int ice_contacts_length,
				   unsigned char *msg, int msg_length, const unsigned char *original_addr, int original_port);
extern void snom_aud_sendmsg(struct socket *skt, char *data, int len, int dport);

extern const unsigned char snom_aud_linear2ulaw[16384];
extern const short snom_aud_ulaw2linear[256];
static void init_ac97(void);
static void reset_ac97 (void);
static void hardware_platform(void);

extern void snom_aud_set_audio (unsigned int audio_settings); 
void set_handset (unsigned int aValue);
void set_ringer(unsigned int aValue); 
void set_headset(unsigned int aValue);
void set_loopback(unsigned int aValue);
void set_handset_mic(unsigned int aValue);
void set_headset_mic(unsigned int aValue);
void set_speaker_phone(unsigned int aValue);
unsigned int reverse_volume(unsigned int aValue);
unsigned short adjust_adc_gain(unsigned int volume);
void set_external_io(unsigned int aValue);
extern int global_serial_num;
extern unsigned int hardware_audio_device;

static void jit_recv_reset(int c);
static void jit_recv_rtp(int c, unsigned char *content, int len, unsigned char *adr, int port);

// structure for defining the ICE contacts:
#define MAX_NUM_ICE 5
struct ice_contact_container {
  int length;
  int accept; // true if we should accept results on this container
  struct ice_contact contacts[MAX_NUM_ICE]; // at most 5 at the same time
};

static struct ice_contact_container ice_contacts[MAX_NUM_RTP];

struct dtmf_state_t {
  int key;
  int pos0;
  int pos1;
};

static void set_dtmf(int key, struct dtmf_state_t *dtmf);

#define CPM_BUF_SIZE (CONFIG_SOUND_SNOM_BYTES_PER_SAMPLE * CONFIG_SOUND_SNOM_SAMPLES_PER_8KHZ_SAMPLE * \
	CONFIG_SOUND_SNOM_AUD_BLOCK_SIZE)

/* Shutdown control pins on port C. */ 

#define AC97_RST (1<<10)
#define SHUTDOWN_SPEAKER_PIN (1<<2)
#define HEADSET_LED (1<<3)

#define GPIO0  (1<<7) 
#define GPIO1  (1<<6) 
#define GPIO2  (1<<5) 
#define GPIO3  (1<<4) 
#define GPIO4  (1<<3) 
#define GPIO5  (1<<2)
#define GPIO6  (1<<1)
#define GPIO7  (1<<0)

#ifdef CONFIG_SNOM105

#define WARM_RESET  	GPIO0 // warm reset to codec  (for cpld)
#define CLK_EN_48KHZ	GPIO1 // clk enable for 48 Khz counter, altera
#define MIC_CONTROL 	GPIO3 // handset/headset mic control
#define VCC		GPIO2 // always high output from altera, wired to vcc

#else // CONFIG_SNOM105

#define WARM_RESET  	GPIO4 // warm reset to codec  (for cpld)
#define CLK_EN_48KHZ	GPIO5 // clk enable for 48 Khz counter, altera
#define MIC_CONTROL 	GPIO6 // handsfree mic control
#define VCC		GPIO7 // always high output from altera, wired to vcc

#endif // CONFIG_SNOM105


#ifdef CONFIG_SNOM105
#define ADC_GAIN 0x800d
#define HANDSET_DAC_GAIN 0x0505
#define DAC_GAIN_SPHONE 0x0c0c
#define DAC_RINGER 0x0b0b
#else
#define ADC_GAIN 0x8006
#define HANDSET_DAC_GAIN 0x0606
#define HEADSET_DAC_GAIN 0x0a0a
#define DAC_GAIN_SPHONE 0x0b0b
#define DAC_RINGER 0x0808
#define DAC_GAIN_SPHONE_A 0x0d0d
#endif

#define HANDSET_MIC 0x0000
#define HEADSET_MIC 0x0100
#define LPBK 0x0080
#define SIDE_TONE 0x000a
#define MIC_BOOST 0x0040
#define SOUND_3D 0x2000
#define HEADSET_RJ14 0x8000
#define HEADSET_PC 0x0000

#define HEADSET_IN (1<<3) // pc12 
static unsigned short ac97_speaker = 0x8000;
static unsigned short ac97_handset = 0x8000;
static unsigned short ac97_headset = 0x8000;
static unsigned short ac97_adc = ADC_GAIN;
static unsigned short ac97_dac = DAC_GAIN_SPHONE | 0x8000;// closed at start up
static unsigned short ac97_gp = HANDSET_MIC;
static unsigned short ac97_headset_mic = HEADSET_PC;
static unsigned short ac97_adc_handset;
static unsigned short ac97_adc_sphone;
static unsigned short ac97_dac_sphone;
static unsigned short ac97_mic = SIDE_TONE;// only for side tone

#define MODE_WR 2 /* I dont understand this, anyway. */

#define HW_NUM_BLOCKS CONFIG_SOUND_SNOM_HW_NUM_BLOCKS /* should be 2 */
#define SW_NUM_BLOCKS 32

/* buffers for the audio streams: */
static short au_buffer[SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE];
static unsigned int local_time = 0; /* timestamp on the local machine */
static int tos = IPTOS_PREC_CRITIC_ECP;
// static unsigned int send_time = 0; /* time of last packet sent */
static unsigned int au_rd_pos = 0;
static unsigned int au_wr_pos = 0;

/* statistics for the jitter buffer */
#define WRITE_ADVANCE 160
#define JITTER_SIZE 16

struct socket *rtp_socket[MAX_NUM_RTP] = { 0 };
struct socket *rtcp_socket[MAX_NUM_RTP] = { 0 };
static int local_port[MAX_NUM_RTP] = { 0 };
static unsigned int packet_size[MAX_NUM_RTP] = { 160 };
static unsigned int packet_type[MAX_NUM_RTP] = { 0 }; /* ulaw */
static unsigned long local_ssrc[MAX_NUM_RTP]; /* ssrc identifier */
static int need_marker[MAX_NUM_RTP] = { 2 }; /* bool */
static int sym_rtp[MAX_NUM_RTP] = { 0 }; /* bool */
static unsigned int packet_seqnr[MAX_NUM_RTP] = { 0 };
static struct sockaddr_in remote_adr[MAX_NUM_RTP];

extern short cng_on[MAX_NUM_RTP];
extern unsigned char cng_vol[MAX_NUM_RTP];
static int silence_suppression = 0;
unsigned int sp_vol = 10;
static int handset_volume = 0;

static int stat_rx[MAX_NUM_RTP];
static int stat_tx[MAX_NUM_RTP];

// the read_buffer is used for recording purposes:
#define READ_BUFFER_LENGTH (SNOM_AUD_BLOCK_SIZE * 100)
static short read_buffer[READ_BUFFER_LENGTH]; // max. 1000 ms
static int read_wpos = 0;
static int read_rpos = 0;
static int read_recording = 0; // boolean that determines if we are recording

int snom_aud_get_stat(struct snom_aud_rtp_stat_s *stat) {
  int c = stat->channel;
  if (c < 0 || c >= MAX_NUM_RTP) return -EINVAL;
  stat->rx_packets = stat_rx[c];
  stat->tx_packets = stat_tx[c];
  stat_rx[c] = 0;
  stat_tx[c] = 0;
  return 0;
}

static _INLINE_ void check_c(char *s, int c) {
  //   if (c < 0 || c >= MAX_NUM_RTP) printk("%s: %d\n", s, c);
}

// check if we should do AGC (which is the case if 
static int do_agc = 1;
static void check_for_agc(void) {
  int i;
  do_agc = 1;
  for(i = 0; i < MAX_NUM_RTP; i++) {
    if (packet_type[i] == 18) {
      do_agc = 0;
      break;
    }
  }
}

// check if we are ready for sending and receiving:
static _INLINE_ int jit_rtp_active(int c) {
  // say if a channel is sending or not:
  check_c("jit_rtp_active", c);
  return rtp_socket[c] && remote_adr[c].sin_addr.s_addr != 0;
}

// we might not be ready for sending, but for receiving:
static _INLINE_ int jit_rtp_listen(int c) {
  // say if a channel is sending or not:
  check_c("jit_rtp_listen", c);
  return rtp_socket[c] != 0;
}

static int mixer_mode = SNOM_AUD_MIXER_RTP_AUDIO;
static int half_duplex = 0;

/* The bloody half duplex mode is now implemented as state
   machine. That makes the transition from the speaker roles more
   smooth and easier to understand in software. */
#define HD_SEND 0 // Microphone on, speaker off
#define HD_RECV 1 // Speaker turned on, Microphone off
static int duplex_state = HD_SEND;


#define AUDIO_STARTING 0
#define AUDIO_RUNNING 1
static int audio_state = AUDIO_STARTING;

#define CLIP(s) ((short)((s) > 32767 ? 32767 : ((s) < -32767 ? -32767 : (s))))

// intermediate buffer for decoded data:
static short jit_decode_buffer[MAX_NUM_RTP][320];
static int jit_decode_pos[MAX_NUM_RTP];
static int jit_decode_len[MAX_NUM_RTP];
static int jit_decode_start[MAX_NUM_RTP];

// jitter buffer for incoming packets:
static unsigned char jit_recv_buffer[MAX_NUM_RTP][JITTER_SIZE][320];
static int jit_recv_len[MAX_NUM_RTP][JITTER_SIZE];
static int jit_recv_type[MAX_NUM_RTP][JITTER_SIZE]; // (0..255, includes marker)
static int jit_recv_timestamp[MAX_NUM_RTP][JITTER_SIZE];

// intermedia buffer for unencoded outgoing data:
static short jit_send_buffer[MAX_NUM_RTP][640];
static int jit_send_len[MAX_NUM_RTP];

// buffer for encoded outgoing data:
static unsigned char jit_encoded_buffer[MAX_NUM_RTP][640 + 12]; // 12 bytes for rtp header
static int jit_encoded_len[MAX_NUM_RTP]; // in encoded form
static int jit_unencoded_len[MAX_NUM_RTP]; // in encoded form
static int jit_encoded_ts[MAX_NUM_RTP];

// statistics on the energy levels presented to speaker and from mic:
static unsigned long energy_lc = 0; // local current
static unsigned long energy_lf = 0; // local filtered
static unsigned long energy_lg = 0; // local filtered, fast decay
static unsigned long energy_rc = 0; // remote current
static unsigned long energy_rf = 0; // remote filtered

/* Now come the devices ... */

typedef struct {
	
  /* CPM Buffer Descriptor pointers.*/
  cbd_t		     *rx_bd_base;
  cbd_t		     *rx_cur;
  cbd_t		     *tx_bd_base;
  cbd_t		     *tx_cur;
	
  /* Underruns */
  int                is_underrun;
	
} snom_aud_physical_dev;

typedef struct {
  int is_open;
} snom_aud_info_t;

static snom_aud_info_t info;
static char *snom_aud_name = "CPM SNOM low level audio driver";
static snom_aud_physical_dev phy;
static int snom_aud_minor_num = 0;
static int snom_aud_mode = 0;

static void (*snom_aud_call_back)(void);

/* Forward decl. */

void snom_aud_reset (snom_aud_info_t *info);
void snom_aud_set_tx_gain2(int level);

#define CPM_RX_ONLY 11
#define CPM_TX_ONLY 22
#define CPM_RXTX 33

static _INLINE_ void restart_cpm(int eMode)
{
  /* Get pointer to Communication Processor */

  volatile cpm8xx_t *cp = cpmp;
  volatile scc_t    *sp = &cp->cp_scc[0];
	
  // printk("\ns");
       
  /* Send the CPM a stop transmitter command. They say it is not
     really necessary but who knows. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_STOP_TX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* reset transmitter and receiver by disabling it. */
	
  sp->scc_gsmrl &= ~(SCC_GSMRL_ENR | SCC_GSMRL_ENT);

  /* Enable all interrupts and clear all pending events. */
	
  sp->scc_sccm |= SCCM_GLR | SCCM_GLT | SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX;
  sp->scc_scce = 0xffff;
	
  /* Send the CPM an initialize transmitter command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_INIT_TRX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  phy.is_underrun = 0;
	
  /* Clear pending TX interrupts. */
	
  sp->scc_scce = 0xffff;
	
  /* Send the CPM a restart transmitter command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_RESTART_TX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* Enable it ... */
	
  switch(eMode) {
  case CPM_RX_ONLY :
    sp->scc_gsmrl |= SCC_GSMRL_ENR;
    break;
  case CPM_TX_ONLY :
    sp->scc_gsmrl |= SCC_GSMRL_ENT;
    break;
  case CPM_RXTX :
    sp->scc_gsmrl |= SCC_GSMRL_ENR | SCC_GSMRL_ENT;
    break;
  }
}

/*
 * Automatic Gain Control (AGC). The algorithm we use here reduces the
 * gain immediately when we are reaching the threshold value
 * (MAX_SHORT). Otherwise, the gain is increased every 100 ms if one
 * value was above a threshold (speech detection).
 */ 

static long agc_gain = 32768 / 4; // 32768 means times 4, the maximum amplification

static void init_agc(void) {
  agc_gain = 32768 / 4;
}

// AGC for 80 samples:
static _INLINE_ void agc_data(short *sample) {
  int idx;
  const int old_gain = agc_gain;

  // This factor is used to make the step from one gain factor to
  // another more smooth:
  static const long factor1[80] = {
    0, 410, 819, 1229, 1638, 2048, 2458, 2867, 3277, 3686,
    4096, 4506, 4915, 5325, 5734, 6144, 6554, 6963, 7373, 7782,
    8192, 8602, 9011, 9421, 9830, 10240, 10650, 11059, 11469, 11878,
    12288, 12698, 13107, 13517, 13926, 14336, 14746, 15155, 15565, 15974,
    16384, 16794, 17203, 17613, 18022, 18432, 18842, 19251, 19661, 20070,
    20480, 20890, 21299, 21709, 22118, 22528, 22938, 23347, 23757, 24166,
    24576, 24986, 25395, 25805, 26214, 26624, 27034, 27443, 27853, 28262,
    28672, 29082, 29491, 29901, 30310, 30720, 31130, 31539, 31949, 32358
  };

  // gain table: transform raw energy (divided by 0x400000):

  static const long gain_table[] = { 
    
    32768, 32768, 32768, 32768, 32768, 32768, 31769, 30992, 30366,
    29848, 29409, 29031, 28702, 28412, 28153, 27921, 27710, 27274,
    26875, 26508, 26169, 25854, 25561, 25287, 25031, 24790, 24563,
    24349, 24146, 23954, 23772, 23598, 23433, 23276, 23125, 22981,
    22843, 22710, 22583, 22461, 22343, 22230, 22121, 22015, 21913,
    21815, 21720, 21628, 21538, 21452, 21368, 21287, 21208, 21131,
    21056, 20984, 20913, 20844, 20777, 20712, 20648, 20586, 20526,
    20467, 20409, 20352, 20297, 20243, 20191, 20139, 20089, 20040,
    19991, 19944, 19898, 19852, 19808, 19764, 19721, 19679, 19638,
    19598, 19558, 19519, 19481, 19443, 19406, 19370, 19334, 19299,
    19264, 19231, 19197, 19164, 19132, 19100, 19069, 19038, 19008,
    18978, 18949, 18920, 18891, 18863, 18835, 18808, 18781, 18754,
    18728, 18702, 18677, 18652, 18627, 18602, 18578, 18554, 18531,
    18507, 18485, 18462, 18440, 18417, 18396, 18374, 18353, 18332,
    18311, 18291, 18270, 18250, 18230, 18211, 18191, 18172, 18153,
    18135, 18116, 18098, 18080, 18062, 18044, 18026, 18009, 17992,
    17975, 17958, 17941, 17925, 17909, 17893, 17877, 17861, 17845,
    17830, 17814, 17799, 17784, 17769, 17754, 17740, 17725, 17711,
    17697, 17682, 17668, 17655, 17641, 17627, 17614, 17601, 17587,
    17574, 17561, 17548, 17535, 17523, 17510, 17498, 17485, 17473,
    17461, 17449, 17437, 17425, 17413, 17402, 17390, 17379, 17367,
    17356, 17345, 17334, 17323, 17312, 17301, 17290, 17280, 17269,
    17258, 17248, 17238, 17227, 17217, 17207, 17197, 17187, 17177,
    17167, 17157, 17148, 17138, 17129, 17119, 17110, 17100, 17091,
    17082, 17073, 17063, 17054, 17045, 17036, 17028, 17019, 17010,
    17001, 16993, 16984, 16976, 16967, 16959, 16950, 16942, 16934,
    16926, 16918, 16910, 16902, 16894, 16886, 16878, 16870, 16862,
    16854, 16847, 16839, 16831, 16824, 16816, 16809, 16802, 16794,
    16787, 16780, 16772, 16765, 16758, 16751, 16744, 16737, 16730,
    16723, 16716, 16709, 16702, 16695, 16689, 16682, 16675, 16669,
    16662, 16656, 16649, 16643, 16636, 16630, 16623, 16617, 16611,
    16604, 16598, 16592, 16586, 16579, 16573, 16567, 16561, 16555,
    16549, 16543, 16537, 16531, 16526, 16520, 16514, 16508, 16502,
    16497, 16491, 16485, 16480, 16474, 16468, 16463, 16457, 16452,
    16446, 16441, 16436, 16430, 16425, 16419, 16414, 16409, 16404,
    16398, 16393, 16388, 16384 
  };

#define XSCALE 1048576

  // update the energy if the current energy is at least 25 % of the filtered energy:
  if (energy_lc > energy_lf / 4) {
    int index = energy_lf / XSCALE;
    int fraction = energy_lf - (index * XSCALE);
    if (index < sizeof(gain_table) / sizeof(gain_table[0]) - 1) {
      agc_gain = gain_table[index] + ((gain_table[index + 1] - gain_table[index]) * (fraction / 16)) / (XSCALE / 16);
    }
    else {
      // lowest amplification (saturation):
      agc_gain = gain_table[sizeof(gain_table) / sizeof(gain_table[0]) - 1];
    }
  }

  // amplify the samples according to the gain:
  for(idx = 0; idx < 80; idx++) {
    const long fac = factor1[idx];
    const int mix = (fac * agc_gain + (32768 - fac) * old_gain) / 32768;
    const long result = (mix * (long)(sample[idx])) / (32768 / 2);
    sample[idx] = CLIP(result);
  }  
}


/******************************************************************************/

/*
 * Here are the routines that are called from ioctl().
 */

int snom_aud_get_blocksize     (snom_aud_info_t *info){
  return SNOM_AUD_BLOCK_SIZE;
}

int snom_aud_get_caps          (snom_aud_info_t *info) {
	
  /* TBD: is real time correct ? I do not know the exact associated semantics ... */
	
  return DSP_CAP_DUPLEX | DSP_CAP_BATCH | DSP_CAP_REALTIME; 
}

int  snom_aud_get_formats       (snom_aud_info_t *info) {
  return AFMT_S16_BE;
}

void snom_aud_set_blocking_mode (snom_aud_info_t *info, int f_flags) {
  /* the device never blocks */
}

int  snom_aud_set_format        (snom_aud_info_t *info, int format) {
  if(format == AFMT_S16_BE) {
    return 0;
  } else {
    return -EINVAL;
  }
}

#define AUDIO_VOL 0x0000000f // volume of left speaker is ignored at the moment
#define AUDIO_DEV 0x00000f00

void snom_aud_set_audio (unsigned int audio_settings)
{
  unsigned int  volume;
  hardware_audio_device = (audio_settings & AUDIO_DEV) >> 8;
  volume = audio_settings & AUDIO_VOL;

  // printk("snom_aud_set_audio: %0xd dev:%d vol:%d\n", audio_settings, device, volume); // use carefully, because its so slow

  switch(hardware_audio_device){
  case 0: set_handset (volume); break; // handset
  case 1: set_ringer(volume);break; // casing speaker
  case 2: set_headset(volume); break; // headset
  case 3: set_loopback(volume); break; // loopback
  case 4: set_handset_mic(volume);break; // mic selection
  case 5: set_headset_mic(volume);break;
  case 6: set_speaker_phone(volume);break;//handsfree mic
  case 7: set_external_io(volume);break;//external audio input
  default:
  }
}

unsigned int reverse_volume(unsigned int volume)
{
  unsigned int ac97_value;

  if (volume < 0) volume = 1;
  if (volume > 15) volume = 15;
 
  ac97_value = 15 - volume;

  return ac97_value; 
}

unsigned short adjust_adc_gain(unsigned int volume)
{
  unsigned short ac97_value;
  switch(volume){
  case 1: ac97_value = 5-4;break; 
  case 2: ac97_value = 6-4;break; //default 
  case 3: ac97_value = 7-4;break;
  case 4: ac97_value = 8-4;break; 
  default:ac97_value = 6-4;break;
  }
  return ac97_value;
}

// set volume (0..15)
void set_handset(unsigned int volume)
{
  handset_volume = volume;
  ac97_handset &= ~0x8000; 
  if(!volume) {
    ac97_handset |= 0x8000;//speaker  mute bit set
    ac97_dac |= 0x8000;
  }
  else {
    ac97_handset = reverse_volume(volume) | (reverse_volume(volume)) << 8;
    ac97_dac = HANDSET_DAC_GAIN;       
  }
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x02,ac97_handset); // mute on/off
}

void hardware_platform(void)
{
#ifdef CONFIG_SNOM105
  ac97_adc_handset = 0x5;
#else
  if(global_serial_num >= SNOM200_B){
    ac97_adc_handset = 0x5; //4
    ac97_adc_sphone = 0xe; //5
    ac97_dac_sphone = DAC_GAIN_SPHONE;
  }
  else{
    ac97_adc_handset = 0x3; //1
    ac97_adc_sphone = 0xb; //6
    ac97_dac_sphone = DAC_GAIN_SPHONE_A;
  }
  ac97_adc = ac97_adc_sphone;
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
#endif //CONFIG_SNOM105
}

#ifdef CONFIG_SNOM105
void set_ringer(unsigned int volume)
{
  ac97_speaker &= ~0x8000;
  ac97_dac &= ~0x8000;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;
    set_ac97_reg(0x18,ac97_dac);// dac preamp closed
    set_ac97_reg(0x06, ac97_speaker); //mute on sent
    ac97_mic &= ~0x8000; //sidetone enabled
  }
  else {  
    ac97_dac = DAC_RINGER;
    set_ac97_reg(0x18,ac97_dac);// new value for the ringer
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);    
    set_ac97_reg(0x06,ac97_speaker); //mute off, new value 
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is on
 }
  set_ac97_reg(0xe, ac97_mic);
}

void set_handset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set     
  }
  else {
    //if (volume < 0 || volume > 4) return;
    // ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= ac97_adc_handset; 
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic selected
    set_port(PORT_SNOM_CSR, 0, MIC_CONTROL);//handset on
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
}

void set_headset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    set_port(PORT_SNOM_CSR, 0,MIC_CONTROL);//handset 
  }
  else {
  if (volume < 0 || volume > 4) return;
   //ac97_adc |= adjust_adc_gain(volume);
   ac97_adc |= 5; // cs: lowest possible amplification for wide linear area
   set_port(PORT_SNOM_CSR, MIC_CONTROL, 0);// Headset is on     
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) dis/en
}

void set_external_io(unsigned int volume)
{}// only for snom200, snom220 
void set_speaker_phone(unsigned int volume)
{
  ac97_speaker &= ~0x8000;
  ac97_dac &= ~0x8000;
  half_duplex = volume > 0;
  ac97_adc &= ~0x800f;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;
    set_ac97_reg(0x18,ac97_dac);// dac preamp closed
    set_ac97_reg(0x06, ac97_speaker); //mute on sent
    ac97_mic &= ~0x8000; //sidetone enabled
    ac97_adc |= 0x8000;
  }
  else {  
    ac97_dac = DAC_RINGER;
    set_ac97_reg(0x18,ac97_dac);// new value for the ringer
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);    
    set_ac97_reg(0x06,ac97_speaker); //mute off, new value 
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is on
    ac97_adc |= 0x6;
  }
  set_ac97_reg(0xe, ac97_mic);
  set_ac97_reg(0x1c, ac97_adc);
}
/*
{
  ac97_speaker &= ~0x8000;
  half_duplex = volume > 0;
  ac97_dac &= ~0x8000;
  ac97_adc &= ~0x800f;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;
    set_ac97_reg(0x18,ac97_dac);// dac preamp closed
    set_ac97_reg(0x06, ac97_speaker); //mute on sent
    ac97_mic &= ~0x8000; //sidetone enabled
    ac97_adc |= 0x8000; // adc (mic) mute bit set
  }
  else {  
    ac97_dac = DAC_GAIN_SPHONE;
    set_ac97_reg(0x18,ac97_dac);// new value for the ringer
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);    
    set_ac97_reg(0x06,ac97_speaker); //mute off, new value 
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is on
    ac97_adc |= 0x6;
 }
  set_ac97_reg(0xe, ac97_mic);
  set_ac97_reg(0x1c, ac97_adc);
}
*/


void set_headset(unsigned int volume)
{
  ac97_headset &= ~0x8000; 
  if(!volume) {
    ac97_headset |= 0x8000;//speaker  mute bit set
    ac97_dac |= 0x8000;//preamp dac mute bit set
  }
  else {
    unsigned int vol_left, vol_right;
    volume = reverse_volume(volume);
    vol_left  = volume;
    vol_right = volume;
    volume = (vol_left << 8) | vol_right;
    ac97_headset = volume;    
    ac97_dac = HANDSET_DAC_GAIN;       
  }    
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x04,ac97_headset); // mute on/off
}
#else //ifdef snom105
void set_ringer(unsigned int volume)
{
  ac97_speaker &= ~0x8000;
  ac97_dac &= ~0x8000;
  if(!volume) {
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;// mute on preamp
    ac97_mic &= ~0x8000; //sidetone enabled
    set_ac97_reg(0x06, ac97_speaker);// mute on
    set_ac97_reg(0x18,ac97_dac);// sent
  }
  else {  
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);
    ac97_dac = DAC_RINGER;
    set_ac97_reg(0x18,ac97_dac);// sent
    set_ac97_reg(0x06, ac97_speaker); //speaker opened
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is open
  }
  set_ac97_reg(0xe, ac97_mic);//side tone on/off sent
}

void set_handset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set     
  }
  else {
    //if (volume < 0 || volume > 4) return;
    // ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= ac97_adc_handset; //see hardware plaform for the value
    //ac97_adc |= 0xe;
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic selected
    set_port(PORT_SNOM_CSR, 0, MIC_CONTROL);//handset on
    //ac97_mic &= ~MIC_BOOST; //mic boost disabled
    //set_ac97_reg(0xe, ac97_mic); 
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
}
void set_headset_mic(unsigned int volume)
{
  ac97_adc &= ~0x800f;
  ac97_headset_mic &= ~0x8000;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    ac97_gp &= ~HEADSET_MIC; //headset_mic dis
    set_port(PORT_B, 0,HEADSET_LED);//disconnected 
  }
  else {
    if (volume < 0 || volume > 4) return;
    //ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= 0x9;
    set_port(PORT_B, HEADSET_LED, 0);// connected
    ac97_gp |= HEADSET_MIC;
    ac97_headset_mic |= HEADSET_RJ14;
  }
  set_ac97_reg(0x20, ac97_gp);
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) dis/en
  set_ac97_reg(0x26,ac97_headset_mic); 
}

void set_external_io(unsigned int volume)
{
  unsigned int vol_left, vol_right;
  if(global_serial_num <  SNOM200_B) return; // doesn't work for snom200a
  ac97_headset &= ~0x8000;
  ac97_adc &= ~0x800f;
  ac97_dac &= ~0x8000;
  ac97_headset_mic &= ~0x8000;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    ac97_gp &= ~HEADSET_MIC; //headset_mic dis
    ac97_headset |= 0x8000;//speaker  mute bit set
    ac97_dac |= 0x8000;
  }
  else {
    // if (volume < 0 || volume > 4) return;
    //ac97_adc |= adjust_adc_gain(volume);
    ac97_adc |= 0x9;
    ac97_gp |= HEADSET_MIC;
    ac97_headset_mic |= HEADSET_PC;
    volume = reverse_volume(volume);
    vol_left  = volume;
    vol_right = volume;
    volume = (vol_left << 8) | vol_right;
    ac97_headset = volume;    
    ac97_dac = HEADSET_DAC_GAIN;
  }
  set_ac97_reg(0x20, ac97_gp);
  set_ac97_reg(0x26,ac97_headset_mic); //external audio from pc audio connector
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) dis/en
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x04,ac97_headset); // mute on/off
}

void set_speaker_phone(unsigned int volume)
{
  sp_vol = volume;
  half_duplex = volume > 0;
  ac97_adc &= ~0x800f;
  ac97_speaker &= ~0x8000;
  ac97_dac &= ~0x8000;
  if(!volume){
    ac97_adc |= 0x8000; // adc (mic) mute bit set
    set_port(PORT_C, SHUTDOWN_SPEAKER_PIN,0);//power amp is off
    ac97_speaker |= 0x8000; //mute on
    ac97_dac |= 0x8000;// mute on
    ac97_mic &= ~0x8000; //sidetone enabled
    set_ac97_reg(0x06, ac97_speaker);// mute on
  }
  else {
    ac97_mic |= 0x8000;//side tone disabled 
    ac97_speaker = reverse_volume(volume);     
    set_ac97_reg(0x06, ac97_speaker); //speaker opened
    ac97_adc |= ac97_adc_sphone;
    ac97_dac = ac97_dac_sphone;
    ac97_gp &= ~HEADSET_MIC;
    set_ac97_reg(0x20, ac97_gp); //handset mic/hfmic  selected
    set_port(PORT_C,0, SHUTDOWN_SPEAKER_PIN);// power amp is open
    //set_port(PORT_SNOM_CSR, MIC_CONTROL,0);//handsfree mic on
  }
  set_ac97_reg(0x1c, ac97_adc); // adc (mic) en (+value)/dis
  set_ac97_reg(0xe, ac97_mic);//side tone level sent
  set_ac97_reg(0x18,ac97_dac);// new value for dac gain
}

void set_headset(unsigned int volume)
{
  if(global_serial_num <  SNOM200_B) set_handset(volume);
   ac97_headset &= ~0x8000;
   ac97_dac &= ~0x8000;
  if(!volume) {
    ac97_headset |= 0x8000;//speaker  mute bit set
    ac97_dac |= 0x8000;
  }
  else {
    unsigned int vol_left, vol_right;
    volume = reverse_volume(volume);
    vol_left  = volume;
    vol_right = volume;
    volume = (vol_left << 8) | vol_right;
    ac97_headset = volume;    
    ac97_dac = HEADSET_DAC_GAIN;       
  }    
  set_ac97_reg(0x18,ac97_dac);
  set_ac97_reg(0x04,ac97_headset); // mute on/off
}
#endif //ifdef snom105


void set_loopback(unsigned int volume) // volume means device in that case
{
  set_handset_mic (4);

  switch(volume){
    case 1: set_handset (15); break; // handset
    case 2: set_ringer (10); break; // casing speaker
    case 3: set_headset (15); break; // headset
    default: {
      set_handset_mic (0);
      set_handset (0);
      set_ringer (0);
      set_headset (0);
    }
  }

  if(!volume) 
    ac97_gp &=~LPBK;       
  else  
    ac97_gp |= LPBK;    
  set_ac97_reg(0x20, ac97_gp);
}

void snom_aud_set_amp_gain(int level)
{}

void snom_aud_set_tx_gain(int level)
{}

void snom_aud_set_tx_gain2(int level)
{}

void snom_aud_set_rx_gain(int level)
{}

void snom_aud_shutdown_speaker(int bTrue)
{}

void snom_aud_shutdown(int bTrue)
{}

void snom_aud_set_mixer_mode(int mode) {
  unsigned long flags;
	
  save_flags(flags); cli();
  mixer_mode = mode;

  // reset the audio buffer (exception: "knocking"):
  au_rd_pos = au_wr_pos;

  restore_flags(flags);
}

/* table for DTMF tones */
/* gawk -v s=23 -v r=2 -v vol=8000 'BEGIN{for(i=0;i<s;i++) printf("%d, ", sin(3.1415927*2*r*i/s)*vol ); exit;}' */
static const short freq_697[]  = { /* 23 samples, 2 reps */
  0, 4156, 7103, 7981, 6535, 3187, -1089, -5048, -7538, -7832, -5846, -2158, 2158, 5846, 7832, 7538, 5048, 1089, -3187, -6535, -7981, -7103, -4156
};

static const short freq_770[]  = { /* 52 samples, 5 reps */
  0, 4544, 7480, 7767, 5304, 964, -3717, -7083, -7941, -5988, -1914, 2836, 6583, 7999, 6583, 2836, -1914, -5988, -7941, -7083, -3717, 964, 5304, 7767, 7480, 4544, -0, -4544, -7480, -7767, -5304, -964, 3717, 7083, 7941, 5988, 1914, -2836, -6583, -7999, -6583, -2836, 1914, 5988, 7941, 7083, 3717, -964, -5304, -7767, -7480, -4544
};

static const short freq_852[]  = { /* 47 samples, 5 reps */
  0, 4957, 7782, 7256, 3608, -1593, -6109, -7995, -6440, -2113, 3123, 7015, 7888, 5366, 534, -4527, -7640, -7465, -4076, 1066, 5750, 7959, 6743, 2624, -2624, -6743, -7959, -5750, -1066, 4076, 7465, 7640, 4527, -534, -5366, -7888, -7015, -3123, 2113, 6440, 7995, 6109, 1593, -3608, -7256, -7782, -4957
};

static const short freq_941[]  = { /* 17 samples, 2 reps */
  0, 5389, 7965, 6384, 1469, -4211, -7694, -7161, -2889, 2889, 7161, 7694, 4211, -1469, -6384, -7965, -5389
};

static const short freq_1209[] = { /* 53 samples, 8 reps */
  0, 6499, 7578, 2336, -4854, -7996, -4469, 2785, 7717, 6212, -473, -6764, -7413, -1879, 5222, 7968, 4068, -3224, -7828, -5902, 946, 7005, 7222, 1415, -5572, -7912, -3653, 3653, 7912, 5572, -1415, -7222, -7005, -946, 5902, 7828, 3224, -4068, -7968, -5222, 1879, 7413, 6764, 473, -6212, -7717, -2785, 4469, 7996, 4854, -2336, -7578, -6499
};

static const short freq_1336[] = { /* 6 samples, 1 rep */
  0, 6928, 6928, -0, -6928, -6928 
};

static const short freq_1477[] = { /* 65 samples, 12 rep */
  0, 7334, 5858, -2655, -7978, -3717, 5009, 7719, 1155, -6795, -6583, 1537, 7811, 4702, -4055, -7941, -2287, 6114, 7171, -386, -7480, -5588, 3016, 7997, 3371, -5304, -7608, -772, 6991, 6356, -1914, -7885, -4384, 4384, 7885, 1914, -6356, -6991, 772, 7608, 5304, -3371, -7997, -3016, 5588, 7480, 386, -7171, -6114, 2287, 7941, 4055, -4702, -7811, -1537, 6583, 6795, -1155, -7719, -5009, 3717, 7978, 2655, -5858, -7334  
};

static const short *dtmf_freq[16][2] = {
  { freq_941, freq_1336 }, // 0
  { freq_697, freq_1209 }, // 1
  { freq_697, freq_1336 }, // 2
  { freq_697, freq_1477 }, // 3
  { freq_770, freq_1209 }, // 4
  { freq_770, freq_1336 }, // 5
  { freq_770, freq_1477 }, // 6
  { freq_852, freq_1209 }, // 7
  { freq_852, freq_1336 }, // 8
  { freq_852, freq_1477 }, // 9
  { freq_941, freq_1209 }, // #
  { freq_941, freq_1477 }, // *
  { freq_941, freq_941 }, // fixme
  { freq_941, freq_941 }, // fixme
  { freq_941, freq_941 }, // fixme
  { freq_941, freq_941 }, // fixme
};

static const int dtmf_len[16][2] = {
  { sizeof(freq_941) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 0
  { sizeof(freq_697) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // 1
  { sizeof(freq_697) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 2
  { sizeof(freq_697) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // 3
  { sizeof(freq_770) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // 4
  { sizeof(freq_770) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 5
  { sizeof(freq_770) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // 6
  { sizeof(freq_852) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // 7
  { sizeof(freq_852) / sizeof(short), sizeof(freq_1336) / sizeof(short) }, // 8
  { sizeof(freq_852) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // 9
  { sizeof(freq_941) / sizeof(short), sizeof(freq_1209) / sizeof(short) }, // #
  { sizeof(freq_941) / sizeof(short), sizeof(freq_1477) / sizeof(short) }, // *
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
  { sizeof(freq_941) / sizeof(short), sizeof(freq_941)  / sizeof(short) }, // fixme
};

struct dtmf_state_t internal_dtmf; // outgoing
struct dtmf_state_t external_dtmf[MAX_NUM_RTP]; // incoming from RTP
int play_dtmf = 0; // bool for enabling internal dtmf

#define DTMF_STATE_NO_DTMF 0
#define DTMF_STATE_SEND_DTMF 1
#define DTMF_STATE_SEND_LAST 2

static int dtmf_key[MAX_NUM_RTP] = { 0 };
static int dtmf_state[MAX_NUM_RTP] = { DTMF_STATE_NO_DTMF };
static int dtmf_shutoff[MAX_NUM_RTP] = { 0 }; // when we are playing dtmf of the other side
// static int dtmf_in_duration = 0; // duration of current incoming dtmf sound
static int dtmf_recv_type[MAX_NUM_RTP] = { 101 }; // har_wired?
static unsigned long dtmf_recv_ssrc_id[MAX_NUM_RTP];
static int dtmf_recv_ssrc_valid[MAX_NUM_RTP] = { 0 };
static int dtmf_duration[MAX_NUM_RTP] = { 0 }; // duration of the outgoing dtmf so far
static int dtmf_recv_pos[MAX_NUM_RTP] = { 0 }; // last received position

// set dtmf key: bits:
// 0..3: key
// 4: key valid
// 5: play sound (on local speaker)
// 6: send dtmf rtp (rfc 2833)
// 7: send tone inband (on rtp packets)
// 8..15: payload type
// 15..17: channel
void snom_aud_set_dtmf(int key) {
  unsigned long flags;

  int channel = key >> 16;
  if (channel < 0 || channel >= MAX_NUM_RTP) return;
  key &= 0xffff; 

  // make sure we dont send both in- and out-of-band:
  if (key & 0x40) {
    key &= ~0x80;
  }

  save_flags(flags); cli();

  switch (dtmf_state[channel]) {
  case DTMF_STATE_NO_DTMF:
    if (key & 16) {
      dtmf_state[channel] = DTMF_STATE_SEND_DTMF;
      dtmf_duration[channel] = 0;
      need_marker[channel] = 1;
    }
    dtmf_key[channel] = key;
    break;
  case DTMF_STATE_SEND_DTMF:
    if (!(key & 16)) {
      dtmf_state[channel] = DTMF_STATE_SEND_LAST;
    }
    else
      // set the new key: fixme: if we did not send the last key
      // termination code, its probably too early to send the next
      // one. However, what can we do?
      dtmf_key[channel] = key;
    break;
  case DTMF_STATE_SEND_LAST:
    // fixme: if we did not send the last key termination code, its
    // probably too early to send the next one. However, what can we
    // do?
    if (key & 16) {
      dtmf_duration[channel] = 0;
      dtmf_state[channel] = DTMF_STATE_SEND_DTMF;
    }
    dtmf_key[channel] = key;
    break;
  }

  if (key & 32) {
    set_dtmf(dtmf_key[channel] & 15, &internal_dtmf);
    play_dtmf = 1;
  }
  else {
    play_dtmf = 0;
  }

  restore_flags(flags);
}

static void clear_buffers(int restart) {
  int j;
  volatile cbd_t  *bdp;
  // printk("c");
  bdp = phy.tx_cur = phy.tx_bd_base;

  for (j = 0; j < HW_NUM_BLOCKS - 1; j++) {
    bdp->cbd_sc = BD_SC_READY; // ints only on rx side
    bdp->cbd_datlen = CPM_BUF_SIZE;
    write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 
    bdp++;
  }
  bdp->cbd_sc = BD_SC_READY | BD_SC_WRAP; // ints only on rx side
  bdp->cbd_datlen = CPM_BUF_SIZE;
  write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 
	
  bdp = phy.rx_cur = phy.rx_bd_base;
	
  for (j=0; j<(HW_NUM_BLOCKS-1); j++) {
    bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_INTRPT);
    bdp->cbd_datlen = CPM_BUF_SIZE;
    write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 
    bdp++;
  }
  bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT);
  bdp->cbd_datlen = CPM_BUF_SIZE;
  write_to_hardware(__va(bdp->cbd_bufaddr), NULL); 

#if 0
  /* mark all frames as clean */
  for(j = 0; j < SW_NUM_BLOCKS; j++) {
    // tx_buffer.avail[j] = 0;
    // rx_buffer.avail[j] = 0;
  }
#endif

  if (restart) {
    restart_cpm(CPM_RXTX);
    
    /* init automatic gain control */
    init_agc();
  }
}

void snom_aud_set_tos(int t) {
  tos = t;
}

// set the destination for a socket:
int snom_aud_set_destination(struct socket *rtp_socket, struct in_addr *addr) {
  struct rtable *rt = 0;
  if(ip_route_output(&rt, addr->s_addr, rtp_socket->sk->saddr,
		     RT_TOS(rtp_socket->sk->protinfo.af_inet.tos) | RTO_CONN | rtp_socket->sk->localroute,
		     rtp_socket->sk->bound_dev_if)) return 0;
  __sk_dst_set(rtp_socket->sk, &rt->u.dst);
  return 1;
}

int snom_aud_set_default_destination(int c) {
  return snom_aud_set_destination(rtp_socket[c], &remote_adr[c].sin_addr);
}

// start the ice process:
int snom_aud_set_ice_parameters(struct snom_aud_ice_parameter_s *parm) {
  int i;
  const int c = parm->channel;
  const int l = parm->length;
  if(c < 0 || c >= MAX_NUM_RTP) return -EINVAL;
  if(l < 0 || l >= MAX_NUM_ICE) return -EINVAL;
  ice_contacts[c].length = 0; // avoid race conditions
  for(i = 0; i < l; i++) {
    if(parm->addr[i] == 0) return -EINVAL;
    ice_contacts[c].contacts[i].dest_adr = parm->addr[i];
    ice_contacts[c].contacts[i].dest_port = parm->port[i];
    memcpy(ice_contacts[c].contacts[i].id, &parm->id[i * 16], 16);
  }

  ice_contacts[c].accept = 1;
  ice_contacts[c].length = l;
  return 0;
}

// set the parameters of a specific audio rtp channel:
int snom_aud_set_rtp_parameters2(int c, struct snom_aud_rtp_parameter_s *parm) {
  int rc;
  struct sockaddr_in sock;
  unsigned long flags;
  save_flags(flags); cli();

  /* set the type & length: */
#if 0
  printk("rtp:");
  printk("%d ", c);
  printk("%08x", parm->raddr);
  printk(":%d", parm->remote_port);
  printk(" %d\n", parm->local_port);
#endif
  check_c("snom_aud_set_rtp_parameters2", c);
  packet_type[c] = parm->type;

  check_for_agc();
     {
	/* ulaw or alaw */
	if      (parm->length <= 80)  packet_size[c] = 80;
	else if (parm->length <= 160) packet_size[c] = 160;
	else if (parm->length <= 240) packet_size[c] = 240;
	else                          packet_size[c] = 320;
      }

  // printk("rtp: a=%08x p=%d t=%d l=%d\n", (int)parm->raddr, (int)parm->remote_port, (int)packet_type, (int)packet_size);

  /* set the local ssrc */
  local_ssrc[c] = parm->ssrc;

  /* A change in the local socket? */
  if (local_port[c] != parm->local_port) {
    /* close the local port */
    if (rtp_socket[c] != 0) {
      sock_release(rtp_socket[c]);
      rtp_socket[c] = 0;
    }
    if (rtcp_socket[c] != 0) {
      sock_release(rtcp_socket[c]);
      rtcp_socket[c] = 0;
    }
    local_port[c] = 0;
    play_dtmf = 0;
  
    /* open a new port? */
    if (parm->local_port > 0) {
      jit_recv_reset(c);
      jit_decode_start[c] = 1;
  
      /* create a socket: */
      // printk("rtp: open socket %d\n", parm->local_port);
      rc = sock_create(PF_INET, SOCK_DGRAM, 0, &rtp_socket[c]);
      if (rc < 0) {
	rtp_socket[c] = 0; /* this socket cannot be used! */
	// printk("snom_aud_set_rtp_parameters: Could not create socket\n");
	goto ret_release;
      }
      
      rc = sock_create(PF_INET, SOCK_DGRAM, 0, &rtcp_socket[c]);
      if (rc < 0) {
	rtcp_socket[c] = 0; /* this socket cannot be used! */
	goto ret_release;
      }
      
      /* set up the local port: */
      memset(&sock, 0, sizeof(sock));
      sock.sin_family = AF_INET;
      sock.sin_addr.s_addr = htonl(INADDR_ANY);
      sock.sin_port = htons(parm->local_port);
      rc = rtp_socket[c]->ops->bind(rtp_socket[c], (struct sockaddr *)&sock, sizeof(sock));
      if (rc != 0) {
	// printk("snom_aud_set_rtp_parameters: Could not bind\n");
	goto ret_release;
      }
      
      memset(&sock, 0, sizeof(sock));
      sock.sin_family = AF_INET;
      sock.sin_addr.s_addr = htonl(INADDR_ANY);
      sock.sin_port = htons(parm->local_port + 1);
      rc = rtcp_socket[c]->ops->bind(rtcp_socket[c], (struct sockaddr *)&sock, sizeof(sock));
      if (rc != 0) {
	// printk("snom_aud_set_rtp_parameters: Could not bind\n");
	goto ret_release;
      }
      
      /* set the socket type of service: IP_TOS = 5 */
      rtp_socket[c]->sk->protinfo.af_inet.tos = tos;
      // rtp_socket[c]->sk->priority = TC_PRIO_INTERACTIVE;
      rtp_socket[c]->sk->priority = 0; // VLAN -> does it help?
      rtp_socket[c]->sk->protinfo.af_inet.pmtudisc = IP_PMTUDISC_DONT; // always increase the identity
      
      // clear_buffers(1);
      local_port[c] = parm->local_port;
    } /* local socket required */
  } /* change in local socket */

  /* prepare the destination */
  if (parm->transmit &&
      (remote_adr[c].sin_addr.s_addr != parm->raddr || remote_adr[c].sin_port != parm->remote_port) &&
      rtp_socket[c] != 0 && rtcp_socket[c] != 0) {
    need_marker[c] = 2;
    dtmf_key[c] = 0;
    dtmf_state[c] = DTMF_STATE_NO_DTMF;
    remote_adr[c].sin_addr.s_addr = parm->raddr;
    remote_adr[c].sin_port = parm->remote_port;
    remote_adr[c].sin_family = AF_INET;
    sym_rtp[c] = parm->sym_rtp;
    jit_unencoded_len[c] = 0;
    jit_encoded_len[c] = 0; // pretend there is nothing encoded
    
    if (remote_adr[c].sin_addr.s_addr != 0) {
      if(!snom_aud_set_destination(rtp_socket[c], &remote_adr[c].sin_addr)) {
 	// printk("rtp: ip_route_output failed!\n");
 	goto ret_release;
      }
      
      if(!snom_aud_set_destination(rtcp_socket[c], &remote_adr[c].sin_addr)) {
 	// printk("rtcp: ip_route_output failed!\n");	
 	goto ret_release;
      }
    } // remote_adr available
  }

  restore_flags(flags);
  return 0;

 ret_release:
  if (rtp_socket[c] != 0) {
    sock_release(rtp_socket[c]);
  }
  if (rtcp_socket[c] != 0) {
    sock_release(rtcp_socket[c]);
  }
  rtp_socket[c] = 0;
  rtcp_socket[c] = 0;
  clear_buffers(0);
  restore_flags(flags);
  return rc;
}

// this function is kept for backward compatibility. It sets the
// parameters for the RTP channel 0 and disables all other channels.
int snom_aud_set_rtp_parameters(struct snom_aud_rtp_parameter_s *parm) {
  return snom_aud_set_rtp_parameters2(0, parm);
}

// start or stop the recording:
int snom_aud_start_recording(int on) {
  unsigned long flags;
  save_flags(flags); cli();

  if(on) {
    read_wpos = 0;
    read_rpos = 0;
  }
  read_recording = on;
  restore_flags(flags);
  return 0;
}

/******************************************************************************/

/*
 * Call back mechanism
 */

void snom_aud_set_call_back(void (*cb)(void))
{
  unsigned long flags;
	
  save_flags(flags); cli();
	
  snom_aud_call_back = cb;
	
  restore_flags(flags);
}

static _INLINE_ void shutdown_cpm(void)
{	
  volatile cpm8xx_t *cp = cpmp;
  volatile scc_t    *sp = &cp->cp_scc[0];
  int i;

  /* Disable reception and transmission. */
	
  sp->scc_gsmrl &= ~(SCC_GSMRL_ENT | SCC_GSMRL_ENR); 
	
  /* Enable all interrupts and clear all pending events. */
	
  sp->scc_sccm |= SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX; // SCCM_GLR | SCCM_GLT
  sp->scc_scce = 0xffff;
	
  // printk("<sd>");
  clear_buffers(0);

  local_time = 0;
  for(i = 0; i < MAX_NUM_RTP; i++) jit_recv_reset(i);

  phy.is_underrun = 1;
}

void snom_aud_reset(snom_aud_info_t *info) {
  unsigned long flags;
  int i;
  save_flags(flags); cli();
  
  for(i = 0; i < MAX_NUM_RTP; i++) {
    if (rtp_socket[i] != 0)
      sock_release(rtp_socket[i]);
    if (rtcp_socket[i] != 0)
      sock_release(rtcp_socket[i]);
    rtp_socket[i] = 0;
    rtcp_socket[i] = 0;
    local_port[i] = 0;
  }
 clear_buffers(0);

  restore_flags(flags);
}

/*
 * real time rtp part
 */

void snom_aud_sendmsg(struct socket *skt, char *data, int len, int dport) {
  struct sock *sk = skt->sk;
  struct sk_buff *skb;
  struct udphdr *uh;
  unsigned char *msg;
  int rc;

  /* fixme: get hand on device and read of physical header size */
  skb = alloc_skb(len + sizeof(struct udphdr) + sizeof(struct iphdr)
		  + 16 /* ethernet header size */ + 15, GFP_ATOMIC);
  if (skb == NULL) {
    // printk("rtp: no skb\n");
    return;
  }

  skb_reserve(skb, (16 /* ethernet header size */ + sizeof(struct udphdr) + sizeof(struct iphdr) + 15) & ~15);
  skb->sk  = sk;
  
  /* put the data on the packet */
  msg = (unsigned char *) skb_put(skb, len);
  memcpy(msg, data, len);

  /* add udp header */
  uh = (struct udphdr *) skb_push(skb, sizeof(struct udphdr));
  uh->source = sk->sport;
  uh->dest = dport;
  uh->len = len + 8; // length with! udp header
  uh->check = 0; // no checksum
  skb->h.uh = uh;

  // put it on the cable:
  rc = ip_queue_xmit(skb);
  if (rc != 0) {
    // printk("Failed to send msg: %d\n", rc);
  }
  return;
}

/* check if something is available for sending: */
static _INLINE_ void snom_aud_check_outgoing(void) {
  unsigned long flags;
  int c;

  /* make sure we have a socket: */
  save_flags(flags); cli();

  // check if we can encode something:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (jit_rtp_active(c)) {
      const int type = packet_type[c];
      int *length = &jit_send_len[c];
      int *el = &jit_encoded_len[c];
      int *ul = &jit_unencoded_len[c];
      unsigned char *buffer = jit_encoded_buffer[c];

	if (*length >= 80) {
	  // all other codecs work on 80 samples:
	  short *src = jit_send_buffer[c];
	  int l = *length;
	  unsigned char *dst = &buffer[*el + 12];
	  int maxlen = sizeof(jit_encoded_buffer[0]) - (*el + 12);
	  int len;
	  if (l < 160) {
	    // we dont have too much in the buffer (running out of CPU)
	    restore_flags(flags);
	    len = snom_aud_encode(c, type, src, l, dst, maxlen);
	    save_flags(flags); cli();
	    *el += len;
	    if (*ul == 0) jit_encoded_ts[c] = local_time;
	    *ul += *length;
	  }
	  *length = 0;
	}
    }
  } // for c
  restore_flags(flags);

  // now check if we can send something:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (jit_rtp_active(c) && jit_unencoded_len[c] >= packet_size[c] && jit_encoded_len[c] > 0) {
      if (dtmf_state[c] == DTMF_STATE_NO_DTMF || (dtmf_key[c] & 0x80) != 0) {
	unsigned char *buffer = jit_encoded_buffer[c];
	unsigned long ssrc = local_ssrc[c];
	int ts = jit_encoded_ts[c];
	static int first_silence = 1;
	static int silence = 0;
	unsigned char vol;

	vol = (unsigned char)(energy_lf/1024);

	buffer[0] = 0x80;
	if(silence_suppression){
	  if(global_serial_num >= SNOM200_B){
	    silence = energy_lf < 10000 ? 1 : 0;
	  }else{
	    silence = energy_lf < 10000 ? 1 : 0;
	  }
	}
	if(silence){
	  if(first_silence){
	    buffer[1] = 13;
	    buffer[12] = vol & 0x7F;
	    //printk("energy_lf = %d\n",energy_lf);
	    //printk("vol       = %d\n",vol);
	  }
	}else{
	  buffer[1] = packet_type[c];
	}
	if (need_marker[c] > 0) {
	  need_marker[c]--;
	  if (need_marker[c] > 0) {
	    // clear the output buffer:
	    jit_unencoded_len[c] = 0;
	    jit_encoded_len[c] = 0;
	    return; // skip first packet so that the jitter is minimized
	  }
	  buffer[1] |= 0x80;
	}
  
	/* sequence number: */
	buffer[2] = (unsigned char)(packet_seqnr[c] >> 8);
	buffer[3] = (unsigned char)(packet_seqnr[c] & 0xff);
    
	/* timestamp */
	buffer[4] = (unsigned char)((ts >> 24) & 0xff);
	buffer[5] = (unsigned char)((ts >> 16) & 0xff);
	buffer[6] = (unsigned char)((ts >> 8) & 0xff);
	buffer[7] = (unsigned char)(ts & 0xff);

	/* ssrc */
	buffer[8]  = (unsigned char)((ssrc >> 24) & 0xff);
	buffer[9]  = (unsigned char)((ssrc >> 16) & 0xff);
	buffer[10] = (unsigned char)((ssrc >> 8) & 0xff);
	buffer[11] = (unsigned char)(ssrc & 0xff);

	// printk("t=%d\n", jit_encoded_len[c]);
	// if (c == 0) printk("%d", packet_type[c]);
	if(silence){
	  if(first_silence){
	    first_silence = 0;
	    stat_tx[c]++;
	    snom_aud_sendmsg(rtp_socket[c], buffer, 13, remote_adr[c].sin_port);
	  }
	}else{
	  first_silence = 1;
	  stat_tx[c]++;
	  snom_aud_sendmsg(rtp_socket[c], buffer, jit_encoded_len[c] + 12, remote_adr[c].sin_port);
	}
	packet_seqnr[c]++;
	// if (c > 0) printk("q");
	jit_unencoded_len[c] = 0;
	jit_encoded_len[c] = 0;
      } // if no dtmf or dtmf inband

      /* RTP out of band signalling: */
      if (dtmf_state[c] != DTMF_STATE_NO_DTMF && (dtmf_key[c] & 0x40) != 0) {
	unsigned char *buffer = jit_encoded_buffer[c];
	unsigned long ssrc = local_ssrc[c];
	int ts = jit_encoded_ts[c];

	buffer[0] = 0x80; // RTP
	buffer[1] = (unsigned char)((dtmf_key[c] >> 8) & 0x7f); // media-type
	if (need_marker[c]) {
	  buffer[1] |= 0x80;
	  need_marker[c] = 0;
	}

	/* sequence number: */
	buffer[2] = (unsigned char)(packet_seqnr[c] >> 8);
	buffer[3] = (unsigned char)(packet_seqnr[c] & 0xff);
  
	/* timestamp */
	buffer[4] = (unsigned char)((ts >> 24) & 0xff);
	buffer[5] = (unsigned char)((ts >> 16) & 0xff);
	buffer[6] = (unsigned char)((ts >> 8) & 0xff);
	buffer[7] = (unsigned char)(ts & 0xff);
    
	/* ssrc */
	buffer[8]  = (unsigned char)((ssrc >> 24) & 0xff);
	buffer[9]  = (unsigned char)((ssrc >> 16) & 0xff);
	buffer[10] = (unsigned char)((ssrc >> 8) & 0xff);
	buffer[11] = (unsigned char)(ssrc & 0xff);

	buffer[12] = (unsigned char)(dtmf_key[c] & 15);
	if (dtmf_state[c] == DTMF_STATE_SEND_LAST) {
	  buffer[13] = 0x8a; // volume, 0x80 = end of digit
	  dtmf_state[c] = DTMF_STATE_NO_DTMF;
	  dtmf_key[c] = 0; // reset the key
	}
	else 
	  buffer[13] = 0x0a; // volume
	dtmf_duration[c] += packet_size[c];
	buffer[14] = (unsigned char)((dtmf_duration[c] >> 8) & 0xff); // duration
	buffer[15] = (unsigned char)(dtmf_duration[c] & 0xff);
	stat_tx[c]++;
	snom_aud_sendmsg(rtp_socket[c], buffer, 16, remote_adr[c].sin_port);
	packet_seqnr[c]++;

	// discard the encoded buffer:
	jit_unencoded_len[c] = 0;
	jit_encoded_len[c] = 0;
      } // out-of-band
    } // packet available
  } // for c
}

/* check if the data contains a rtp packet and process it: */
#if 0
static _INLINE_ void snom_aud_process_rtp(unsigned char *data, int length, unsigned int remote_time) {
  static short decode_buffer[640];
  int plen; /* payload length, audio length */
  int start_pos, i;
  unsigned long flags;

  if (data != 0) {
    /* determine the length of the payload: */
    plen = (data[0] & 0xf) * 4 + 12;
    if (data[0] & 0x10) plen += ((int)(data[plen + 2]) << 8) | (int)(data[plen + 3]);
    plen = length - plen;
    if (data[0] & 0x20) plen -= data[length - 1]; /* padding */
    if (plen <= 0 || plen > sizeof(decode_buffer) / sizeof(decode_buffer[0])) return;
    
    /* determine the type and decode the payload: */
    remote_packet_type = data[1] & 0x7f;
    remote_packet_size = snom_aud_decode(remote_packet_type, &data[length - plen], plen, decode_buffer, sizeof(decode_buffer) / sizeof(decode_buffer[0]));
    if (remote_packet_size == 0) return; /* unknown decode type */
  }
  else {
    /* fill the gap: */
    snom_aud_fill(remote_packet_type, remote_packet_size, decode_buffer, sizeof(decode_buffer) / sizeof(decode_buffer[0]));
  }
  save_flags(flags);
  cli();

  /* copy the data: */
  start_pos = remote_time % (sizeof(rx_buffer.buf) / sizeof(rx_buffer.buf[0]));
  if (start_pos + remote_packet_size > (sizeof(rx_buffer.buf) / sizeof(rx_buffer.buf[0]))) {
    /* split up into two memcpy: */
    const int length = (sizeof(rx_buffer.buf) / sizeof(rx_buffer.buf[0])) - start_pos;
    memcpy(&rx_buffer.buf[start_pos], decode_buffer, length * sizeof(rx_buffer.buf[0]));
    for(i = 0; i < length; i += SNOM_AUD_BLOCK_SIZE)
      rx_buffer.avail[(start_pos + i) / SNOM_AUD_BLOCK_SIZE] = 1;
    memcpy(&rx_buffer.buf[0], &decode_buffer[length], (remote_packet_size - length) * sizeof(rx_buffer.buf[0]));
    for(i = 0; i < remote_packet_size - length; i += SNOM_AUD_BLOCK_SIZE)
      rx_buffer.avail[i / SNOM_AUD_BLOCK_SIZE] = 1;
  }
  else {
    memcpy(&rx_buffer.buf[start_pos], decode_buffer, remote_packet_size * sizeof(rx_buffer.buf[0]));
    for(i = 0; i < remote_packet_size; i += SNOM_AUD_BLOCK_SIZE)
      rx_buffer.avail[(start_pos + i) / SNOM_AUD_BLOCK_SIZE] = 1;
  }
  restore_flags(flags);
}
#endif

/* check if we have incoming packets: */
static _INLINE_ void snom_aud_check_incoming(void) {
  unsigned long flags;
  int c;
  for(c = 0; c < MAX_NUM_RTP; c++) {
    /* rtp handling: */
    if (rtp_socket[c]) {
      /* get all packets that are available: */
      while(1) {
	struct sk_buff *skb;
	unsigned char *adr;
	int len, err;
      
	/* check if there is a packet ready: (code from udp.c:udp_recvmsg()) */
	skb = skb_recv_datagram(rtp_socket[c]->sk, 0, MSG_DONTWAIT, &err);
	if (!skb) break; 
      
	/* process the packet: */
	len = skb->len - sizeof(struct udphdr);

	/* check if we need to fill up buffers */
	// printk("r=%d\n", len);

	// check if we have a STUN packet:
	adr = skb->h.raw + sizeof(struct udphdr);
	if(len > 20 && (adr[0] == 0 || adr[0] == 1) && adr[1] == 1) {
	  int port = (adr[-8] << 8) | adr[-7];
	  snom_on_stun_available(c, ice_contacts[c].contacts, ice_contacts[c].length, 
				 adr, len, &adr[-16], port);
	}
	else {
	  int port = (adr[-8] << 8) | adr[-7];
	  save_flags(flags); cli();
	  jit_recv_rtp(c, adr, len, &adr[-16], port);
	  restore_flags(flags);
	}
	skb_free_datagram(rtp_socket[c]->sk, skb);
      }
    } /* rtp */

    /* rtcp handling: */
    if (rtcp_socket[c]) {
      while(1) {
	struct sk_buff *skb;
	int err;

	/* just eat all packets from the buffer: */
	skb = skb_recv_datagram(rtcp_socket[c]->sk, 0, MSG_DONTWAIT, &err);
	if (!skb) break;
	skb_free_datagram(rtcp_socket[c]->sk, skb);
      } /* while */
    } // rtcp
  }
} /* snom_check_incoming */

#define MAX_NUM_PENDING_AC97_CMDS 128
static int ac97_fifo_rx = 0;
static int ac97_fifo_tx = 0;
static unsigned long ac97_regs[128]; 
static int delay_sending_commands = 100000;

// Update the filtered energy: Quick for high energy, slow for low energy
static long update_energy(long old_e, long new_e) {
  if (new_e > 100000000) 
    return old_e - (old_e / 2) + new_e / 2;
  else if (new_e > 10000000) 
    return old_e - (old_e / 4) + new_e / 4;
  else if (new_e > 1000000) 
    return old_e - (old_e / 8) + new_e / 8;
  else if (new_e > 100000) 
    return old_e - (old_e / 16) + new_e / 16;
  else 
    return old_e - (old_e / 32) + new_e / 32;
}

static struct pending_command {
  unsigned long reg;
  unsigned long value;
} ac97_fifo_buffer[MAX_NUM_PENDING_AC97_CMDS];

// write a block of short samples (16 bit) to the hardware: This
// includes the calculation of the current energy level presented to
// the speaker.
static void write_to_hardware(unsigned char *raw, short *linear) {
  static short last_sf = 0;
  short sf[6];
  short diff1, diff2;
  int i;
  unsigned long *p = (unsigned long*)raw;

  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    int sub;
    register unsigned long int0, int1, int2;
    // short sm = ((i % 20) - 10) * 1000;
    short sm = (linear == NULL ? 0 : ((*linear++) & 0xffff));
    sf[5] = sm;

    diff1 = (sf[5] - last_sf) >> 1;
    diff2 = ((int)(sf[5] - last_sf) * 171) / 1024; // divide by 6
    sf[2] = last_sf + diff1;
    sf[0] = last_sf + diff2;
    sf[1] = sf[0] + diff2;
    sf[3] = sf[2] + diff2;
    sf[4] = sf[3] + diff2;
    last_sf = sf[5]; // remember for next time
    
    for(sub = 0; sub < 6; sub++) {
      unsigned long sample = (unsigned short)(sf[sub]);
      int0 = 0x98000000; // pure audio
      int1 = sample >> 8;
      int2 = (sample << 24) | (sample << 4);

      // Need to fill in a command?
      if (ac97_fifo_rx != ac97_fifo_tx) {
        // writing a sample and a command
        const register unsigned long reg = ac97_fifo_buffer[ac97_fifo_rx].reg & 0xff;
        const register unsigned long val = ac97_fifo_buffer[ac97_fifo_rx].value & 0xffff;
        
        ac97_regs[reg] = val;
        
        int0 |= 0x60000000 | (reg << 8);
        int1 |= val << 12;	
        
        // increment the rx pointer
        ac97_fifo_rx = ((ac97_fifo_rx + 1) % MAX_NUM_PENDING_AC97_CMDS);
      }
      
      // write the three registers to memory:
      p[0] = int0;
      p[1] = int1;
      p[2] = int2;    
      p += 3; // go to the next frame
    } // sub 
  }
}


/*------------------------------------------------------------------------*
 * 2nd order high pass filter with cut off frequency at 140 Hz.           *
 *  Input are divided by two in the filtering process.                    *
 *-----------------------------------------------------------------------*/


/* Static values to be preserved between calls */
/* y[] values is keep in double precision      */

static short x_0 = 0;
static short x_1 = 0;
static long y_1 = 0;
static long y_2 = 0;

#define C_B140_0 (1899)
#define C_B140_1 (-3798)
#define C_B140_2 (1899)

#define C_A140_0 (4096)
#define C_A140_1 (7807)
#define C_A140_2 (-3733)

// read from the hardware and update the current energy level.
static void read_from_hardware(short *linear, unsigned char *raw) {
  int i, sub;
  unsigned long *p = (unsigned long*)raw;
  short x_2;
  long tmp;
  energy_lc = 0;

  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    int result;

    // first step of filter:
    int val = 0;
    int count = 0;
    for(sub = 0; sub < 3; sub++) {
      if ((p[0] & 0x18000000) != 0) {
        val += (short)((p[2] >> 4) & 0xffff);
        count++;
      }
      p += 3; // go to the next frame
    } // sub

    // do a digital filter with cut off at 280 Hz:
    x_2 = x_1;
    x_1 = x_0;
    if(count == 3)
      x_0 = val / 3;
    else if(count == 2)
      x_0 = val / 2;
    else
      x_0 = val;

    /*  y[i] = b[0]*x[i]/2 + b[1]*x[i-1]/2 + b140[2]*x[i-2]/2  */
    /*                     + a[1]*y[i-1] + a[2] * y[i-2];      */

    tmp = (y_1 >> 12) * C_A140_1 + (y_2 >> 12) * C_A140_2 + 
      (x_0 * C_B140_0 + x_1 * C_B140_1 + x_2 * C_B140_2) * 16;

    result = (tmp + 0x80) >> 16;
    y_2 = y_1;
    y_1 = tmp;

    // second filter step:
    val = 0;
    count = 0;
    for(sub = 0; sub < 3; sub++) {
      if ((p[0] & 0x18000000) != 0) {
        val += (short)((p[2] >> 4) & 0xffff);
        count++;
      }
      p += 3; // go to the next frame
    } // sub
    
    x_2 = x_1;
    x_1 = x_0;
    if(count == 3)
      x_0 = val / 3;
    else if(count == 2)
      x_0 = val / 2;
    else
      x_0 = val;

    /*  y[i] = b[0]*x[i]/2 + b[1]*x[i-1]/2 + b140[2]*x[i-2]/2  */
    /*                     + a[1]*y[i-1] + a[2] * y[i-2];      */

    tmp = (y_1 >> 12) * C_A140_1 + (y_2 >> 12) * C_A140_2 + 
      (x_0 * C_B140_0 + x_1 * C_B140_1 + x_2 * C_B140_2) * 16;

    result += (tmp + 0x80) >> 16;
    y_2 = y_1;
    y_1 = tmp;

    energy_lc += (result * result) / 256;
    *linear++ = result;
  } /* i */

  // update the filtered energy:
  energy_lg = energy_lg - (energy_lg / 4) + energy_lc / 4;
  energy_lf = update_energy(energy_lf, energy_lc);
}

typedef struct _ac97_slot {
  unsigned short slot_1; //slot0 is defined in int0 
  unsigned short slot_2;
} t_ac97_slot;

static void set_ac97_reg(int reg, unsigned long val) {
  ac97_fifo_buffer[ac97_fifo_tx].reg = reg;
  ac97_fifo_buffer[ac97_fifo_tx].value = val;
  ac97_fifo_tx = ((ac97_fifo_tx + 1) % MAX_NUM_PENDING_AC97_CMDS);

  // read back: (debug)
#if 0
  ac97_fifo_buffer[ac97_fifo_tx].reg = reg | 0x80;
  ac97_fifo_buffer[ac97_fifo_tx].value = 0;
  ac97_fifo_tx = ((ac97_fifo_tx + 1) % MAX_NUM_PENDING_AC97_CMDS);
#endif
}
 
// get 80 samples of dtmf:
static void get_dtmf(struct dtmf_state_t *dtmf, short *buffer) {
  int i;
  const int k = dtmf->key;
  const int l0 = dtmf_len[k][0];
  const int l1 = dtmf_len[k][1];
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    const short *p0 = dtmf_freq[k][0];
    const short *p1 = dtmf_freq[k][1];
    *buffer++ = p0[dtmf->pos0++] + p1[dtmf->pos1++];
    if (dtmf->pos0 >= l0) dtmf->pos0 = 0;
    if (dtmf->pos1 >= l1) dtmf->pos1 = 0;
  }
}

// set a dtmf key:
static void set_dtmf(int key, struct dtmf_state_t *dtmf) {
  key &= 15;
  if (dtmf->key != key) {
    dtmf->key = key;
    dtmf->pos0 = 0;
    dtmf->pos1 = 0;
  }
}

// get audio from an audio channel:
short *jit_get_audio(const int c) {
  int timestamp;
  int i;
  int packet;
  int type;
  unsigned char *buffer;

  check_c("jit_get_audio", c);
  if (local_port[c] == 0) return 0;

  // do we have the data from a previous decode step?
  if (jit_decode_len[c] >= SNOM_AUD_BLOCK_SIZE) {
    short *result = &jit_decode_buffer[c][jit_decode_pos[c]];
    // if (jit_decode_pos[c] < 0 || jit_decode_pos[c] + SNOM_AUD_BLOCK_SIZE > 320) printk("jit_get_audio");
    jit_decode_pos[c] += SNOM_AUD_BLOCK_SIZE;
    jit_decode_len[c] -= SNOM_AUD_BLOCK_SIZE;
    return result;
  }

  // get some more data: search for the lowest timestamp
  packet = -1;
  if (jit_decode_start[c]) {
    // in the beginning, search for the biggest! timestamp:
    for(i = 0; i < JITTER_SIZE; i++) {
      if (jit_recv_len[c][i] > 0) {
        const int ts = jit_recv_timestamp[c][i];
        if (ts - timestamp > 0 || packet < 0) {
          timestamp = ts;
          packet = i;
        }
      }
    } // for i
    if (packet >= 0) {
      jit_decode_start[c] = 0; // no more starting
      for(i = 0; i < JITTER_SIZE; i++) {
        if (i != packet) jit_recv_len[c][i] = 0;
      } // for i      
    } // packet >= 0
  } // decode start
  else {
    for(i = 0; i < JITTER_SIZE; i++) {
      if (jit_recv_len[c][i] > 0) {
        const int ts = jit_recv_timestamp[c][i];
        if (ts - timestamp < 0 || packet < 0) {
          timestamp = ts;
          packet = i;
        }
      }
    } // for i
  }

  if (packet < 0) {
    dtmf_recv_ssrc_valid[c] = 0; // underrun: accept any ssrc from now on!
    check_c("snom_aud_fill", c);
    jit_decode_len[c] = snom_aud_fill(c, jit_decode_buffer[c],
				      sizeof(jit_decode_buffer[0]) / sizeof(jit_decode_buffer[0][0]));
  }
  else {
    // decode the first part of the found packet:
    type = jit_recv_type[c][packet] & 0x7f;
    buffer = jit_recv_buffer[c][packet];
    // if (c > 0) printk("%d", type);
    jit_decode_pos[c] = 0;
    if (type == dtmf_recv_type[c]) {
      set_dtmf(buffer[0], &external_dtmf[c]);
      get_dtmf(&external_dtmf[c], jit_decode_buffer[c]);
      jit_decode_len[c] = SNOM_AUD_BLOCK_SIZE;
      jit_recv_len[c][packet] = 0; // this packet is used!
      dtmf_shutoff[c] = SNOM_AUD_BLOCK_SIZE * 3;
    }
    else {
      // no DTMF:
      int *length = &jit_recv_len[c][packet];
      int src_size, dec_size;
#if 0
      if (c > 0) {
        int x;
        printk("recv: t=%d", type);
        for(x = 0; x < *length; x++) printk(" %02x", (int)(buffer[x]));
        printk("\n");
      }
#endif
      snom_aud_get_encoded_length(type, *length, &src_size, &dec_size);
      // if (src_size == 0) printk("s=%d", type);
      jit_decode_len[c] = snom_aud_decode(c, 
        type,
        buffer,
        src_size,
        jit_decode_buffer[c],
        sizeof(jit_decode_buffer[0]) / sizeof(jit_decode_buffer[0][0]));
      // if (jit_decode_len[c] == 0) printk("0 t=%d s=%d d=%d", type, src_size, dec_size);
      *length = 0; // this packet is used.
    }
  } // packet available
  
  if (jit_decode_len[c] >= SNOM_AUD_BLOCK_SIZE) {
    short *result = jit_decode_buffer[c];
    jit_decode_pos[c] += SNOM_AUD_BLOCK_SIZE;
    jit_decode_len[c] -= SNOM_AUD_BLOCK_SIZE;
    return result;
  }
  else {
    dtmf_recv_ssrc_valid[c] = 0; // underrun: accept any ssrc from now on!
    return 0;
  }
}

// reset a jitter buffer:
static void jit_recv_reset(int c) {
  int i;
  unsigned long flags;	
  save_flags(flags); cli();

  check_c("jit_recv_reset", c);
  jit_decode_len[c] = 0;
  for(i = 0; i < JITTER_SIZE; i++) {
    jit_recv_len[c][i] = 0;
  }
  dtmf_recv_ssrc_valid[c] = 0;
  snom_aud_reset_codec(c);
  restore_flags(flags);
}

// find a free parking place and put a RTP fragment there:
static void jit_place_content(int c, int type, int len, int ts, unsigned char *buffer) {
  int i;
  int found = 0;
  unsigned long flags;	
  save_flags(flags); cli();
  check_c("jit_place_content", c);
  // if (len == 0) printk("jit_place_content: len == 0\n");
  for(i = 0; i < JITTER_SIZE; i++) {
    if (jit_recv_len[c][i] == 0) {
      if (!found) {
        jit_recv_len[c][i] = len;
        jit_recv_type[c][i] = type;
        jit_recv_timestamp[c][i] = ts;
        // if (len > sizeof(jit_recv_buffer[0][0])) printk("jit_place_content: len = %d\n", len);
        memcpy(jit_recv_buffer[c][i], buffer, len);
        found = 1;
      }
    }
    else {
      if (jit_recv_timestamp[c][i] == ts) {
        // printk("d");
        jit_recv_len[c][i] = 0; // double packet!
      }
    }
  } // for i

#if 0
  for(i = 0; i < JITTER_SIZE; i++) {
    if (jit_recv_len[c][i] > 0) used++;
  }
  if (used > 0) printk("%d", used);
#endif
  restore_flags(flags);
}

// process an incoming rtp packet and put the media data into the
// jitter buffer. This is done according to the media type in smallest
// possible portions.
static void jit_recv_rtp(int c, unsigned char *content, int len, unsigned char *adr, int port) {
  int i;

  check_c("jit_recv_rtp", c);
  if (len >= 12) {
    len -= 12;
    if ((content[0] & 0xc0) == 0x80) {
      const int type = content[1] & 0x7f;
      int ts = ((int)(content[4]) << 24) | 
        ((int)(content[5]) << 16) | 
        ((int)(content[6]) << 8) | 
        (int)(content[7]);
      
      unsigned long ssrc = ((int)(content[8]) << 24) | 
        ((int)(content[9]) << 16) | 
        ((int)(content[10]) << 8) | 
        (int)(content[11]);
      
      if( type == 13 && silence_suppression){
        //printk("type = 13\n");
        cng_on[c] = 1;
        cng_vol[c] = content[12];
      }else{
        cng_on[c] = 0;
      }
      
      // check if the SSRC if is correct:
      if (dtmf_recv_ssrc_valid[c]) {
        if (ssrc != dtmf_recv_ssrc_id[c]) {
          return; // wrong id
        }    
      }
      else {
        jit_recv_reset(c);
        dtmf_recv_ssrc_id[c] = ssrc;
        dtmf_recv_ssrc_valid[c] = 1;
      }
      
      // check if we should use the source address for symmetrical RTP:
      if(sym_rtp[c] && (remote_adr[c].sin_port != port || memcmp(&remote_adr[c].sin_addr, adr, sizeof(remote_adr[c].sin_addr)) != 0)) {
        memcpy(&remote_adr[c].sin_addr, adr, sizeof(remote_adr[c].sin_addr));
        remote_adr[c].sin_port = port;
        snom_aud_set_default_destination(c);
      }
      
      stat_rx[c]++;
      if (type == dtmf_recv_type[c] && len == 4) {
        // get the duration of the current digit:
        int duration = content[14] * 256 + content[15];
        
        // check if we have a new digit:
        if (dtmf_recv_pos[c] >= duration) {
          // new digit:
          dtmf_recv_pos[c] = 0;
        }
        
        // put pieces into the receive buffer:
        while(dtmf_recv_pos[c] < duration) {
          jit_place_content(c, content[1], 4, ts + dtmf_recv_pos[c], &content[12]);
          dtmf_recv_pos[c] += SNOM_AUD_BLOCK_SIZE;
        }
      }
      else {
        int frag_length, dec_size;
        snom_aud_get_encoded_length(type, len, &frag_length, &dec_size);
        // if(c > 0) printk("%d", c);
        for(i = 0; i < len; i += frag_length) {
          jit_place_content(c, content[1], dec_size, ts, &content[12 + i]);
          ts += dec_size;
        }
      }
    } /* rtp */
  } /* possible length */
}

// ICE returned a result: 
void snom_aud_ice_result(int c, struct in_addr *adr, int port) {
  if(ice_contacts[c].accept) {
    memcpy(&remote_adr[c].sin_addr, adr, sizeof(remote_adr[c].sin_addr));
    remote_adr[c].sin_port = port;
    snom_aud_set_default_destination(c);
    ice_contacts[c].accept = 0; // stop the process
    sym_rtp[c] = 0; // dont accept symmetrical RTP any more
  }
}

// put 80 samples into the output buffer. This action may start the
// lower half processing, but the encoding is not done here.
static inline void set_rtp_audio(int c, short *clipped_buffer) {
  int *len = &jit_send_len[c];
  check_c("set_rtp_audio", c);
  if (*len + SNOM_AUD_BLOCK_SIZE <= sizeof(jit_send_buffer[0]) / sizeof(jit_send_buffer[0][0])) {
    memcpy(&(jit_send_buffer[c][*len]), clipped_buffer, SNOM_AUD_BLOCK_SIZE * sizeof(short));
    *len += SNOM_AUD_BLOCK_SIZE;
  }
}

// add 80 samples
static _INLINE_ void clr_signal(int *dest) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE / 8; i++) {
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
    *dest++ = 0;
  }
}

static _INLINE_ void add_signal(int *dest, const short *src) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE / 8; i++) {
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
    *dest++ += *src++;
  }
}

static _INLINE_ void set_signal(int *dest, const short *src) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE / 8; i++) {
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
    *dest++ = *src++;
  }
}

static _INLINE_ void clip_signal(short *dest, const int *src) {
  int i;
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    const int c = *src++;
    *dest++ = CLIP(c);
  }
}

// generic routing to handle all the audio stuff for one packet:
//
// out = 80 shorts output buffer
// in = 80 shorts input from microphone/dtmf (may be zero if turned off)
// rtp = bool if rtp should be used
// synth = 80 synthesize samples (if available)
// echo = bool if in should be copied to out (dtmf loopback)
//
static inline void generate_output(short *out, short *in, int rtp, const short *synth, int echo) {
  // get the audio from all channels:
  static int buffer[SNOM_AUD_BLOCK_SIZE];
  static int buffer2[SNOM_AUD_BLOCK_SIZE];
  static short clipped_buffer[SNOM_AUD_BLOCK_SIZE];
  short *rtp_in[MAX_NUM_RTP];
  int c; // channel index
  int i;

  // white noise for CNG:
  /*static const short rand_data[SNOM_AUD_BLOCK_SIZE * 4] = {
    27, -5, -29, 43, 22, 7, 23, 16, -36, -26, 29, 40, -4, 48, 40, 39,
    -7, -13, -21, -37, 10, -49, 11, -31, 36, -13, 22, -17, -14, 43,
    21, 44, -4, -27, -5, -32, -46, 21, 35, 29, 25, 34, -11, 46, 46,
    35, 30, -40, -1, 4, 31, 49, -7, -44, -25, 39, -37, -12, 12, 32,
    43, -31, -3, 27, -50, 30, -37, 1, 33, 1, -29, 6, 8, 15, -3, -50,
    32, 28, 8, 8, 46, -48, 19, -2, 42, 0, -31, 26, -44, -16, 2, 20, 8,
    48, -25, 2, -15, -8, 10, -21, -26, 5, 16, 14, 28, 17, 36, 31, -38,
    -22, 22, 27, -35, 31, -2, -39, 44, 40, 32, -20, -35, -31, -27,
    -38, -2, -33, 43, -42, 28, -7, -17, 26, 49, -11, 7, -11, -1, -23,
    30, 6, -7, -39, 19, 6, -9, 38, -2, 28, 12, 26, 30, 0, 6, -1, 48,
    44, 6, -14, 36, -14, -23, 27, 30, 16, -29, -24, -47, 28, -16, -26,
    41, 47, 47, -38, 44, -25, -14, 7, 24, 2, -29, 10, -5, -11, 20,
    -17, 5, -41, 30, -22, 3, -11, 20, -28, -22, -2, -28, 11, -47, 34,
    39, 31, 46, -16, 30, 25, 16, -47, -9, -5, 10, 21, -28, 33, -24,
    -15, 9, 8, -3, -35, 12, 0, 44, 48, 18, 42, -2, 0, 39, 10, -27,
    -20, 31, 11, -2, 26, 0, -40, 7, 36, -43, 44, 49, -32, -46, 25, 33,
    -16, 34, 2, -26, -31, 33, -46, -31, -24, -33, 29, -17, 46, -50,
    24, -10, 7, -41, 33, 31, -18, -35, 36, -11, 14, -29, 39, -23, 21,
    20, 15, 34, 14, -17, -21, -34, 14, 8, 9, -49, 44, 43, 28, 7, 39,
    -42, -43, -17, 42, -37, -43, -9, -14, 39, -37, 0, -6, 20, 18, -47,
    -43, 19, -38, 40, -23, 30, 2, -31, 32, -44, -32, 12, -49
    };*/

  static const short rand_data[] = {
97,-35,-7,60,33,96,-47,14,26,95,41,42,-25,71,32,-99,98,76,-24,-2,
-93,72,-40,81,97,-42,1,44,-25,-6,5,52,6,-60,-32,-13,93,-15,53,-79,
-74,62,-55,62,10,-36,-74,62,-72,-6,-49,-8,-98,-64,-79,3,61,-39,97,99,
-48,13,61,63,51,60,-48,76,79,-84,-65,-27,-66,83,-76,-34,7,18,-40,-66,
2,97,-51,86,-5,69,11,-1,97,33,-93,26,-62,59,36,-62,86,21,-38,87,
52,-43,70,-54,68,57,1,-10,-89,-73,-29,19,67,-63,37,-3,-98,56,-50,-95,
-95,50,7,2,68,24,-53,83,13,81,-34,-37,23,88,-8,-8,97,-34,72,94,
-79,-6,94,-79,96,64,64,-72,-11,44,-59,57,-97,63,9,55,11,39,90,-10,
-89,-26,-19,-67,51,60,-68,-83,20,-14,85,-36,-50,-72,34,56,68,-12,40,60,
-32,-99,-53,-7,95,-96,-19,63,-90,84,-72,76,-15,-94,24,-66,-48,-53,-86,-27,
-54,46,-85,-41,24,-79,-1,-8,13,70,21,80,-95,41,94,-18,-31,-25,7,53,
52,-91,75,66,48,70,33,-74,73,-11,53,-35,65,-28,-66,61,75,-81,-97,44,
-9,50,15,-98,85,38,-96,10,-96,74,92,3,97,39,25,13,-90,-93,76,-81,
-37,-25,6,-54,-69,-12,73,97,25,83,-71,23,86,-12,79,30,38,7,22,56,
32,-17,-45,-39,-70,60,98,1,-96,-49,-37,-28,-100,41,-34,81,34,95,-1,3,
47,43,25,-53,-23,70,20,39,38,16,-86,75,-30,-78,-84,-41,-71,-34,-21,64,
71,74,-48,-4,72,50,-76,61,75,28,-94,-82,64,64,-9,94,48,-85,96,-42,
23,-85,-52,-39,31,33,80,-34,-47,-92,68,85,-67,57,90,-50,15,96,-19,90,
-60,10,47,-85,48,-4,26,54,42,70,-13,-3,76,-81,-33,-71,-29,4,-52,97,
-96,14,67,98,-90,33,-97,55,38,-72,3,16,5,-90,6,15,24,-67,-26,-92,
32,-85,-29,-37,33,23,-92,40,-63,-68,-71,58,66,-32,92,88,92,-90,-50,-59,
-1,-21,15,34,77,-67,64,-25,62,9,-9,-17,98,43,52,9,80,-78,-42,-69,
-26,47,55,-93,-54,-26,71,-87,80,-64,81,53,-15,36,95,-91,-92,-92,-66,-18,
-87,-37,27,87,70,-5,-37,35,52,-64,49,-28,-59,17,-74,-31,-9,26,-78,-80,
-6,43,73,-76,41,-61,31,83,-17,-64,-8,45,-72,-5,-80,-11,-26,-85,-94,78,
-34,51,23,39,-55,2,-49,-49,-29,-13,-41,1,74,-64,58,-19,-88,88,61,56,
68,-31,33,10,38,19,-21,-49,95,43,78,56,-62,55,31,73,53,77,49,74,
60,18,34,-33,39,-12,18,22,-17,-62,74,-7,-12,94,-37,64,-60,35,49,-9,
-57,-91,97,31,-20,46,-83,52,-47,65,93,84,-2,-22,-85,-10,-5,26,86,-74,
39,74,7,-81,-66,-1,-65,-11,-59,59,16,-49,41,92,9,20,45,-41,-46,46,
-78,-42,-52,6,2,55,-68,16,26,92,43,-16,13,-68,51,76,-96,-53,10,-4,
8,-46,-22,-79,25,-58,-20,88,22,-8,30,-77,-37,-31,45,-19,-57,-97,10,84,
-6,48,23,38,51,38,-43,-37,-19,-98,18,43,-31,-2,19,57,39,49,-88,49,
-85,66,-1,17,66,-7,42,-49,-16,80,87,99,99,-79,6,80,-11,26,-71,-60,
-2,37,84,-83,1,82,8,92,62,-79,-20,-69,-21,-83,-72,-86,-56,95,-63,74,
-77,85,25,-94,-9,-23,-26,31,-6,-26,75,93,61,-17,-51,-26,-39,-93,-39,43,
23,9,-73,-62,-64,-77,6,56,50,68,76,22,-87,-41,-100,-53,91,67,-64,32,
95,81,-42,-31,62,40,-16,24,20,90,13,-19,-2,51,-54,94,1,16,16,-48,
38,-52,3,70,-44,-29,-37,10,-48,-47,-53,-74,-88,-84,-11,-20,22,54,23,79,
-65,78,82,-78,26,-67,-21,-79,65,18,66,34,-63,21,-13,-46,96,-81,13,-49,
-35,-51,75,95,-46,-78,71,-57,-9,-58,-65,45,49,13,-47,29,-76,69,-81,-83,
44,52,-70,-34,-15,-99,-25,-27,41,77,-54,-3,55,16,62,86,79,-56,-60,88,
-38,-3,-89,-23,92,-2,-58,32,80,-18,-92,21,-72,-35,-91,21,-37,-60,9,-41,
26,47,-36,56,-63,28,-97,87,47,62,-66,-10,76,-87,89,-67,-69,99,-17,94,
18,27,59,-58,-34,-36,42,14,20,-29,-85,66,-51,72,84,-42,63,25,11,-88,
62,-88,29,69,-39,60,9,-79,-3,10,-74,-74,-18,58,41,9,66,73,-88,87,
6,81,-23,28,-61,83,-75,44,-9,-93,88,29,-83,-78,36,60,95,21,45,-11,
-65,91,-30,83,61,22,-45,-38,53,-15,-41,59,-86,-39,43,-6,57,61,35,-98,
4,-69,-58,-1,-20,25,-25,-31,-60,21,-40,-66,80,48,-37,-77,74,-82,28,-62,
30,-78,65,2,2,-79,-13,-26,-84,35,-39,-85,-4,40,-5,43,34,-72,87,77,
100,57,-91,97,19,36,-95,66,-36,-49,97,-44,-84,32,73,-38,46,-24,-54,-68,
-41,88,-13,-41,13,-43,-61,44,-36,-12,58,-45,-33,80,40,-6,6,-60,-55,9,
-78,85,-9,100,48,-95,29,38,-59,-49,36,57,76,49,-35,-48,83,-65,-7,-46,
-85,89,-27,32,24,-88,-84,45,14,78,19,-15,-79,2,72,96,-10,34,46,-22,
32,-79,79,72,87,-99,-29,-42,-20,43,-76,39,92,96,33,-82,90,33,-67,-17,
65,-93,37,41,66,63,-43,-45,3,-68,-89,63,-8,88,51,40,21,-69,-71,-4,
78,64,-51,62,17,75,-65,94,-99,-12,84,-75,32,63,10,98,20,-78,30,-14,
13,-66,69,82,78,100,73,-39,77,39,88,-70,-60,64,-63,-5,71,20,60,-13,
65,75,79,-22,38,92,-90,-44,-69,-10,88,-16,-76,-45,52,49,83,89,14,-95,
2,-41,-55,-23,21,-32,-95,-8,-14,74,8,-34,57,-22,47,-9,56,-36,-77,-88,
93,-71,26,18,77,-73,49,87,16,-81,95,88,94,-54,99,97,65,-2,79,-62,
-57,31,47,6,-91,2,-32,-39,-9,1,21,88,20,23,2,34,-51,-69,-26,66,
-17,-49,85,-44,1,10,-95,75,45,63,-97,7,59,33,-64,23,-33,-14,-31,65,
-70,43,25,-53,-30,78,21,-30,84,100,94,-8,79,18,3,41,21,-89,-70,-92,
-67,-51,81,32,-31,74,-100,33,-99,-9,87,-89,61,-66,40,-43,-52,-97,41,19,
-11,-86,-18,-84,42,-62,-20,-6,-1,21,-99,31,-84,54,-14,-77,-34,-96,41,-31,
-57,84,-94,65,-55,-28,-65,-65,-65,-92,70,68,17,93,12,-16,99,7,-55,17,
40,-88,8,27,1,8,-28,33,78,-93,27,-63,-24,91,-99,-93,-76,-33,52,-79,
98,-26,64,-92,-38,-41,38,74,91,-25,-77,-2,-86,21,-74,-61,70,100,5,-40,
-22,-9,47,-27,-65,27,95,-28,96,38,-3,-93,31,10,-32,54,-59,-97,10,80,
-10,42,-8,37,35,-99,-22,-70,97,67,-78,88,79,-35,-38,68,-97,-43,27,-71,
-37,47,67,66,-91,7,-30,-72,15,-80,92,-5,-98,-32,-15,51,37,53,-28,-36,
-36,95,-28,-16,-7,17,-89,83,-87,-38,90,-76,73,-17,47,8,59,27,65,8,
-93,-58,97,71,45,19,-20,-21,-88,65,-19,93,-90,-23,87,15,-17,-45,-39,-17,
52,61,-28,-43,-27,53,-92,76,63,-34,-95,-80,-26,-48,-89,49,-94,78,-91,75,
38,-64,70,24,-14,36,39,-25,-70,-65,-75,-68,58,-12,-37,27,-72,10,5,-80,
-83,-38,-26,50,1,-20,80,-59,-36,-34,-49,-90,83,-32,7,-65,12,-72,15,-35,
-75,-68,-37,-82,22,-41,61,-30,-29,-96,35,-79,-48,2,17,-36,52,-91,-35,-74,
13,60,63,94,-71,-99,-6,98,38,18,97,-87,-94,98,-14,54,-9,46,-95,4,
73,-64,-85,80,-36,38,-74,-1,-66,-29,67,-15,16,81,53,-63,-1,45,93,58,
-96,10,18,-69,37,49,-42,31,-38,-77,-84,-52,26,-22,-4,43,77,8,55,61,
78,-53,40,53,87,-72,89,27,-20,2,45,-27,-40,36,60,-23,-63,63,-12,9,
74,33,22,-25,-60,-24,94,77,50,-24,-66,80,72,30,-66,9,-17,-56,90,-62,
26,50,-86,41,-75,31,55,37,84,-34,61,36,-35,-64,53,-28,-100,-76,-6,-41,
-85,-47,-65,74,-84,-80,10,-54,-49,-9,26,-95,67,20,-94,-82,54,84,86,52,
-59,-4,78,6,-86,-70,74,-12,-17,-33,-41,43,59,-89,-60,-9,100,69,84,10,
-74,55,17,5,-41,93,-94,84,81,-32,-84,59,-100,67,-1,66,44,-99,-28,-80,
34,-3,50,-42,3,13,-7,-98,55,17,-54,49,-33,67,32,-19,95,98,61,32,
11,60,32,-97,-66,-66,19,13,77,51,-10,85,25,-39,-33,47,-95,30,-14,-6,
-34,-91,-30,-90,-18,-72,92,15,-28,17,23,59,48,-6,-61,-55,-67,95,71,92,
-100,-76,65,-28,11,49,-42,17,-65,-2,-37,-50,-73,-10,-79,-7,70,-42,-66,-19,
-86,81,-55,49,45,57,56,75,-65,-96,-44,-1,80,-73,-88,63,67,-87,78,-60,
-91,-41,-85,-94,57,-41,12,-19,-82,-2,-74,10,-70,-69,83,35,7,50,46,-47,
69,62,78,-85,33,-91,-23,-92,14,59,92,35,3,-9,-14,53,-21,4,93,99,
94,28,-85,-15,25,-13,83,-2,88,72,-16,-9,-40,-38,60,41,28,-30,65,-67,
-27,-20,84,52,79,18,32,-91,-51,65,-36,68,-99,-42,-19,12,-50,94,-80,83,
93,-59,96,53,88,15,-30,-15,79,39,59,-92,26,-78,7,-68,-50,30,95,-25,
-82,-43,4,-96,38,21,-46,28,-92,-42,52,61,-74,99,69,18,-73,37,-75,-3,
53,17,99,-18,86,86,59,-93,93,57,-12,19,-100,7,11,5,-2,-87,-86,-40,
-50,-38,-24,72,-37,28,-68,-95,6,-97,35,51,-68,-15,-31,-49,-29,-52,94,85,
-39,-60,38,88,-39,75,-20,-67,34,-16,-9,77,-72,12,71,-98,-35,33,-75,22,
-48,39,-97,-56,-89,31,-48,36,-29,-29,10,-54,-66,52,83,-98,85,-14,-45,-2,
24,31,-33,16,-55,-45,20,74,-73,52,-21,92,7,-55,-100,-65,-34,15,97,-83,
7,25,61,-58,-3,95,-2,62,74,-43,-49,-83,-15,-8,29,-83,47,67,47,-83,
-4,71,-89,43,-46,89,33,73,68,-66,-71,41,-37,95,-46,-5,-30,100,2,47,
90,-30,-46,-30,-98,-71,17,-58,45,-60,78,-58,-33,-27,-12,65,-74,-34,-34,26,
10,33,-15,-14,-42,44,-17,-55,-75,29,-43,-67,70,94,98,-14,-14,93,66,67,
99,54,88,85,-14,8,-71,99,-52,32,92,-22,-24,-17,24,54,94,-28,-74,-16,
17,-94,29,-88,98,1,-21,-57,56,100,96,-88,-5,-30,-5,96,-51,-49,-99,-35,
-89,-78,-45,6,-28,66,42,-63,-16,-53,68,48,-87,-58,19,-24,10,-71,76,-31,
64,67,77,-49,-12,19,3,75,98,38,-78,-12,-47,-15,81,74,-88,81,65,94,
82,75,1,58,-62,-32,20,40,91,72,-78,-55,-31,16,28,26,17,100,-9,-56,
59,11,40,-22,35,35,-39,-73,22,9,-37,-50,-14,-99,-24,6,-72,-77,89,52,
77,22,89,-3,97,23,19,-48,-69,-5,98,-23,-88,57,-68,2,-15,-81,8,99,
-95,51,-53,-82,-84,-80,36,-1,96,-7,33,24,-52,-89,-25,95,56,13,71,-92,
-64,-14,-97,10,89,11,93,45,96,-48,25,-99,24,56,57,-95,-17,25,-91,-1,
-63,-82,10,72,-53,91,47,-2,-47,59,-52,86,52,-59,82,-100,17,60,-99,-89,
38,77,-80,69,-76,24,-45,-39,83,89,87,15,52,-15,27,60,-65,-96,-69,45,
63,86,-82,7,29,19,-78,-28,-75,-42,-16,26,-53,63,-11,-54,-67,4,-61,-18,
-8,78,-77,2,-93,-67,40,4,-22,23,-56,-97,-46,-68,-32,-82,44,25,62,76,
18,44,28,94,-99,-48,-7,52,64,95,99,85,-62,47,62,70,-64,52,-58,69,
11,39,-20,-77,87,-30,-20,90,-94,59,59,-80,49,-1,-66,-84,-89,25,5,-8,
-84,-47,5,-41,-70,-3,-91,-79,-95,29,-37,70,-67,63,-25,-58,27,-47,-17,58,
-27,-50,46,22,-24,2,-13,-36,82,38,95,-19,-92,-59,73,39,-4,100,-67,38,
-64,12,85,-71,-74,64,-84,-61,58,42,-42,-63,42,-79,49,-36,-45,83,1,-40,
88,80,14,32,7,81,-47,-83,87,44,40,-4,-1,63,-59,-88,-69,58,60,-57,
23,90,-65,32,-86,78,-43,-7,68,30,12,-36,-68,94,52,-42,94,-12,-24,-23,
91,92,-11,87,85,92,-66,86,37,-15,-69,68,11,-16,-6,45,56,26,-54,-67,
-95,82,79,14,32,73,46,95,70,-1,47,-64,-40,-48,-32,46,-44,63,63,-79,
12,52,72,-63,66,94,34,-27,-2,55,59,-2,-82,24,83,-15,41,-61,86,-89,
1,48,-10,-57,6,69,3,-89,59,86,-58,-34,82,-38,-83,66,78,48,70,-56,
20,-44,-83,-44,32,-95,58,-27,-4,-50,9,-69,5,3,-60,5,13,-60,74,76,
-58,25,7,64,-79,91,-65,-47,90,-53,82,61,-58,-26,-3,77,-99,80,84,-92,
-83,-83,5,-31,47,64,90,60,-32,77,-48,-14,-20,-90,-19,-36,-91,78,33,-28,
-92,-32,91,-18,15,-79,44,-35,-88,-16,95,99,-87,-55,-11,23,45,-81,46,-26,
88,25,76,-64,-75,-44,-50,63,53,19,-50,79,-27,62,67,-55,37,-36,82,-2,
53,91,73,-88,77,-7,-17,39,65,-54,21,54,-52,61,57,-32,-100,-99,-7,-12,
49,82,-29,-23,86,-65,28,50,24,-76,44,-12,97,41,-49,50,63,-34,-60,-27,
-23,-32,-61,48,48,88,43,80,-27,-17,-48,-97,16,83,33,-29,-10,-17,46,-30,
23,31,-5,49,-75,54,6,-24,-6,-78,-29,-83,-86,-14,-61,-59,-49,65,35,-14,
33,-9,-80,-82,-38,93,77,21,-92,-42,38,-73,31,-80,-91,-46,91,66,-77,-100,
16,-32,-63,-53,-72,-49,64,-90,77,-95,85,-94,-47,-37,-60,84,-51,28,23,68,
72,-13,-77,6,-97,-56,79,87,100,64,43,19,54,40,49,33,45,38,-67,-94,
-5,-36,-92,29,50,-14,-95,68,-40,57,-85,93,43,-57,-80,7,-18,37,-94,5,
-4,-85,37,-71,-29,95,-28,70,24,-30,-13,73,-56,-91,-8,-78,-70,-1,-43,-47,
41,98,-27,-66,46,46,-18,-3,-42,56,-12,-86,58,16,-41,-94,-1,39,-86,60,
-74,-23,98,81,3,18,-34,-48,-85,10,11,-88,39,-19,62,-9,-73,6,-10,72,
68,-92,59,-62,-67,-94,87,81,-83,-67,63,49,12,-68,65,-18,69,21,-21,100,
-26,3,56,-5,-69,39,-53,93,-64,11,-26,41,42,77,-11,-17,85,5,58,70,
33,-21,47,33,-5,64,11,39,27,-57,-70,42,-71,93,-85,63,16,-26,-14,-40,
-13,-85,-87,88,-64,90,95,85,-14,88,-85,51,62,95,-76,60,70,-60,-78,45,
-99,31,-32,32,-49,40,56,78,-17,-98,8,-20,-60,-3,-61,41,80,-63,-85,50,
11,28,-51,-66,-31,31,39,24,-72,56,-64,-89,-58,99,-8,80,40,57,76,-14,
56,22,50,55,-97,89,-68,77,88,77,-9,-45,-78,62,-88,-3,49,23,75,73,
29,1,11,-73,-27,-66,-48,-44,72,6,-1,87,-69,17,2,16,66,-17,79,-23,
-37,-61,-41,55,-70,78,-79,3,77,-14,-71,-2,-7,-94,-14,-51,-80,-88,55,34,
25,98,40,62,40,52,-77,25,-58,13,12,-58,-24,97,-9,13,91,-12,55,-55,
-26,97,67,-58,-24,-85,-20,24,-38,-71,78,7,-28,96,-49,-52,-6,-44,-40,-53,
15,49,38,-8,-97,16,-88,-3,-95,52,-99,1,-56,-16,-51,-38,44,-73,21,45,
-75,70,19,72,-59,-11,35,25,-52,-62,4,-21,-28,52,76,-51,-54,97,4,59,
87,-34,-43,-29,-77,63,61,-87,-85,9,63,-18,-14,91,-28,-89,49,-6,57,-94,
13,24,-32,100,39,-95,96,-15,42,-72,-32,-65,-19,45,-69,80,49,81,-85,-2,
-14,90,41,93,-27,-90,-8,-66,-75,92,21,25,-75,-97,38,-12,-40,2,18,-77,
-61,-21,100,53,97,21,9,-44,84,-64,63,-29,48,67,-99,74,38,72,8,-63,
-68,-32,-84,-32,72,-1,-85,-28,72,48,11,-65,-56,76,-83,9,-74,51,-46,61,
40,55,-83,-64,52,39,-32,-13,18,-94,-29,87,6,-36,53,82,-91,3,87,69,
27,58,95,-57,-27,16,6,66,55,71,12,22,29,-47,-83,-54,43,55,85,-14,
-92,60,24,97,18,68,86,-88,66,21,-7,7,-46,6,-41,79,22,72,-47,7,
94,67,-90,65,12,84,51,95,88,-15,-53,-27,31,-26,-92,-3,84,-38,91,-16,
-82,24,59,63,-22,16,-67,10,-37,-25,30,67,-94,-77,-10,81,-97,-58,27,33,
-4,68,-80,57,59,98,84,63,-32,-10,23,81,77,-30,6,8,-95,-62,-43,76,
21,-6,32,-30,-10,88,57,11,46,12,99,-65,57,-14,60,55,-49,-33,-2,-86,
-42,94,-91,47,33,-6,61,-63,57,-70,-89,56,-72,-95,-79,28,-46,-36,-61,-43,
-69,-41,-4,37,64,2,40,-24,23,-34,97,31,43,53,-52,31,-88,72,89,100,
51,-13,-75,78,-37,-22,-46,-46,56,9,-5,-19,26,-86,-32,44,92,-32,13,16,
-54,-59,69,-6,-14,-98,12,-97,25,-95,2,-40,-66,16,-21,17,15,-70,-61,18,
45,43,21,-29,28,41,-21,-35,-56,52,-24,35,65,-12,-70,70,-83,92,-30,-59,
82,-93,-64,54,46,53,-96,3,-53,68,98,-27,65,25,92,11,36,-50,60,3,
24,-62,-67,-32,-30,-49,100,-33,31,47,-12,-84,-42,35,89,64,82,-49,-20,13,
-49,-97,84,8,-22,92,50,57,58,-5,-87,97,-12,-13,89,-11,70,77,11,-52,
86,-24,16,-86,-42,39,-40,41,-14,79,63,-73,52,-87,-94,-43,-70,32,76,-6,
56,49,-90,-12,87,2,-83,-23,-59,50,-3,37,27,35,-48,-30,66,73,30,-6,
-36,-65,44,-63,58,19,56,-68,12,-80,-4,49,-41,76,-8,-95,66,80,-93,72,
-19,75,-27,-47,-83,42,67,31,-51,-54,23,30,-88,84,33,-8,15,25,43,65,
3,-24,15,25,72,-29,-30,42,-14,46,16,9,23,-6,85,-90,-34,-58,-79,-55,
-81,-52,-33,36,62,-59,50,-88,-23,-84,-84,68,46,-1,42,-24,-59,2,93,-98,
49,-49,95,-37,95,-51,-27,24,9,-43,-41,-16,91,-85,23,93,100,65,65,-28,
-27,27,-53,61,-97,98,66,-88,64,36,87,-61,38,-89,25,10,7,63,-54,60,
-86,11,-61,79,-68,-20,6,23,93,-27,97,-8,27,36,-77,-79,-58,52,-21,4,
96,26,50,77,-96,15,-52,28,-82,-50,-18,55,-95,-22,96,59,-46,-75,-45,99,
82,-17,16,-54,-70,-40,31,-4,-84,-49,88,-15,17,14,70,-42,-90,-27,-68,-46,
-92,-66,-19,19,15,-89,-53,-22,11,20,-56,-40,61,62,42,-44,-54,20,42,71,
80,26,-100,-9,35,95,86,2,38,-73,21,-65,-62,18,-9,10,75,73,-90,82,
98,-71,-100,18,-69,-36,-44,6,26,-7,-85,-63,99,71,38,-26,12,-10,-87,-54,
-5,54,-3,-39,-62,60,87,36,54,-88,-13,5,92,98,91,-55,99,-62,-43,16,
86,81,86,-14,-45,-20,35,-58,58,-53,10,43,-8,-19,-61,68,-17,100,6,58,
-81,-94,-80,-5,30,-18,93,21,93,16,-7,3,-80,-74,23,8,-73,-61,-74,-43,
33,-67,44,-67,-15,68,87,7,99,-91,66,83,-35,3,98,-48,95,-100,-19,39,
41,21,-58,-45,6,66,-19,79,-86,-29,86,72,48,-46,-28,71,-40,-49,29,-58,
-31,-41,18,-70,46,77,24,-48,24,-61,33,79,-86,52,-64,-68,12,-87,52,10,
39,13,53,85,62,-32,-50,52,-52,-62,79,72,-86,4,-63,-29,-93,-29,-66,71,
2,-71,49,62,18,-34,-45,50,96,-19,-48,-36,36,48,-9,72,-87,-19,50,-56,
66,-56,-83,54,87,-87,53,-18,-4,20,81,31,64,-46,-28,61,-81,-41,-70,18,
7,91,81,-39,31,-23,-35,-72,83,-50,-14,-80,50,32,-36,-55,6,44,87,45,
58,62,76,95,-96,55,-87,-96,-65,97,-61,52,41,85,79,-55,28,73,-63,-77,
-60,-17,-82,-69,19,37,81,-77,-67,82,18,69,86,-17,-8,76,-33,-17,-86,38,
11,77,86,-76,-46,31,18,-66,81,86,-75,26,78,65,41,20,-41,-73,-15,-74,
42,57,-46,-100,65,97,84,39,-60,-100,41,-9,12,6,64,-71,21,-10,22,-7,
94,-98,5,-8,48,87,-32,79,-32,-87,-81,47,54,-73,-21,97,25,64,-62,1,
-13,-21,14,80,10,-57,67,-62,-14,-53,69,60,21,-97,-33,-65,30,25,42,42,
58,-80,75,50,-76,87,-46,-66,-46,-29,49,25,-76,49,-20,73,-94,54,13,27,
21,-40,-85,-61,67,-19,-93,86,34,-72,-9,-57,-37,2,41,-85,31,4,71,69,
55,-55,2,28,-87,-2,21,-58,-9,-74,72,95,92,-2,44,93,-85,52,-66,11,
-21,-73,22,-18,35,50,74,5,-92,81,96,-13,-33,54,28,-73,-28,48,-31,57,
44,39,65,-42,-23,-19,74,-89,-80,32,85,30,-1,19,-20,73,-68,-76,-18,11,
81,22,-74,57,9,-87,-85,-76,-37,6,-13,41,-41,24,60,-96,63,-22,49,33,
-5,-46,-32,84,35,55,90,17,3,77,-69,-57,72,83,5,-11,60,-67,80,3,
98,-43,-82,-32,-93,-54,77,-35,10,-93,-97,-60,-66,32,39,-62,-33,73,73,97,
-57,4,70,-46,86,-70,-39,-32,18,54,70,-67,47,-78,100,-1,40,4,-87,86,
-11,-61,93,53,-15,75,-88,4,-55,-17,-55,-10,95,-57,-18,-28,-98,39,78,29,
-60,-62,62,-34,70,-78,-81,-69,-75,96,-8,26,-65,7,-64,45,-57,-13,-12,100,
11,79,-75,-1,36,83,-56,8,-30,-95,-73,-63,16,53,13,20,-100,63,-88,7,
39,-7,-11,84,-18,62,-84,-21,-47,-33,93,91,62,-3,-23,-66,-46,-84,79,38,
55,94,25,-40,65,-37,-31,-89,-25,100,-35,37,-23,-60,-16,-36,69,-6,80,-15,
-88,-72,-63,-66,19,82,35,-53,40,8,32,20,95,41,-85,20,36,77,-74,56,
78,-7,-86,-30,-69,-29,-91,-11,67,71,28,-9,33,17,-4,66,-41,59,2,-58,
86,54,10,7,24,-93,-53,12,8,-85,47,-86,46,52,-53,7,-69,-40,-57,85,
-8,36,55,38,89,4,49,-80,-39,-28,61,-26,-100,73,-57,1,-64,55,82,75,
10,23,-8,-94,64,55,100,40,64,-67,-84,87,40,-33,-35,86,-8,-20,79,49,
73,31,-100,-53,-49,63,-68,-55,-88,83,-71,16,-22,2,-66,-46,-17,-82,-42,-49,
-46,-33,-91,94,-62,43,-95,-87,94,3,27,63,-54,-62,-59,-87,-50,75,88,71,
-19,-81,-71,96,72,-86,76,56,-41,-52,41,-7,-57,-14,48,-33,21,-95,12,23,
3,40,76,-61,-42,63,26,24,-68,6,-12,-3,-31,-7,84,23,-39,84,-46,90,
-21,-33,-10,-34,-88,-34,74,-82,-15,70,-10,30,-37,-66,-57,33,47,-29,39,9,
-22,8,-84,51,-45,-65,61,56,53,-60,34,-62,92,44,-13,2,-80,67,40,-69,
92,-100,-66,-91,51,-70,53,21,30,1,87,-40,-76,78,99,70,34,12,-3,3,
93,26,-46,-37,40,52,65,32,26,37,65,71,-50,65,6,-39,49,-88,96,-13,
1,-74,-7,62,-48,-67,1,62,-22,52,60,-23,88,-43,-39,-31,-99,-76,-27,-95,
66,20,19,15,-37,-95,6,-59,-85,6,67,-50,-21,-91,-90,72,-5,-74,87,78,
-82,-42,34,-26,16,22,-22,8,-56,83,-14,12,-94,-39,22,-79,-49,-71,91,84,
-63,63,-100,-22,-99,44,-72,-15,63,49,-99,23,35,87,75,-69,8,-28,51,85,
-46,93,89,56,49,-69,43,91,-37,24,-39,97,38,-44,77,69,96,-77,-44,89,
89,-90,-33,-47,56,43,85,62,-42,-70,-33,-24,53,-73,66,-47,24,34,2,-29,
-96,35,-74,34,-13,87,-29,80,-23,-83,-42,39,99,-48,34,54,-66,-16,15,63,
47,11,-26,43,-14,-41,8,-56,-60,-9,-69,19,76,33,-73,77,-43,79,40,-97,
-40,5,56,-63,-60,21,-97,-96,39,-61,-65,-82,-2,-62,-37,5,14,-65,-29,-5,
-50,-18,-63,-39,73,-1,-61,61,31,-57,40,8,56,36,-24,-5,-88,60,-65,77,
98,-84,-86,77,37,87,-23,-26,-88,64,-91,-65,-34,36,-54,17,36,-68,-16,-69,
-32,-6,-13,-55,34,-20,-34,63,50,63,34,47,-15,-40,-80,-56,30,-96,92,71,
-21,6,9,57,3,86,-63,-31,-63,-15,62,35,50,-65,91,26,82,50,25,-82,
-63,-80,49,11,-13,-36,97,79,11,52,99,-54,49,9,-46,-53,1,-44,-58,-7,
-1,-73,-75,-42,-39,21,56,-63,-61,68,86,-60,93,-10,5,19,-66,-27,39,-54,
59,-66,-54,41,-58,41,-15,-73,18,34,-59,24,49,-23,83,25,3,10,-40,55,
-82,68,-55,42,57,-3,-45,76,95,-18,-56,41,3,1,13,-48,81,20,81,-90,
96,30,-37,-9,-53,-95,-60,67,-96,-18,-25,-42,5,94,20,-24,22,-58,-65,-34,
75,-65,-49,-48,-21,68,83,38,-6,-15,55,25,27,57,-10,39,-13,-69,-32,3,
22,84,-44,-14,-40,37,20,53,-4,-25,9,94,26,74,-70,-33,44,-29,-21,-27,
67,43,3,-48,-51,88,49,-43,76,-77,-76,-15,36,94,79,-12,-48,-42,-44,-26,
69,2,-73,-10,-24,-1,19,6,-49,97,35,-81,-49,-73,-37,81,61,49,74,97,
39,99,-18,38,-41,-85,61,-25,39,-2,29,87,35,-40,-100,84,90,51,47,-25,
67,-32,-82,13,88,23,-21,-96,-99,-14,53,-29,50,66,-33,56,91,-43,-74,-93,
-73,-60,-65,-35,67,-38,86,69,97,-80,55,-99,77,48,69,89,70,62,68,-13,
56,-8,7,17,9,-58,48,1,-91,-6,-50,6,61,-8,89,14,83,45,76,67,
68,-94,-95,87,-6,-5,54,25,-29,6,-41,-4,-73,-3,6,88,-61,7,-38,59,
96,61,-70,5,-63,-91,-91,22,50,61,64,30,76,-47,-67,98,36,16,43,-85,
51,25,56,-90,42,26,69,-20,6,92,-29,-63,-7,44,68,79,-89,-100,8,-5,
-82,-35,-94,62,-44,-47,22,90,94,-43,-94,-84,72,36,74,-43,-5,51,65,40,
90,-45,-98,-96,81,-42,-24,-69,-100,-81,-30,58,-49,66,-62,-15,-45,-48,-62,73,
-59,-98,-14,-87,-76,-9,9,83,-54,-1,16,-26,12,46,71,93,-21,93,72,-65,
-6,-24,-94,-29,-65,4,12,91,1,56,36,5,-62,42,60,-9,-99,60,-24,-83,
29,52,-90,5,-20,25,-22,-37,-81,84,-93,6,-67,75,90,13,97,-49,-40,-43,
-85,27,-47,94,-81,-87,40,-16,26,58,34,63,23,-28,-7,-56,12,64,79,-61,
66,-50,-21,-82,94,26,55,56,63,-49,-89,73,-21,67,44,100,44,96,10,23,
-93,22,2,-94,77,-2,-37,4,-35,-94,18,51,-28,-79,36,-35,-48,19,78,56,
-28,-4,54,35,80,-31,-30,32,32,-81,-79,-93,-86,-19,41,-99,-83,-1,-99,-29,
87,94,-64,39,73,-9,-72,52,38,-49,-98,-58,-10,64,81,30,-87,-25,55,-48,
-73,2,-2,-43,38,67,14,-48,-82,97,-64,-57,16,-95,-93,10,-55,25,-10,76,
-24,-58,-4,-59,-74,28,64,10,13,4,20,-57,46,-73,88,38,-37,-36,36,-62,
66,-16,-50,66,20,-78,-31,-97,-77,66,83,-95,-18,40,-32,5,38,45,-47,80,
-82,-74,-75,-17,81,54,-94,92,-16,-42,-69,-13,-77,-31,-29,-84,47,75,54,-83,
-95,-17,73,51,-80,38,73,92,-69,15,-26,-91,66,-92,23,38,-57,59,-26,-71,
62,54,27,22,-5,-68,-19,88,94,-64,89,-4,11,15,12,32,-19,44,64,66,
81,87,44,45,-88,61,14,79,-2,-48,43,89,10,91,-73,61,94,-100,90,-6,
-10,78,-11,-42,80,48,-83,79,-97,-3,-94,-24,-44,-61,-98,-95,22,39,-48,-48,
-96,-34,-51,12,-45,20,22,50,-34,-53,62,15,-92,-97,46,-31,-6,-53,-18,81,
23,7,-2,-12,-72,86,-31,45,90,31,-78,3,68,-79,7,-55,2,-59,36,-61,
67,100,5,-43,-89,-51,70,55,-43,-16,98,80,-58,-99,-76,-27,32,-61,38,45,
-65,93,-3,-55,96,95,-20,-82,68,-77,28,-16,-84,66,38,67,97,-21,-10,-48,
-96,-75,56,-65,13,73,54,14,-42,36,65,100,-69,-29,-76,61,-11,43,-45,-10,
76,93,64,85,-28,25,31,15,65,38,-40,94,17,15,75,-76,-47,72,25,-70,
-68,-11,5,-60,-76,96,-96,-30,91,35,-90,2,14,-73,55,-90,-89,-2,5,-64,
-91,82,-88,-27,13,75,-87,-3,-71,-88,27,-41,25,34,-3,30,8,100,-15,21,
66,-34,13,74,15,-36,-66,-2,-10,-71,27,42,23,-84,28,-29,36,-41,73,86,
-45,66,49,-24,82,-5,-55,61,38,96,-33,-44,-13,-51,21,90,-65,51,-9,42,
-82,56,19,27,58,-25,88,-36,-90,50,-66,15,81,-97,30,-65,54,-74,5,56,
90,-90,-3,-56,-63,48,-70,67,-100,36,-49,82,-46,-33,-50,-45,-45,-36,43,30,
-31,88,48,-30,-37,26,57,60,9,39,-47,68,36,37,24,66,-95,38,-69,69,
-6,58,98,43,-95,25,-40,-75,-80,97,18,-46,61,-94,52,22,-25,-46,70,55,
62,51,-27,98,98,-44,48,-34,38,81,-74,10,-16,-36,21,-96,-69,-7,-58
  };

  static short rand_pos = 0;

  // check from where RTP is available:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (rtp && jit_rtp_listen(c)) {
      rtp_in[c] = jit_get_audio(c);
    }
    else
      rtp_in[c] = 0;	
  }

  // update the filtered energy and calculate incoming RTP energy:
  clr_signal(buffer2);
  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (rtp_in[c]) add_signal(buffer2, rtp_in[c]);
  }
  energy_rc = 0; // local current energy
  for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
    const int sm = buffer2[i];
    energy_rc += (sm * sm) / 256;
  }

  energy_rf = energy_rf - (energy_rf / 4) + energy_rc / 4;
  //energy_rf = update_energy(energy_rf, energy_rc);

  // if half duplex mode, we sometimes have to use comfort noise to replace the audio.
  if(in && !play_dtmf) {
    if (half_duplex) {
      if(energy_rf > 100000) {
        duplex_state = HD_RECV;
      } else if(energy_lf > 100000) {
        duplex_state = HD_SEND;
      }
      
      if(energy_lf < 1000 && energy_rf < 1000){ //30000
        duplex_state = HD_SEND;
      }
      
      // check if we should turn the mic off:
      if (duplex_state == HD_RECV) {
        for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
          if (rand_pos >= sizeof(rand_data) / (sizeof(rand_data[0]))){
            rand_pos = 0;
          }
          in[i] = rand_data[rand_pos++]/2;
        }
      }
    }
    // full duplex: supress echo from the handset. The handset
    // normally response with a much lower volume (factor
    // 1000). Therefore we can make a quite strict checking.
    else { 
      static const int threshold_table[16] =
      {
        8 * 700, // 0
          8 * 800, // 1
          8 * 900, // 2
          8 * 1000, // 3
          8 * 1100, // 4
          8 * 1200, // 5
          8 * 1500, // 6
          8 * 1800, // 7
          16 * 2200, // 8
          16 * 2600, // 9
          16 * 3000, // 10
          16 * 3400, // 11
          16 * 3800, // 12
          16 * 4200, // 13
          16 * 4600, // 14
          16 * 5000, // 15
      };
      int threshold = threshold_table[handset_volume & 15];
      threshold = (threshold * (energy_rf / 0x100000)) / 0x18; // 0x1800000 is really very loud
      if(energy_rf < 30000 || energy_lg > threshold) {
        // the remote site is so silent so there is no echo OR the
        // local sound is so loud so that you cant really hear the
        // echo:
        duplex_state = HD_SEND;
      }
      else {
        const unsigned int factor = energy_lg > 0 ? energy_rf / energy_lg : 1000000;
        if(factor > 1000) {
          duplex_state = HD_RECV;
        }
      }
      
      // check if we should turn the mic off:
      if (duplex_state == HD_RECV) {
        for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
          if (rand_pos >= sizeof(rand_data) / (sizeof(rand_data[0]))){
            rand_pos = 0;
          }
          in[i] = rand_data[rand_pos++]/4;
        }
      }
    }
  } // in & !dtmf

  // generate speaker output:
  if (synth) set_signal(buffer, synth);
  else clr_signal(buffer);
  if (!half_duplex || duplex_state == HD_RECV) {
    // normal output mode:    
    for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
      buffer[i] += buffer2[i];
    }
  }
  if (echo && in) add_signal(buffer, in);
  clip_signal(out, buffer);
  
  // generate rtp streams:
  for(c = 0; c < MAX_NUM_RTP; c++) {
    int d;
    if (!jit_rtp_active(c)) continue; // this channel is not active
    
    // initialize the buffer with the microphone input:
    if (in) set_signal(buffer, in);
    else clr_signal(buffer);
    
    // mix the other channels:
    for(d = 0; d < MAX_NUM_RTP; d++) {
      if (c == d) continue; // not mixing to itself
      if (!rtp_in[d]) continue;
      add_signal(buffer, rtp_in[d]);
    } // for d
    
    clip_signal(clipped_buffer, buffer);
    set_rtp_audio(c, clipped_buffer);
  } // for c

  // generate recording data:
  if(read_recording) {
    int d;

    // initialize the buffer with the microphone input:
    if (in) set_signal(buffer, in);
    else clr_signal(buffer);

    for(d = 0; d < MAX_NUM_RTP; d++) {
      if (!rtp_in[d]) continue;
      add_signal(buffer, rtp_in[d]);
    } // for d

    clip_signal(&read_buffer[read_wpos], buffer);
    read_wpos += SNOM_AUD_BLOCK_SIZE;
    if(read_wpos >= READ_BUFFER_LENGTH) {
      read_wpos = 0; // wrap
    }
  }
}

// process the next block. This procedure is called when the input is
// available and the output needs to be refilled.
static void next_block(short *output, short *input, int echo) {
  int audio_size;
  if (au_wr_pos >= au_rd_pos) 
    audio_size = au_wr_pos - au_rd_pos;
  else
    audio_size = sizeof(au_buffer) / sizeof(au_buffer[0]) - (au_rd_pos - au_wr_pos);
  
  // buffer size hysteris to avoid fast starting:
  if (audio_state == AUDIO_STARTING) {
    if (audio_size < (3 * SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) / 4) audio_size = 0;
    else audio_state = AUDIO_RUNNING;
  }
  else {
    if (audio_size < SNOM_AUD_BLOCK_SIZE) audio_state = AUDIO_STARTING;
  }

  // generate the output:
  switch (mixer_mode) {
  case SNOM_AUD_MIXER_ONLY_RTP:
    generate_output(output, input, 1, 0, echo);
    break;
  case SNOM_AUD_MIXER_ONLY_AUDIO:
    if (audio_size >= SNOM_AUD_BLOCK_SIZE) {
      // if (au_rd_pos < 0 || au_rd_pos + SNOM_AUD_BLOCK_SIZE > SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) printk("au_rd_pos out of range\n");
      generate_output(output, 0, 0, &au_buffer[au_rd_pos], echo);
    }
    else
      generate_output(output, 0, 0, 0, echo);
    break;
  case SNOM_AUD_MIXER_RTP_AUDIO: 
    /* if rtp is available, take it; otherwise audio. This mode has
       been obsoleted! */
  case SNOM_AUD_MIXER_AUDIO_RTP:
    if (audio_size >= SNOM_AUD_BLOCK_SIZE) {
      // if (au_rd_pos < 0 || au_rd_pos + SNOM_AUD_BLOCK_SIZE > SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) printk("au_rd_pos out of range\n");
      generate_output(output, input, 1, &au_buffer[au_rd_pos], echo);
    }
    else
      generate_output(output, input, 1, 0, echo);
    break;
  default: /* SNOM_AUD_MIXER_NOTHING */
    generate_output(output, 0, 0, 0, echo);
    break;
  }

  /* advance the audio read position ANYWAY */
  if (audio_size > 0) {
    au_rd_pos += SNOM_AUD_BLOCK_SIZE;
    if (au_rd_pos >= sizeof(au_buffer) / sizeof(au_buffer[0])) au_rd_pos = 0;
  }
}

/* try to refill a rx and tx buffer */
static int refill_buffer(void) {
  static short tx_location[SNOM_AUD_BLOCK_SIZE];
  static short rx_location[SNOM_AUD_BLOCK_SIZE];
#ifdef CONFIG_GIPS
  ulong flags;
#endif
  volatile cbd_t *tx_bdp = phy.tx_bd_base;
  volatile cbd_t *rx_bdp = phy.rx_bd_base;
  int i;
  int retval = 0;

  /* refill both rx and tx buffer if both buffers are "empty" and "ready" */
  for (i = 0; i < HW_NUM_BLOCKS; i++) {
    int rx_avail = !(rx_bdp->cbd_sc & BD_SC_EMPTY);

    if (rx_avail) {
      unsigned char *rx = __va(rx_bdp->cbd_bufaddr);

      rx_bdp->cbd_sc = (rx_bdp->cbd_sc | BD_SC_EMPTY) & (~BD_SC_OV);

      local_time += SNOM_AUD_BLOCK_SIZE;

      retval = 1;

      if (play_dtmf) { // play synthesis sound
        // overwrite the microphone samples:
        get_dtmf(&internal_dtmf, rx_location);
        next_block(tx_location, rx_location, 1);
      }
      else {
        // read the microphone samples:
        read_from_hardware(rx_location, rx);
        if (do_agc)
          agc_data(rx_location);
        next_block(tx_location, rx_location, 0);
      }
    }
    rx_bdp++;
  }

  for (i = 0; i < HW_NUM_BLOCKS; i++) {
    int tx_avail = !(tx_bdp->cbd_sc & BD_SC_READY);

    if (tx_avail) {
      unsigned char *tx = __va(tx_bdp->cbd_bufaddr);
      tx_bdp->cbd_sc = (tx_bdp->cbd_sc | BD_SC_READY) & (~BD_SC_OV);
      write_to_hardware(tx, tx_location);
    } /* if buffer available */
    tx_bdp++;
  } /* for */

  return retval;
}

void snom_aud_do_tasklet(unsigned long unused) {
  /* check if we have a incoming rtp packet ready: */
  snom_aud_check_incoming();
  
  /* check if we have enough data to send a packet: */
  snom_aud_check_outgoing();
}

DECLARE_TASKLET(snom_aud_tasklet, snom_aud_do_tasklet, 0);

/*
 * This is the snom_aud driver's interrupt routine for a single port
 */
static void snom_aud_interrupt(void) {
  u_char events;
  volatile scc_t *sccp;
  int run_tasklet = 0;

  sccp = &cpmp->cp_scc[0];
  events = sccp->scc_scce;

  if (events & (SCCM_GLR | SCCM_GLT | SCCM_TXE | SCCM_BSY)) { /* optimize, since this seldomly happens */
    if (events & SCCM_TXE) {
      // printk("<t>");
      clear_buffers(1);
      delay_sending_commands = 1000;
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SCCM_BSY) {
      // printk("<b>");
      clear_buffers(1);
      delay_sending_commands = 1000;
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SCCM_GLR) {
      // printk("<g>");
      clear_buffers(1);
      if (refill_buffer()) run_tasklet = 1;
    }
    if (events & SCCM_GLT) {
      // printk("<l>");
      clear_buffers(1);
      if (refill_buffer()) run_tasklet = 1;
    }    
  }
  
  if (events & SCCM_RX) {
    if (refill_buffer()) run_tasklet = 1;
  }
  else if (events & SCCM_TX) {
    if (refill_buffer()) run_tasklet = 1;
  }
  
  /* check if we have to wake up interruptable processes */
  if (snom_aud_call_back) {
    snom_aud_call_back();
  }
    
  if (run_tasklet) {
    tasklet_schedule(&snom_aud_tasklet);
  }

  sccp->scc_scce = events; 
  // printk("</i>");
}


/*
 * -------------------------------------------------------------------
 * Here ends the snom_aud interrupt routines.
 * -------------------------------------------------------------------
 */

// calculate how many *bytes* can be written:
int snom_aud_write_avail(snom_aud_info_t *info) {
  unsigned long flags;
  int rem;
  save_flags(flags); cli();
  
  /* determine how many words are available in the buffer (keep one buffer spare) */
  if (au_wr_pos >= au_rd_pos)
    rem = sizeof(au_buffer) - (au_wr_pos - au_rd_pos) * sizeof(short) - SNOM_AUD_BLOCK_SIZE * sizeof(short);
  else
    rem = (au_rd_pos - au_wr_pos) * sizeof(short) - SNOM_AUD_BLOCK_SIZE * sizeof(short);  

  restore_flags(flags);
  // if (rem < 500) printk("<%d>", rem);
  return rem;
}

// calculate how many *bytes* can be read:
int snom_aud_read_avail(snom_aud_info_t *info) {
  unsigned long flags;
  int length;
  save_flags(flags); cli();
  
  // get the length:
  if(read_recording) {
    length = read_wpos - read_rpos;
    if(length < 0) {
      length += READ_BUFFER_LENGTH;
    }
  }
  else {
    length = 0;
  }
  
  restore_flags(flags);
  return length * sizeof(short);
}

/*
 * This routine is used to handle the "bottom half" processing for the
 * snom_aud driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * snom_aud_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using snom_aud_sched_event(), and they get done here.
 */

/* Only short words are written to the driver! (pcm16) */
int snom_aud_write(snom_aud_info_t *info, const unsigned char *buf, int count) {
  unsigned long flags;
  int rem = snom_aud_write_avail(info);
  int written = 0;

  if (rem < SNOM_AUD_BLOCK_SIZE) return -EAGAIN;
  if (rem < count) count = rem;

  save_flags(flags); cli();

  while(count >= SNOM_AUD_BLOCK_SIZE * 2) {
    // printk("w");
    // if (au_wr_pos < 0 || au_wr_pos + SNOM_AUD_BLOCK_SIZE > SW_NUM_BLOCKS * SNOM_AUD_BLOCK_SIZE) printk("au_rd_pos out of range\n");
    memcpy(&au_buffer[au_wr_pos], buf, SNOM_AUD_BLOCK_SIZE * 2);
    count -= SNOM_AUD_BLOCK_SIZE * 2;
    buf += SNOM_AUD_BLOCK_SIZE * 2;
    written += SNOM_AUD_BLOCK_SIZE * 2;
    au_wr_pos = (au_wr_pos + SNOM_AUD_BLOCK_SIZE) % (sizeof(au_buffer) / sizeof(au_buffer[0]));
  }

  restore_flags(flags);

  if (written > 0)
    return written;
  else
    return -EAGAIN;
}


int snom_aud_read(snom_aud_info_t *info, unsigned char *buf, int count) {
  int length;
  short *buffer = (short*)buf;
  unsigned long flags;

  // check if we are recording:
  if(!read_recording) return -EAGAIN;
  save_flags(flags); cli();

  // get the length:
  length = read_wpos - read_rpos;
  if(length < 0) {
    length += READ_BUFFER_LENGTH;
  }

  if(length == 0) {
    restore_flags(flags);
    return -EAGAIN;
  }

  // copy the data:
  while(read_wpos != read_rpos) {
    int i;
    short *r = &read_buffer[read_rpos];
    for(i = 0; i < SNOM_AUD_BLOCK_SIZE; i++) {
      *buffer++ = *r++;
    }
    read_rpos += SNOM_AUD_BLOCK_SIZE;
    if(read_rpos >= READ_BUFFER_LENGTH) {
      read_rpos = 0;
    }
  }

  // return:
  restore_flags(flags);
  return length * sizeof(short);
}

/*
 * ------------------------------------------------------------
 * snom_aud_close()
 * 
 * This routine is called when the snom_aud port gets closed.
 * ------------------------------------------------------------
 */
void snom_aud_close(snom_aud_info_t *p_info) {
  unsigned long flags;
  int c;
  save_flags(flags); cli();
	
  if(p_info == &info)
    info.is_open = 0;
	
  shutdown_cpm();

  for(c = 0; c < MAX_NUM_RTP; c++) {
    if (rtp_socket[c] != 0) {
      sock_release(rtp_socket[c]);
      rtp_socket[c] = 0;
    }
    if (rtcp_socket[c] != 0) {
      sock_release(rtcp_socket[c]);
      rtcp_socket[c] = 0;
    }
    packet_size[c] = 160;
    packet_type[c] = 0;
    need_marker[c] = 0;
    local_port[c] = 0;
    dtmf_key[c] = 0;
    dtmf_shutoff[c] = 0;
    dtmf_duration[c] = 0;
    dtmf_state[c] = DTMF_STATE_NO_DTMF;
  }
  mixer_mode = SNOM_AUD_MIXER_RTP_AUDIO;

  restore_flags(flags);
}

/*
 * ------------------------------------------------------------
 * snom_aud_open() and friends
 * ------------------------------------------------------------
 */

/*
 * This routine is called whenever a snom_aud port is opened.
 */
int snom_aud_open(int minor_num, snom_aud_info_t **pp_info, int mode, int f_flags) {
  unsigned long flags;
  int ret_val;
	
  save_flags(flags); cli();
	
  if(minor_num != 0 && minor_num  != 2) {
    ret_val = -EINVAL;
    goto ret_open;
  }

  if((minor_num == 2) && (mode != MODE_WR)) {
    ret_val = -EINVAL;
    goto ret_open;
  }

  *pp_info = &info;
	
  /* Busy ? */
  if((*pp_info)->is_open) {
    ret_val = -EBUSY;
    goto ret_open;
  }
	
  (*pp_info)->is_open = 1;
  snom_aud_mode	      = mode;
  snom_aud_minor_num  = minor_num;

  shutdown_cpm();
  restart_cpm(CPM_RXTX);
  ret_val = 0;
	
 ret_open:
  restore_flags(flags);
  return ret_val;
}

/*
 * ---------------------------------------------------------------------
 * snom_aud_init() is called at boot-time to initialize the snom_aud driver.
 * ---------------------------------------------------------------------
 */


/*
 * This routine prints out the appropriate snom_aud driver version
 * number, and identifies which options were configured into this
 * driver.
 */
static _INLINE_ void show_snom_aud_version(void) {
  printk(KERN_INFO "%s $Revision: 1.98 $ $Date: 2004/09/22 17:44:21 $\n", snom_aud_name);
}

/*
 * Initializes the AC97
 */
static void reset_ac97 (void) {
  int j;
  // enable power for 24 MHz (VCC) and stop SYNC counter(CLK_EN_48KHZ) and deactivate handsfree mode:
  set_port(PORT_SNOM_CSR, VCC | CLK_EN_48KHZ, 0);
  set_port(PORT_SNOM_CSR, 0, MIC_CONTROL);//low = handset
  // deactivate WARM_RESET warm reset 
  set_port(PORT_SNOM_CSR, 0, WARM_RESET);

  // Set the port directions:
  set_port_dir(PORT_SNOM_CSR, OUTPUT_PINS, VCC | MIC_CONTROL | CLK_EN_48KHZ |WARM_RESET);
  #define PC12 (1<<3)
  set_port_dir(PORT_C, INPUT_PINS, PC12);// not to disturb headset spk audio

  // release AC97 reset:
  set_port(PORT_C, AC97_RST, 0);
  set_port_dir(PORT_C, OUTPUT_PINS, AC97_RST);
  set_port(PORT_B, 0,HEADSET_LED); //low, NOT CONNECTED
  set_port_dir(PORT_B, OUTPUT_PINS, HEADSET_LED);

  // Perform hard reset:
  for (j = 0; j < 10000; j++);
  set_port(PORT_C, 0, AC97_RST); // reset
  set_port(PORT_SNOM_CSR, 0, CLK_EN_48KHZ); // reset counter
  for (j = 0; j < 1000; j++);
  set_port(PORT_SNOM_CSR, CLK_EN_48KHZ, 0); // restart counter
  for (j = 0; j < 10000; j++);
  set_port(PORT_C, AC97_RST, 0); // end reset
  for (j = 0; j < 1000000; j++);
}

static void init_ac97(void) {
  //configuration commands:
  set_ac97_reg(0x26, 0x0); // all powered 
  set_ac97_reg(0x2a, 0x1); // enables variable rate audio

  //set_ac97_reg(0x2c, 0x3e80); // 16khz dac
  //set_ac97_reg(0x32,0x1f40);// 8Khz adc
  set_ac97_reg(0x2c, 0x1f40); // 8khz dac
  set_ac97_reg(0x32,0x3e80); // 16Khz adc, better transmision
  //set_ac97_reg(0x26,ac97_headset_mic); //headset mic from RJ14


  //audio commands:
  set_ac97_reg(0x02, ac97_handset); // main amp is muted, max. gain  
  set_ac97_reg(0x0e, ac97_mic);//  no mic boost, (sidetone = on, value 1f)   
  set_ac97_reg(0x1c,ac97_adc); // 0x0006 is starting value
  set_ac97_reg(0x18, ac97_dac); // PCM is an input for mux, gain 1/2
  set_ac97_reg(0x06,ac97_speaker); //mono aut max. mute on  
	       
  //set_ac97_reg(0x04, 0x0000); // alt out enabled
  //set_ac97_reg((0x7c | 0x80),0); //read r 0x7c, vendor id1 
  //set_ac97_reg(0x5E, 0x0100); // AC mode control, da -> main out, no mux
  //set_ac97_reg(0x26, 0x0f00); // adc, pcm analog mix power down
  // set_ac97_reg(0x20, 0x0080); // loop back
 
  set_port_dir(PORT_C, OUTPUT_PINS, SHUTDOWN_SPEAKER_PIN);
}

static void port_init(void) {
  // get the serial number
  register int serial_num;
  register unsigned char *puc = (unsigned char *) 0xfe030000; 
  puc += 3; 
  serial_num = *puc; 
  serial_num <<= 8; 
  puc++; 
  serial_num |= *puc; 
  serial_num <<= 8; 
  puc++; 
  serial_num |= *puc;
  global_serial_num = serial_num;
}

/*
 * This function is used to configure the second time multiplexed channel 
 * (TDMb) of the time slot assigner (TSA) of the serial interface (SI) to
 * route out data to the second serial management controller (SCC1). It
 * also configures the port pins to do their special function.
 */
// tsa is programed to take 12 bytes (96 bits) and ignore rest (up to 256 bits)
// sample is one bit late form sync
static void tsa_init(void)
{
  volatile	cpm8xx_t	*cp;
  volatile	immap_t		*immap;
  volatile      unsigned long   *si_ram;
  int i;

  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* and to internal registers */
	
  immap = (immap_t *)IMAP_ADDR;	
	
  /* Program SI RAM to route first 96 bits (12 bytes) to SCC1 and to
     ignore the rest. */
  
  /* get 32 bit pointer */
	
  si_ram = (unsigned long *) &(cp->cp_siram);

  /* reset the memory */
  for(i = 0; i < 128; i++) si_ram[i] = 0x00010000L;
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = SCC1, 12 byte
     CNT = 11 and BYT = 1 and NOT last entry for both RXb and TX000000b entry.
  */
	
  si_ram[32] = (1L << 22) | (11L << 18) | (1L << 17);
  si_ram[96] = (1L << 22) | (11L << 18) | (1L << 17);

  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 16 byte CNT = 15 and BYT = 1 and NOT last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[33] = (0L << 22) | (15L << 18) | (1L << 17);
  si_ram[97] = (0L << 22) | (15L << 18) | (1L << 17);
	
  /* normal mode (no loop), no swtr, SSEL = none, CSEL = time slot not 
     used, 4 byte CNT = 3 and BYT = 1 and is last entry for both 
     RXb and TXb entry.
  */
	
  si_ram[34] = (0L << 22) | (3L << 18) | (1L << 17) | (1L << 16);
  si_ram[98] = (0L << 22) | (3L << 18) | (1L << 17) | (1L << 16);

  /* Set the SIMODE register to SCC1 connected to SI, normal mode,
     RFSDx = 00 0 bit delay, DSCx = 0 normal speed, common receive and
     transmit clocks CRTb = 1, STZb = 0 normal, CE = 0 data sent on
     rising and received on falling clock edge, FEb = 0 frame sync edge
     falling edge, GMb = 0, TFSDb = 00 0 bit delay.
  */
	
  cp->cp_simode = 0x01410000 | (cp->cp_simode & 0xffff);
  // cp->cp_simode = 0x05400000 | (cp->cp_simode & 0xffff);
  cp->cp_sicr = 0x40 | (cp->cp_sicr & 0xffffff00);

  /* Configure PA2 as L1RCLKB, PA10 as L1RXBD, PA11 as L1TXDB and
     PC6 as L1RSYNCB 
  */
	
  immap->im_ioport.iop_papar |=  0x2030; /* use PA2, PA10 and PA11 as SPR */
  immap->im_ioport.iop_padir |=  0x0030; /* L1TXDB, L1RXDB, L1RCLKB, from table 34-1 */
  immap->im_ioport.iop_padir &= ~0x2000; /* make PA2 and PA10 input */
  immap->im_ioport.iop_pcpar |=  0x0200; /* use PC6 as L1RSZNCB, see table 34-11 */
	
  /* Set SIGMR to enable TDMb, static TDMa and TDMb RAM division */
	
  cp->cp_sigmr = 0x0a;
}

/* 
 * This routine configures the SCC1 to transparent mode, sets up the 
 * required buffers and their descriptors, resets the event mask and 
 * allocates and activates the interrupts.
 */
static void scc1_init(void) {
  int                 j;
  uint                mem_addr, dp_addr;
  volatile immap_t    *immap;
  volatile cbd_t      *bdp;
  volatile scc_t      *sp; // conf. registers for scc
  volatile sccp_t     *up; // parameter ram structure for scc
  volatile cpm8xx_t   *cp; // whole cpm structure
	
  /* Get pointer to Communication Processor */
	
  cp = cpmp;
	
  /* addr internal registers */
	
  immap = (immap_t *)IMAP_ADDR;
	
  memset((void *) &phy, 0, sizeof(snom_aud_physical_dev)); /* clear buffers */
	
  /* We need to allocate a transmit and receive buffer
   * descriptors from dual port ram, and a character
   * buffer area from host mem.
   */

  dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * HW_NUM_BLOCKS);
  /* Allocate space for FIFOs in the host memory. */

  mem_addr = m8xx_cpm_hostalloc(HW_NUM_BLOCKS * CPM_BUF_SIZE);
  memset((void *) mem_addr, 0, HW_NUM_BLOCKS * CPM_BUF_SIZE);
	
  /* Set the physical address of the host memory
   * buffers in the buffer descriptors, and the
   * virtual address for us to work with.
   */
	
  bdp = (cbd_t *)&cp->cp_dpmem[dp_addr];
  phy.rx_cur = phy.rx_bd_base = (cbd_t *)bdp;
	
  for (j=0; j<(HW_NUM_BLOCKS-1); j++) {
    bdp->cbd_bufaddr = __pa(mem_addr);
    bdp->cbd_sc = (BD_SC_EMPTY | BD_SC_INTRPT);
    bdp->cbd_datlen = CPM_BUF_SIZE;
    mem_addr += CPM_BUF_SIZE;
    bdp++;
  }
	
  bdp->cbd_bufaddr = __pa(mem_addr);
  bdp->cbd_sc = (BD_SC_WRAP | BD_SC_EMPTY | BD_SC_INTRPT);
  bdp->cbd_datlen = CPM_BUF_SIZE;
	
  sp = &cp->cp_scc[0];
  up = (sccp_t *)&cp->cp_dparam[PROFF_SCC1];
  up->scc_rbase = dp_addr;
	
  dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * HW_NUM_BLOCKS);
  /* Allocate space for FIFOs in the host memory.
   */
  mem_addr = m8xx_cpm_hostalloc(HW_NUM_BLOCKS * CPM_BUF_SIZE);
  memset((void *) mem_addr, 0, HW_NUM_BLOCKS * CPM_BUF_SIZE);
	
  /* Set the physical address of the host memory
   * buffers in the buffer descriptors, and the
   * virtual address for us to work with.
   */
  bdp = (cbd_t *)&cp->cp_dpmem[dp_addr];
  phy.tx_cur = phy.tx_bd_base = (cbd_t *)bdp;
	
  for (j=0; j<(HW_NUM_BLOCKS-1); j++) {
    bdp->cbd_bufaddr = __pa(mem_addr);
    bdp->cbd_sc = BD_SC_READY;
    bdp->cbd_datlen = CPM_BUF_SIZE; 
    mem_addr += CPM_BUF_SIZE;
    bdp++;
  }
  bdp->cbd_bufaddr = __pa(mem_addr);
  bdp->cbd_sc = BD_SC_READY | BD_SC_WRAP;
  bdp->cbd_datlen = CPM_BUF_SIZE;
	
  up->scc_tbase = dp_addr;
	
  /* Set up the uart parameters in the parameter ram. */
	
  up->scc_rfcr = SCC_EB;
  up->scc_tfcr = SCC_EB;
	
  /* Set this to 1 for now, so we get single character interrupts.
     Using idle charater time requires some additional tuning.
  */
	
  up->scc_mrblr = CPM_BUF_SIZE;
	
  /* Send the CPM an initialize command. */
	
  while (cp->cp_cpcr & CPM_CR_FLG);
  cp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_INIT_TRX) | CPM_CR_FLG;
  while (cp->cp_cpcr & CPM_CR_FLG);
	
  /* Set SDMAs' U-bus arbitration priority level to level 5 which
     is normal operation. */
	
  immap->im_siu_conf.sc_sdcr = 1;
	
  /* Set transparent mode, reversed data, normal operation (i.e.
     no loopback) and do NOT enable interrupts.
  */
       
  sp->scc_gsmrh =
    /* total transparent mode */
    SCC_GSMRH_TRX | SCC_GSMRH_TTX | 
    /* inherent synchronization, see 29.4.2.2 */
    SCC_GSMRH_CDP | SCC_GSMRH_CTSP | SCC_GSMRH_CDS | SCC_GSMRH_CTSS |
    /* transmit is synchronous to receiving */
    // SCC_GSMRH_TXSY |
    /* reverse byte order: */
    SCC_GSMRH_REVD;
  sp->scc_gsmrl = 
    /* no adjustments to sample point */
    SCC_GSMRL_EDGE_NONE; 

  /* Enable all interrupts and clear all pending events. */
	
  sp->scc_sccm |=  SCCM_TXE | SCCM_BSY | SCCM_TX | SCCM_RX; //| SCCM_GLT SCCM_GLR  |
  sp->scc_scce = 0xffff;
	
  /* Install interrupt handler. */
	
  cpm_install_handler(CPMVEC_SCC1, (void *) snom_aud_interrupt, NULL);
}

static void init_structures(void) {
  int i, j;

  /* init the mem structures */
  for(j = 0; j < MAX_NUM_RTP; j++) {
    dtmf_recv_type[j] = 101;
#if 0
    for(i = 0; i < JITTER_SIZE; i++) {
      memset(jit_recv_buffer[j][i], 0, sizeof(jit_recv_buffer[0][0]));
    }
    
    memset(jit_decode_buffer[j], 0, sizeof(jit_decode_buffer[0]));
#endif

    rtp_socket[j] = 0;
    rtcp_socket[j] = 0;
    local_port[j] = 0;
    packet_size[j] = 0;
    packet_type[j] = 0;
    local_ssrc[j] = 0;
    need_marker[j] = 2;
    packet_seqnr[j] = 0;
    remote_adr[j].sin_addr.s_addr = 0;
    remote_adr[j].sin_port = 0;
    dtmf_recv_ssrc_valid[j] = 0;
    ice_contacts[j].length = 0;
    sym_rtp[j] = 0;

    jit_decode_pos[j] = 0;
    jit_decode_len[j] = 0;
    jit_decode_start[j] = 1;

    for(i = 0; i < JITTER_SIZE; i++) {
      jit_recv_len[j][i] = 0;
      jit_recv_type[j][i] = 0;
      jit_recv_timestamp[j][i] = 0;
    }

    jit_send_len[j] = 0;

    jit_encoded_len[j] = 0;
    jit_unencoded_len[j] = 0;

    external_dtmf[j].key = 0;
    external_dtmf[j].pos0 = 0;
    external_dtmf[j].pos1 = 0;
  }

  internal_dtmf.key = 0;
  internal_dtmf.pos0 = 0;
  internal_dtmf.pos1 = 0;
}

int __init snom_aud_init(void) {
  extern void snomcodec_init(void);
  
#if 0
  printk("snom_aud_init:\n");
#endif
	
  show_snom_aud_version();
  init_structures();
 
  reset_ac97();
  tsa_init();
  scc1_init();
  init_ac97();
  port_init();
  hardware_platform();
  snomcodec_init();

  return 0;
}

#endif // CONFIG_SOUND_SNOM || CONFIG_SOUND_SNOM2 
