/*
 *  arch/ppc/8xx_io/snomdisp200.c
 *
 *  Copyright (C) 2002, snom technology AG
 *
 *  This code is for the LCD module connected to the SNOM 200 board.
 *
 *  SNOM 200: it implements only write to the display and no light. 
 * 
 */

#include <linux/config.h>
#include <linux/types.h>

#ifdef CONFIG_SNOM_DISP

#include <string.h>

#include <asm/port.h>
#include <asm/mpc8xx.h>

#include "commproc.h"
#include "snomaud.h"

static int contrast_val;
extern void hardware_platform(void);

int global_serial_num;

/*****************************************************************
This driver is used to connect the  CS290 DISPLAY (CONTR. SSD1815B)
 to the SNOM  200 board and  thus be able to configure it.  
 

 The data lines are expect to be connected as follows:

28 CS   - PB27 active low
27 RST  - PB29 active low
26 D/C  - PB31 
25 WR   - PB30 make low to write
24 RD   - PC4  make low for read (which we never do)
23 DB0  - PB14 
22 DB1  - PB15 
21 DB2  - PB16 
20 DB3  - PB17 
19 DB4  - PB18 
18 DB5  - PB22 
17 DB6  - PB23 
16 DB7  - PB26 
02 C68/80 -  Gnd

for other signals on disp con. see sch ip5_disp+kdb_2V1
**************************************************************/

#define DELAY		for(j=0;j<50;j++);
#define RST_DELAY	for(j=0;j<2500;j++);
#define READ_DELAY      for(j=0;j<10;j++); //200ns
extern int status_read(void);
extern int display_status =0;

/* The data bus and these lines are located on port B. */

#define DB0_B             (1<<17)
#define DB1_B             (1<<16)
#define DB2_B             (1<<15)
#define DB3_B             (1<<14)
#define DB4_B             (1<<13)
#define DB5_B             (1<<9)
#define DB6_B             (1<<8)
#define DB7_B             (1<<5)

#define CS_B              (1<<4)
#define RST_B             (1<<2)
#define WR_B              (1<<1)
#define DC_B              (1<<0)

/* This is on port C. */

#define RD_C              (1<<11)

/* Command values */

#define SET_LOWER_COL          0x00
#define SET_HIGH_COL           0x10
#define SET_INIT_RES_RATIO     0x23
#define SET_POWER_ON           0x2F
#define SET_START_LINE         0x40
#define SET_CONT_CTRL_REG      0x81
#define SET_SEG_REMAP_OFF      0xA0
#define SET_ALTERNATE_BIAS     0xA3
#define SET_DISPLAY_OFF        0xAE
#define SET_DISPLAY_ON         0xAF
#define SET_PAGE_ADDR          0xB0
#define SET_COM_REMAP_OFF      0xC0
#define SET_READ_MOD_WRITE     0xE0
#define SET_END_READ_MOD_WRITE 0xEE
#define SET_BIAS_CTRL_REG      0xA9
#define SET_1_4_BIAS_RATIO     0xAB

#define SET_EXT_COMMAND        0xA9

#define SET_INIT_CONTRAST_A      0x26
#define SET_INIT_CONTRAST_B      0x20


/* Character table */

static unsigned char const char_code[256 * 5] =
{
0x00, 0x00, 0x00, 0x00, 0x00,  //   00    0
0x1C, 0x14, 0x1C, 0x22, 0x7F,  //   01    1    Loudspeaker
0x80, 0xC8, 0xA5, 0x13, 0x01,  //   02    2    Separator
0x10, 0x60, 0x10, 0x08, 0x04,  //   03    3    Check
0x18, 0x24, 0x18, 0x24, 0x18,  //   04    4    Infinity
0x00, 0x00, 0x00, 0x00, 0x00,  //   05    5
0x00, 0x00, 0x00, 0x00, 0x00,  //   06    6
0x00, 0x00, 0x00, 0x00, 0x00,  //   07    7
0x00, 0x00, 0x00, 0x00, 0x00,  //   08    8
0x00, 0x00, 0x00, 0x00, 0x00,  //   09    9
0x00, 0x00, 0x00, 0x00, 0x00,  //   0A    10
0x00, 0x00, 0x00, 0x00, 0x00,  //   0B    11
0x00, 0x00, 0x00, 0x00, 0x00,  //   0C    12
0x00, 0x00, 0x00, 0x00, 0x00,  //   0D    13
0x00, 0x00, 0x00, 0x00, 0x00,  //   0E    14
0x00, 0x00, 0x00, 0x00, 0x00,  //   0F    15
0x7F, 0x3E, 0x1C, 0x08, 0x08,  //   10    16   
0x08, 0x08, 0x1C, 0x3E, 0x7F,  //   11    17   
0x14, 0x22, 0x7F, 0x22, 0x14,  //   12    18   
0x00, 0x4F, 0x00, 0x4F, 0x00,  //   13    19   
0x06, 0x0F, 0x7F, 0x01, 0x7F,  //   14    20   
0x0A, 0x55, 0x55, 0x55, 0x28,  //   15    21   
0x60, 0x60, 0x60, 0x60, 0x60,  //   16    22   
0x40, 0x52, 0x7F, 0x52, 0x40,  //   17    23   
0x04, 0x02, 0x7F, 0x02, 0x04,  //   18    24   
0x10, 0x20, 0x7F, 0x20, 0x10,  //   19    25   Arrow down
0x08, 0x08, 0x2A, 0x1C, 0x08,  //   1A    26   Arrow up
0x08, 0x1C, 0x2A, 0x08, 0x08,  //   1B    27   
0x0C, 0x08, 0x08, 0x08, 0x08,  //   1C    28   
0x08, 0x1C, 0x08, 0x1C, 0x08,  //   1D    29   
0x60, 0x78, 0x7E, 0x78, 0x60,  //   1E    30   
0x06, 0x1E, 0x7E, 0x1E, 0x06,  //   1F    31   
0x00, 0x00, 0x00, 0x00, 0x00,  //   20    32    
0x00, 0x00, 0x4F, 0x00, 0x00,  //   21    33   !
0x00, 0x07, 0x00, 0x07, 0x00,  //   22    34   "
0x14, 0x7F, 0x14, 0x7F, 0x14,  //   23    35   #
0x24, 0x2A, 0x7F, 0x2A, 0x12,  //   24    36   $
0x23, 0x13, 0x08, 0x64, 0x62,  //   25    37   %
0x36, 0x49, 0x55, 0x22, 0x50,  //   26    38   &
0x00, 0x05, 0x03, 0x00, 0x00,  //   27    39   '
0x00, 0x1C, 0x22, 0x41, 0x00,  //   28    40   (
0x00, 0x41, 0x22, 0x1C, 0x00,  //   29    41   )
0x14, 0x08, 0x3E, 0x08, 0x14,  //   2A    42   *
0x08, 0x08, 0x3E, 0x08, 0x08,  //   2B    43   +
0x00, 0xA0, 0x60, 0x00, 0x00,  //   2C    44   ,
0x08, 0x08, 0x08, 0x08, 0x08,  //   2D    45   -
0x00, 0x60, 0x60, 0x00, 0x00,  //   2E    46   .
0x20, 0x10, 0x08, 0x04, 0x02,  //   2F    47   /
0x3E, 0x51, 0x49, 0x45, 0x3E,  //   30    48   0
0x00, 0x42, 0x7F, 0x40, 0x00,  //   31    49   1
0x42, 0x61, 0x51, 0x49, 0x46,  //   32    50   2
0x21, 0x41, 0x45, 0x4B, 0x31,  //   33    51   3
0x18, 0x14, 0x12, 0x7F, 0x10,  //   34    52   4
0x27, 0x45, 0x45, 0x45, 0x39,  //   35    53   5
0x3C, 0x4A, 0x49, 0x49, 0x31,  //   36    54   6
0x01, 0x71, 0x09, 0x05, 0x03,  //   37    55   7
0x36, 0x49, 0x49, 0x49, 0x36,  //   38    56   8
0x06, 0x49, 0x49, 0x29, 0x1E,  //   39    57   9
0x00, 0x36, 0x36, 0x00, 0x00,  //   3A    58   :
0x00, 0x56, 0x36, 0x00, 0x00,  //   3B    59   ;
0x08, 0x14, 0x22, 0x41, 0x00,  //   3C    60   <
0x14, 0x14, 0x14, 0x14, 0x14,  //   3D    61   =
0x00, 0x41, 0x22, 0x14, 0x08,  //   3E    62   >
0x02, 0x01, 0x51, 0x09, 0x06,  //   3F    63   ?
0x3E, 0x41, 0x4D, 0x53, 0x1E,  //   40    64   @
0x7E, 0x11, 0x11, 0x11, 0x7E,  //   41    65   A
0x7F, 0x49, 0x49, 0x49, 0x36,  //   42    66   B
0x3E, 0x41, 0x41, 0x41, 0x22,  //   43    67   C
0x7F, 0x41, 0x41, 0x22, 0x1C,  //   44    68   D
0x7F, 0x49, 0x49, 0x49, 0x41,  //   45    69   E
0x7F, 0x09, 0x09, 0x09, 0x01,  //   46    70   F
0x3E, 0x41, 0x49, 0x49, 0x7A,  //   47    71   G
0x7F, 0x08, 0x08, 0x08, 0x7F,  //   48    72   H
0x00, 0x41, 0x7F, 0x41, 0x00,  //   49    73   I
0x20, 0x40, 0x41, 0x3F, 0x01,  //   4A    74   J
0x7F, 0x08, 0x14, 0x22, 0x41,  //   4B    75   K
0x7F, 0x40, 0x40, 0x40, 0x40,  //   4C    76   L
0x7F, 0x02, 0x0C, 0x02, 0x7F,  //   4D    77   M
0x7F, 0x04, 0x08, 0x10, 0x7F,  //   4E    78   N
0x3E, 0x41, 0x41, 0x41, 0x3E,  //   4F    79   O
0x7F, 0x09, 0x09, 0x09, 0x06,  //   50    80   P
0x3E, 0x41, 0x51, 0x21, 0x5E,  //   51    81   Q
0x7F, 0x09, 0x19, 0x29, 0x46,  //   52    82   R
0x46, 0x49, 0x49, 0x49, 0x31,  //   53    83   S
0x01, 0x01, 0x7F, 0x01, 0x01,  //   54    84   T
0x3F, 0x40, 0x40, 0x40, 0x3F,  //   55    85   U
0x1F, 0x20, 0x40, 0x20, 0x1F,  //   56    86   V
0x3F, 0x40, 0x38, 0x40, 0x3F,  //   57    87   W
0x63, 0x14, 0x08, 0x14, 0x63,  //   58    88   X
0x07, 0x08, 0x70, 0x08, 0x07,  //   59    89   Y
0x61, 0x51, 0x49, 0x45, 0x43,  //   5A    90   Z
0x00, 0x7F, 0x41, 0x41, 0x00,  //   5B    91   [
0x02, 0x04, 0x08, 0x10, 0x20,  //   5C    92   Backslash
0x00, 0x41, 0x41, 0x7F, 0x00,  //   5D    93   ]
0x04, 0x02, 0x01, 0x02, 0x04,  //   5E    94   ^
0x40, 0x40, 0x40, 0x40, 0x40,  //   5F    95   _
0x00, 0x03, 0x05, 0x00, 0x00,  //   60    96   Apostrophe
0x20, 0x54, 0x54, 0x54, 0x78,  //   61    97   a
0x7F, 0x48, 0x44, 0x44, 0x38,  //   62    98   b
0x38, 0x44, 0x44, 0x44, 0x20,  //   63    99   c
0x38, 0x44, 0x44, 0x48, 0x7F,  //   64    100  d
0x38, 0x54, 0x54, 0x54, 0x18,  //   65    101  e
0x08, 0xFE, 0x09, 0x01, 0x02,  //   66    102  f
0x18, 0xA4, 0xA4, 0xA4, 0x7C,  //   67    103  g
0x7F, 0x08, 0x04, 0x04, 0x78,  //   68    104  h
0x00, 0x44, 0x7D, 0x40, 0x00,  //   69    105  i
0x40, 0x80, 0x84, 0xFD, 0x00,  //   6A    106  j
0x7F, 0x10, 0x28, 0x44, 0x00,  //   6B    107  k
0x00, 0x41, 0x7F, 0x40, 0x00,  //   6C    108  l
0x7C, 0x04, 0x38, 0x04, 0x78,  //   6D    109  m
0x7C, 0x08, 0x04, 0x04, 0x78,  //   6E    110  n
0x38, 0x44, 0x44, 0x44, 0x38,  //   6F    111  o
0xFC, 0x24, 0x24, 0x24, 0x18,  //   70    112  p
0x18, 0x24, 0x24, 0x24, 0xFC,  //   71    113  q
0x7C, 0x08, 0x04, 0x04, 0x00,  //   72    114  r
0x48, 0x54, 0x54, 0x54, 0x20,  //   73    115  s
0x04, 0x3F, 0x44, 0x40, 0x20,  //   74    116  t
0x3C, 0x40, 0x40, 0x20, 0x7C,  //   75    117  u
0x1C, 0x20, 0x40, 0x20, 0x1C,  //   76    118  v
0x3C, 0x40, 0x30, 0x40, 0x3C,  //   77    119  w
0x44, 0x28, 0x10, 0x28, 0x44,  //   78    120  x
0x1C, 0xA0, 0xA0, 0xA0, 0x7C,  //   79    121  y
0x44, 0x64, 0x54, 0x4C, 0x44,  //   7A    122  z
0x00, 0x08, 0x36, 0x41, 0x00,  //   7B    123  {
0x00, 0x00, 0x7F, 0x00, 0x00,  //   7C    124  |
0x00, 0x41, 0x36, 0x08, 0x00,  //   7D    125  }
0x08, 0x04, 0x08, 0x08, 0x04,  //   7E    126  ~
0x70, 0x48, 0x44, 0x48, 0x70,  //   7F    127  
0x1E, 0xA1, 0xE1, 0x21, 0x12,  //   80    128  
0x38, 0x41, 0x40, 0x41, 0x78,  //   81    129  ue
0x38, 0x54, 0x56, 0x55, 0x18,  //   82    130  
0x20, 0x56, 0x55, 0x56, 0x78,  //   83    131  
0x20, 0x55, 0x54, 0x55, 0x78,  //   84    132  ae
0x20, 0x55, 0x56, 0x54, 0x78,  //   85    133  
0x20, 0x57, 0x55, 0x57, 0x78,  //   86    134  
0x18, 0xA4, 0xA4, 0xE4, 0x28,  //   87    135  
0x38, 0x56, 0x55, 0x56, 0x18,  //   88    136  
0x38, 0x55, 0x54, 0x55, 0x18,  //   89    137  
0x38, 0x55, 0x56, 0x54, 0x18,  //   8A    138  
0x00, 0x45, 0x7C, 0x41, 0x00,  //   8B    139  
0x00, 0x46, 0x7D, 0x42, 0x00,  //   8C    140  
0x00, 0x45, 0x7E, 0x40, 0x00,  //   8D    141  
0x78, 0x15, 0x14, 0x15, 0x78,  //   8E    142  Ae
0x78, 0x17, 0x15, 0x17, 0x78,  //   8F    143  
0x7C, 0x54, 0x56, 0x55, 0x44,  //   90    144  
0x74, 0x54, 0x78, 0x54, 0x58,  //   91    145  
0x7E, 0x09, 0x7F, 0x49, 0x41,  //   92    146  
0x30, 0x4A, 0x49, 0x4A, 0x30,  //   93    147  
0x30, 0x49, 0x48, 0x49, 0x30,  //   94    148  oe
0x30, 0x49, 0x4A, 0x48, 0x30,  //   95    149  
0x38, 0x42, 0x41, 0x20, 0x78,  //   96    150  
0x38, 0x41, 0x42, 0x20, 0x78,  //   97    151  
0x1C, 0xA1, 0xA0, 0xA1, 0x7C,  //   98    152  
0x38, 0x45, 0x44, 0x45, 0x38,  //   99    153  
0x3C, 0x41, 0x40, 0x41, 0x3C,  //   9A    154  Ue
0x1C, 0x22, 0x7F, 0x22, 0x22,  //   9B    155  Cent
0x48, 0x7E, 0x49, 0x41, 0x42,  //   9C    156  
0x15, 0x16, 0x7C, 0x16, 0x15,  //   9D    157  Yen
0x3F, 0x05, 0x17, 0x78, 0x50,  //   9E    158  Pt
0x40, 0x48, 0x3E, 0x09, 0x01,  //   9F    159  
0x20, 0x54, 0x56, 0x55, 0x78,  //   A0    160  
0x00, 0x44, 0x7E, 0x41, 0x00,  //   A1    161  
0x30, 0x48, 0x4A, 0x49, 0x30,  //   A2    162  
0x38, 0x40, 0x42, 0x21, 0x78,  //   A3    163  
0x7A, 0x11, 0x0A, 0x0A, 0x71,  //   A4    164  
0x7A, 0x09, 0x12, 0x22, 0x79,  //   A5    165  
0x4C, 0x55, 0x55, 0x55, 0x5E,  //   A6    166  
0x4E, 0x51, 0x51, 0x51, 0x4E,  //   A7    167  
0x30, 0x48, 0x45, 0x40, 0x20,  //   A8    168  
0x38, 0x08, 0x08, 0x08, 0x08,  //   A9    169  like 170, mirrored
0x08, 0x08, 0x08, 0x08, 0x38,  //   AA    170  
0x17, 0x08, 0x94, 0xD0, 0xB0,  //   AB    171  
0x17, 0x08, 0x74, 0x40, 0xF0,  //   AC    172  
0x00, 0x00, 0x79, 0x00, 0x00,  //   AD    173  
0x08, 0x14, 0x2A, 0x14, 0x22,  //   AE    174  
0x22, 0x14, 0x2A, 0x14, 0x08,  //   AF    175  
0x38, 0x45, 0x46, 0x45, 0x20,  //   B0    176  small c with caron
0x38, 0x55, 0x56, 0x55, 0x18,  //   B1    177  small e with caron
0x7C, 0x09, 0x06, 0x05, 0x78,  //   B2    178  small n with caron
0x7C, 0x09, 0x06, 0x05, 0x08,  //   B3    179  small r with caron
0x48, 0x55, 0x56, 0x55, 0x20,  //   B4    180  small s with caron
0x44, 0x65, 0x56, 0x4D, 0x44,  //   B5    181  small z with caron
0x3C, 0x40, 0x42, 0x21, 0x7C,  //   B6    182  small u with acute
0x20, 0x54, 0x56, 0x55, 0x78,  //   B7    183  small a with acute
0x38, 0x44, 0x46, 0x45, 0x20,  //   B8    184  small c with acute
0x48, 0x54, 0x56, 0x55, 0x20,  //   B9    185  small s with acute
0x00, 0x49, 0x7F, 0x44, 0x00,  //   BA    186  small l with stroke
0x20, 0x54, 0x54, 0xF8, 0x80,  //   BB    187  small a with ogonek
0x44, 0x64, 0x55, 0x4C, 0x44,  //   BC    188  small z with dot above
0x1C, 0x2A, 0xEA, 0xAA, 0x8C,  //   BD    189  small e with ogonek
0x7C, 0x08, 0x12, 0x21, 0x7C,  //   BE    190  capital n with acute
0x3F, 0x29, 0xE9, 0xA9, 0xA1,  //   BF    191  capital e with ogonek
0x78, 0x15, 0x16, 0x14, 0x78,  //   C0    192  capital a with grave
0x78, 0x14, 0x16, 0x15, 0x78,  //   C1    193  capital a with acute
0x78, 0x16, 0x15, 0x16, 0x78,  //   C2    194  capital a with circumflex
0x7A, 0x15, 0x16, 0x16, 0x79,  //   C3    195  capital a with tilde
0x38, 0x45, 0x46, 0x45, 0x28,  //   C4    196  capital c with caron
0x7C, 0x55, 0x56, 0x55, 0x44,  //   C5    197  capital e with caron
0x7C, 0x15, 0x16, 0x35, 0x48,  //   C6    198  capital r with caron
0x48, 0x55, 0x56, 0x55, 0x24,  //   C7    199  capital s with caron
0x7C, 0x55, 0x56, 0x54, 0x44,  //   C8    200  capital e with grave
0x44, 0x65, 0x56, 0x4D, 0x44,  //   C9    201  capital z with caron
0x7C, 0x56, 0x55, 0x56, 0x44,  //   CA    202  capital e with circumflex
0x7C, 0x55, 0x54, 0x55, 0x44,  //   CB    203  capital e with diaeresis
0x00, 0x45, 0x7E, 0x44, 0x00,  //   CC    204  capital i with grave
0x00, 0x44, 0x7E, 0x44, 0x00,  //   CD    205  capital i with acute
0x00, 0x46, 0x7D, 0x46, 0x00,  //   CE    206  capital i with circumflex
0x00, 0x45, 0x7C, 0x45, 0x00,  //   CF    207  capital i with diaeresis
0x08, 0x7F, 0x49, 0x22, 0x1C,  //   D0    208  capital eth
0x62, 0x52, 0x4B, 0x46, 0x42,  //   D1    209  capital z with dot above
0x38, 0x45, 0x46, 0x44, 0x38,  //   D2    210  capital o with grave
0x38, 0x44, 0x46, 0x45, 0x38,  //   D3    211  capital o with acute
0x38, 0x46, 0x45, 0x46, 0x38,  //   D4    212  capital o with circumflex
0x3A, 0x45, 0x46, 0x46, 0x39,  //   D5    213  capital o with tilde
0x7C, 0x09, 0x12, 0x21, 0x7C,  //   D6    214  capital n with caron
0x7E, 0x09, 0x09, 0xC9, 0xBE,  //   D7    215  capital a with ogonek
0x3C, 0x41, 0x42, 0x40, 0x3C,  //   D8    216  capital u with grave
0x3C, 0x40, 0x40, 0x41, 0x3C,  //   D9    217  capital u with acute
0x3C, 0x42, 0x41, 0x42, 0x3C,  //   DA    218  capital u with circumflex
0x00, 0x78, 0x41, 0x41, 0x41,  //   DB    219  empty box start
0x41, 0x41, 0x78, 0x00, 0x00,  //   DC    220  empty box end
0x0C, 0x10, 0x62, 0x11, 0x0C,  //   DD    221  capital y with acute
0x7F, 0x24, 0x24, 0x24, 0x18,  //   DE    222  captial THORN
0x00, 0x78, 0x41, 0x55, 0x49,  //   DF    223  crossed box start
0x55, 0x41, 0x78, 0x00, 0x00,  //   E0    224  crossed box end
0x7E, 0x01, 0x4D, 0x4A, 0x30,  //   E1    225  sz
0x00, 0x44, 0x7F, 0x41, 0x00,  //   E2    226  small i with acute
0x62, 0x55, 0x56, 0x56, 0x79,  //   E3    227  small a with tilde
0x7f, 0x63, 0x55, 0x49, 0x49,  //   E4    228  mailbox start
0x49, 0x55, 0x63, 0x7f, 0x00,  //   E5    229  mailbox end
0x7E, 0x10, 0x10, 0x08, 0x1E,  //   E6    230  
0x38, 0x44, 0x46, 0x45, 0x38,  //   E7    231  small o with acute
0x5C, 0x32, 0x2A, 0x26, 0x1D,  //   E8    232  Danish  (capital)
0x02, 0x73, 0x49, 0x45, 0x45,  //   E9    233  solid phone start
0x45, 0x49, 0x73, 0x02, 0x00,  //   EA    234  solid phone end
0x02, 0x73, 0x79, 0x55, 0x7D,  //   EB    235  empty phone start
0x55, 0x79, 0x73, 0x02, 0x00,  //   EC    236  empty phone end
0x18, 0x64, 0x3C, 0x26, 0x18,  //   ED    237  Danish oe (lower case)
0x1C, 0x2A, 0x49, 0x41, 0x22,  //   EE    238  Euro Symbol
0x7E, 0x09, 0x06, 0x05, 0x78,  //   EF    239  small n with tilde
0x30, 0x49, 0x4D, 0x4A, 0x3D,  //   F0    240  small eth
0x4E, 0x51, 0x71, 0x11, 0x10,  //   F1    241  empty is just for testing !!!
0x7E, 0x09, 0x06, 0x05, 0x78,  //   F2    242  empty is just for testing !!!
0x7C, 0x08, 0x06, 0x05, 0x78,  //   F3    243  small n with acute
0x32, 0x49, 0x4A, 0x4A, 0x31,  //   F4    244  small o with tilde
0x00, 0x00, 0x00, 0x00, 0x00,  //   F5    245
0x08, 0x08, 0x2A, 0x08, 0x08,  //   F6    246  'Division'
0x9C, 0xA2, 0xE2, 0x22, 0x10,  //   F7    247  
0x00, 0x02, 0x05, 0x02, 0x00,  //   F8    248   
0x00, 0x00, 0x18, 0x18, 0x00,  //   F9    249  'Multiplication'
0x3E, 0x41, 0x7F, 0x49, 0x41,  //   FA    250  capital ligature oe
0x38, 0x44, 0x38, 0x54, 0x5C,  //   FB    251  small ligature oe
0x0C, 0x11, 0x60, 0x11, 0x0C,  //   FC    252  capital Y with diaeresis
0x18, 0xA0, 0xA2, 0xA1, 0x78,  //   FD    253  small y with acute
0xFF, 0x18, 0x24, 0x24, 0x18,  //   FE    254  small THORN
0xFF, 0xFF, 0xFF, 0xFF, 0xFF   //   FF    255
};

/* read/write: The default state is CS = 1, D = output, WR = 1, DC =
   1, RST = 1, RD = 1 */

/* Write command */

static void write_command(unsigned int command)
{
  int portb_set_val = 0;
  int portb_clear_val = 0;
  int j; // for DELAY

  if (command & (1<<0)) portb_set_val |= DB0_B; else portb_clear_val |= DB0_B;
  if (command & (1<<1)) portb_set_val |= DB1_B; else portb_clear_val |= DB1_B;
  if (command & (1<<2)) portb_set_val |= DB2_B; else portb_clear_val |= DB2_B;
  if (command & (1<<3)) portb_set_val |= DB3_B; else portb_clear_val |= DB3_B;
  if (command & (1<<4)) portb_set_val |= DB4_B; else portb_clear_val |= DB4_B;
  if (command & (1<<5)) portb_set_val |= DB5_B; else portb_clear_val |= DB5_B;
  if (command & (1<<6)) portb_set_val |= DB6_B; else portb_clear_val |= DB6_B;
  if (command & (1<<7)) portb_set_val |= DB7_B; else portb_clear_val |= DB7_B;

  set_port(PORT_B, portb_set_val, portb_clear_val | CS_B | WR_B | DC_B);
  DELAY;
  set_port(PORT_B, DC_B | CS_B | WR_B, 0);
}

/* Write data */
static void write_data(unsigned int data)
{ 
  int portb_set_val = 0;
  int portb_clear_val = 0;
  int j;

  if (data & (1<<0)) portb_set_val |= DB0_B; else portb_clear_val |= DB0_B;
  if (data & (1<<1)) portb_set_val |= DB1_B; else portb_clear_val |= DB1_B;
  if (data & (1<<2)) portb_set_val |= DB2_B; else portb_clear_val |= DB2_B;
  if (data & (1<<3)) portb_set_val |= DB3_B; else portb_clear_val |= DB3_B;
  if (data & (1<<4)) portb_set_val |= DB4_B; else portb_clear_val |= DB4_B;
  if (data & (1<<5)) portb_set_val |= DB5_B; else portb_clear_val |= DB5_B;
  if (data & (1<<6)) portb_set_val |= DB6_B; else portb_clear_val |= DB6_B;
  if (data & (1<<7)) portb_set_val |= DB7_B; else portb_clear_val |= DB7_B;

  set_port(PORT_B, portb_set_val | DC_B, portb_clear_val | CS_B | WR_B);
  DELAY;
  set_port(PORT_B, CS_B | WR_B, 0);
}

int status_read(void)
{
  int r, disp_status;  
  int j; // DELAY

  set_port_dir(PORT_B, INPUT_PINS,
	       DB0_B | DB1_B | DB2_B | DB3_B |
	       DB4_B | DB5_B | DB6_B | DB7_B);
  set_port(PORT_B, 0, DC_B); // reading status is a command
  set_port(PORT_C, 0, RD_C); //read starts  
  DELAY;
  r = read_port(PORT_B);
  set_port(PORT_B, DC_B, 0);
  set_port(PORT_C, RD_C, 0); // read ends
  set_port_dir(PORT_B, OUTPUT_PINS,
	       DB0_B | DB1_B | DB2_B | DB3_B |
	       DB4_B | DB5_B | DB6_B | DB7_B);

  disp_status = 0;
  if (r & DB0_B) disp_status |= (1<<0);
  if (r & DB1_B) disp_status |= (1<<1);
  if (r & DB2_B) disp_status |= (1<<2);
  if (r & DB3_B) disp_status |= (1<<3);
  if (r & DB4_B) disp_status |= (1<<4);
  if (r & DB5_B) disp_status |= (1<<5);
  if (r & DB6_B) disp_status |= (1<<6); 
  if (r & DB7_B) disp_status |= (1<<7);
  return disp_status;
}

static int text_x = 0, text_y = 0;

static void clear_line(int line)
{
   int cnt = 132;

   write_command(SET_PAGE_ADDR | line);
   write_command(SET_LOWER_COL |  0);
   write_command(SET_HIGH_COL  |  0);
    
   do
   {
      write_data(0);
   }
   while (cnt-- > 0);
}

void clear_snom_display(void)
{
   clear_line(0);
   clear_line(1);
   clear_line(2);

   text_x = text_y = 0;
}

void hardware_platform (void)
{
  //short  phone_type;
  if (global_serial_num >= SNOM200_B)
    contrast_val = SET_INIT_CONTRAST_B;
    //phone_type = 1; 
    else    
    contrast_val = SET_INIT_CONTRAST_A;
    //phone_type = 0;
   // return phone_type;
}


 
static void ssd1815b_init(int cold)
{
  int j;
  static int initial = 1;

  /*** Set the ports to the correct mode. */
  if (cold) {
    set_port_dir(PORT_B, OUTPUT_PINS,
		 DB0_B | DB1_B | DB2_B | DB3_B |
		 DB4_B | DB5_B | DB6_B | DB7_B |
		 CS_B | RST_B | WR_B  | DC_B);
    
    set_port_dir(PORT_C, OUTPUT_PINS, RD_C);
  }

  /* reset */
  if (initial || cold) {
    initial = 0;
    // set default config:
    set_port(PORT_C, RD_C, 0);
    set_port(PORT_B, CS_B | WR_B | RST_B | DC_B, 0);
    RST_DELAY;
  }

  if (cold) {
    // perform reset cycle:
    set_port(PORT_B, 0, RST_B);
    RST_DELAY;
    set_port(PORT_B, RST_B, 0);
  }
  else {
    write_command(0xE2);  // Software reset
  }

  while(1) {
    int r = status_read();
    if ((r & 0x10) == 0) break;
  }

  write_command(SET_DISPLAY_OFF);      // AE
  write_command(SET_SEG_REMAP_OFF);    // A0
  write_command(SET_COM_REMAP_OFF);    // C0
  write_command(SET_INIT_RES_RATIO);   // 23
  write_command(SET_CONT_CTRL_REG);    // 81
  write_command((unsigned char)contrast_val);  
  write_command(SET_EXT_COMMAND);      // A9
  write_command(0xc1);                 // 21 kHz Bias = 1/6
  write_command(SET_ALTERNATE_BIAS); // A3
  write_command(SET_POWER_ON);         // 2F
  write_command(SET_DISPLAY_ON);       // AF
  //write_command(SET_BIAS_CTRL_REG);  // A9
  //write_command(SET_1_4_BIAS_RATIO); // AB
  //write_command(0xd4); // set total frame phases
  //write_command(0x30); // 16 phases

  if (cold)
    clear_snom_display();
}

/* 0 - idle, 1 - got x, 2 - got y, 3 - still need to bytes.  */

static int write_state = 0;
static int x = 0, y = 0, len = 0;
static int invert = 0;

int write_snom_display(const unsigned char * buf, int count)
{
  int l;
  //temp = status_read();
  //if(temp != display_status) ssd1815b_init();

  if(count < 0)
    return -1;

  for(l = 0; l < count; l++) {

    switch(write_state) {
    case 0:
      {
	x = *((const unsigned char *) buf);
	write_state = 1;
	break;
      }
    case 1:
      {
	y = *((const unsigned char *) buf);
	write_state = 2;
	break;
      }
    case 2:
      {
	len = *((const unsigned char *) buf);

   	write_command(SET_PAGE_ADDR |  y);

	if(len == 0) // text ?
	{
	   write_state = 4;

	   text_y = y;
           text_x = x;

	   // setup correct text position.
	
	   x = x * 5;
           if(x >= 85)
		x += 11;
	   else if(x >= 35)
		x += (x/5 - 6);
		
  	   write_command(SET_LOWER_COL |  (x & 0xf));
  	   write_command(SET_HIGH_COL  |  ((x >> 4) & 0xf));
	}
	else
	{
	   write_state = 3;

	   // setup correct position.
	
   	   write_command(SET_PAGE_ADDR |  y);
  	   write_command(SET_LOWER_COL |  (x & 0xf));
  	   write_command(SET_HIGH_COL  |  ((x >> 4) & 0xf));
	}
	break;
      }
    case 3:
      {
	write_data(*buf);
	len--;
	if (len == 0) write_state = 0;
	break;
      }
    case 4:		// get text length
      {
	len = *((const unsigned char *) buf);

	write_state = 5;
	break;
      }
    case 5:
      {
	int i = *buf * 5;

	if(*buf == '\n') { // provoke wrap
          text_x = 24;
 	} 
	else if (*buf == 11) { // reset device (hard reset)
 	  ssd1815b_init(0); // FIXMEE!! if 1 put 1 here the phone crashes! cs.
 	}
	else if (*buf == 12) { // reset device (soft reset)
 	  ssd1815b_init(0);
 	}
	else if (*buf == 14) { // invert on
 	  invert = 255;
 	}
 	else if (*buf == 15) { // invert off
 	  invert = 0;
 	}
 	else {
	    int x;
	
	    x = text_x * 5;
            if(x >= 85)
	      x += 11;
	    else if(x >= 35)
	      x += (x/5 - 6);
		
	// Handle special chars here

	   switch(*buf)
	   {
    	       case 0xe5:
    	       case 0xea:
	       case 0xec:
	       case 0xde:
	       case 0xe0:
		x--;
	   }

	   if(text_y == 0) {
   	      write_command(SET_PAGE_ADDR);
  	      write_command(SET_LOWER_COL |  (x & 0xf));
  	      write_command(SET_HIGH_COL  |  ((x >> 4) & 0xf));

 	      write_data(char_code[i++] ^ invert);
 	      write_data(char_code[i++] ^ invert);
 	      write_data(char_code[i++] ^ invert);
 	      write_data(char_code[i++] ^ invert);
 	      write_data(char_code[i  ] ^ invert);
	   }
           else if(text_y == 1)
	   {
   	      write_command(SET_PAGE_ADDR |  1);
  	      write_command(SET_LOWER_COL |  (x & 0xf));
  	      write_command(SET_HIGH_COL  |  ((x >> 4) & 0xf));

 	      write_data(((char_code[i++] ^ invert) & 0x7f) << 1);
 	      write_data(((char_code[i++] ^ invert) & 0x7f) << 1);
 	      write_data(((char_code[i++] ^ invert) & 0x7f) << 1);
 	      write_data(((char_code[i++] ^ invert) & 0x7f) << 1);
 	      write_data(((char_code[i  ] ^ invert) & 0x7f) << 1);

   	      write_command(SET_PAGE_ADDR |  2);
  	      write_command(SET_LOWER_COL |  (x & 0xf));
  	      write_command(SET_HIGH_COL  |  ((x >> 4) & 0xf));

	      i = *buf * 5;
 	      write_data((char_code[i++] ^ invert) & 0x80 ? 0x01 : 0x00);
 	      write_data((char_code[i++] ^ invert) & 0x80 ? 0x01 : 0x00);
 	      write_data((char_code[i++] ^ invert) & 0x80 ? 0x01 : 0x00);
 	      write_data((char_code[i++] ^ invert) & 0x80 ? 0x01 : 0x00);
 	      write_data((char_code[i  ] ^ invert) & 0x80 ? 0x01 : 0x00);
	   }
	}

 	if (*buf != 14 && *buf != 15 && *buf != 11 && *buf != 12)
          text_x++;

	if(text_x > 23) {
	   text_x = 0; text_y++;
   	   write_command(SET_PAGE_ADDR |  text_y);
  	   write_command(SET_LOWER_COL);
  	   write_command(SET_HIGH_COL);
	} 

	len--;
	if (len == 0) write_state = 0;
	break;
      }
    default:
      {
	write_state = 0;
	break;
      }
    } // switch 

    buf++;
  }

  return count;
}

void writeln_snom_display(unsigned char *str)
{
   unsigned char aucTmp[30];
   int i, cnt;

   for(i=0;i<24;i++) 
	if(str[i] == 0)
		break;

   cnt = i;

   aucTmp[0] = text_x;
   aucTmp[1] = text_y;
   aucTmp[2] = 0;      // text escape sequence
   aucTmp[3] = cnt;

   for(i=0;i<cnt;i++)
	aucTmp[i+4] = str[i];
   
   write_snom_display(aucTmp, cnt + 4);
}

void text_pos_snom_disp(int xpos, int ypos)
{
	text_x = xpos;
	text_y = ypos;
}

#define GPIO0  (1<<7)
#define GPIO1  (1<<6)
#define GPIO2  (1<<5)
#define GPIO3  (1<<4)

#define PC9    (1<<6)
#define PC15   (1<<0)
static void init_leds(void)
{
   set_port_dir(PORT_C, OUTPUT_PINS, PC9 | PC15);
   set_port_dir(PORT_SNOM_CSR, OUTPUT_PINS, GPIO0 | GPIO1 | GPIO2 | GPIO3);
   set_port(PORT_C, 0,PC9 | PC15);
   set_port(PORT_SNOM_CSR,0, GPIO0 | GPIO1 | GPIO2 | GPIO3);
}

void put_led_snom_display(int nVal)
{
   int bOff = 0;

   if (nVal < 0) {
      bOff = 1;
      nVal = -nVal;
   } 

   switch(nVal) {
      case 1:
	 set_port(PORT_C, !bOff ? PC9 : 0, bOff ? PC9 : 0);
         break;
      case 2:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO0 : 0, bOff ? GPIO0 : 0);
         break;
      case 3:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO1 : 0, bOff ? GPIO1 : 0);
         break;
      case 4:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO2 : 0, bOff ? GPIO2 : 0);
         break;
      case 5:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO3 : 0, bOff ? GPIO3 : 0);
         break;
      case 6:
	 set_port(PORT_C, !bOff ? PC15 : 0, bOff ? PC15 : 0);
         break;
    }
}

void toggle_leds(void) 
{
   static int count = 0;

   put_led_snom_display(count & 1 ? 1 : -1);
   put_led_snom_display(count & 2 ? 2 : -2);
   put_led_snom_display(count & 4 ? 3 : -3);
   put_led_snom_display(count & 8 ? 4 : -4);
   put_led_snom_display(count & 16 ? 5 : -5);

   count++;
}

void put_light_snom_display(int bOn)
{
}

/* Contrast Control */


void inc_contrast_snom_display(int bUp)
{
  if(bUp) {
    contrast_val += 1;
    if(contrast_val > 63) contrast_val = 63;
  } 
  else {
    contrast_val -= 1;
    if(contrast_val < 0) contrast_val = 0;
  }
  write_command(SET_CONT_CTRL_REG);
  write_command((unsigned char)contrast_val);  
}

void set_contrast_snom_display(unsigned char contrast)
{
  contrast_val = contrast;
  if(contrast_val > 63) contrast_val = 63;
  if(contrast_val < 0) contrast_val = 0;

  write_command(SET_CONT_CTRL_REG);
  write_command((unsigned char)contrast_val);  
}

void progress_snom_display(void)
{
  unsigned char aucTmp[2] = { 0xff, 0x00 };

  writeln_snom_display(aucTmp);
}

static void disp_snom_logo(void)
{
  clear_snom_display();
/* writeln_snom_display("        snom 200\n");
   writeln_snom_display("   snom technology AG   "); */
}


int init_snom_display(void)
{
  hardware_platform();
  ssd1815b_init(1);
  disp_snom_logo();
  init_leds();
  //toggle_leds();
  
  write_state = 0;
  display_status = status_read();
  return 0;
}

#endif

