/*
 *  arch/ppc/8xx_io/snomdisp220.c
 *
 *  Copyright (C) 2003, snom technology AG
 *
 *  This code is for the LCD module connected to the SNOM 220 board.
 *
 */

#include <linux/config.h>
#include <linux/types.h>

#ifdef CONFIG_SNOM_DISP

#include <string.h>
// #include <linux/kernel.h> // printk

#include <asm/port.h>
#include <asm/mpc8xx.h>

#include "commproc.h"
#include "snomaud.h"

int global_serial_num;

/*****************************************************************
This driver is used to connect the  CS290 DISPLAY (CONTR. SSD1815B)
 to the SNOM  220 board and  thus be able to configure it.  
 

 The data lines are expect to be connected as follows:

28 CS   - PB27 active low
27 RST  - PB29 active low
26 D/C  - PB31 
25 WR   - PB30 make low to write
24 RD   - PC4  make low for read (which we never do)
23 DB0  - PB14 
22 DB1  - PB15 
21 DB2  - PB16 
20 DB3  - PB17 
19 DB4  - PB18 
18 DB5  - PB22 
17 DB6  - PB23 
16 DB7  - PB26 
Display light is on PA8

for other signals on disp con. see sch ip5_disp+kdb_2V1
**************************************************************/

#define DELAY		for(j=0;j<50;j++);
#define RST_DELAY	for(j=0;j<2500;j++);
#define READ_DELAY      for(j=0;j<10;j++); //200ns
extern int status_read(void);
// extern int display_status =0;

/* The data bus and these lines are located on port B. */

#define DB0_B             (1<<17)
#define DB1_B             (1<<16)
#define DB2_B             (1<<15)
#define DB3_B             (1<<14)
#define DB4_B             (1<<13)
#define DB5_B             (1<<9)
#define DB6_B             (1<<8)
#define DB7_B             (1<<5)

#define CS_B              (1<<4)
#define RST_B             (1<<2)
#define WR_B              (1<<1)
#define DC_B              (1<<0)

/* This is on port C. */

#define RD_C              (1<<11)
/* this is Port A ,display light */
/* Port A */
#define LIGHT              (1<<7)


/* Command values */

#define SET_LOWER_COL          0x00
#define SET_HIGH_COL           0x10
#define SET_INIT_RES_RATIO     0x23
#define SET_POWER_ON           0x2F
#define SET_START_LINE         0x40
#define SET_CONT_CTRL_REG      0x81
#define SET_SEG_REMAP_OFF      0xA0
#define SET_ALTERNATE_BIAS     0xA3
#define SET_DISPLAY_OFF        0xAE
#define SET_DISPLAY_ON         0xAF
#define SET_PAGE_ADDR          0xB0
#define SET_COM_REMAP_OFF      0xC0
#define SET_READ_MOD_WRITE     0xE0
#define SET_END_READ_MOD_WRITE 0xEE
#define SET_BIAS_CTRL_REG      0xA9
#define SET_1_4_BIAS_RATIO     0xAB

#define SET_EXT_COMMAND        0xA9

#define SET_INIT_CONTRAST_A      0x30
#define SET_INIT_CONTRAST_B      0x20

static int contrast_val = SET_INIT_CONTRAST_A;

static const unsigned char arrow_page[] = {0xB0, 0xB1, 0xB2, 0xB4, 0xB5, 0xB6, 0xB7,
                                           0xB0, 0xB1, 0xB2, 0xB4, 0xB5, 0xB6, 0xB7};

/* read/write: The default state is CS = 1, D = output, WR = 1, DC =
   1, RST = 1, RD = 1 */

/* Write command */

static void write_command(unsigned int command)
{
  int portb_set_val = 0;
  int portb_clear_val = 0;
  int j; // for DELAY

  if (command & (1<<0)) portb_set_val |= DB0_B; else portb_clear_val |= DB0_B;
  if (command & (1<<1)) portb_set_val |= DB1_B; else portb_clear_val |= DB1_B;
  if (command & (1<<2)) portb_set_val |= DB2_B; else portb_clear_val |= DB2_B;
  if (command & (1<<3)) portb_set_val |= DB3_B; else portb_clear_val |= DB3_B;
  if (command & (1<<4)) portb_set_val |= DB4_B; else portb_clear_val |= DB4_B;
  if (command & (1<<5)) portb_set_val |= DB5_B; else portb_clear_val |= DB5_B;
  if (command & (1<<6)) portb_set_val |= DB6_B; else portb_clear_val |= DB6_B;
  if (command & (1<<7)) portb_set_val |= DB7_B; else portb_clear_val |= DB7_B;

  set_port(PORT_B, portb_set_val, portb_clear_val | CS_B | WR_B | DC_B);
  DELAY;
  set_port(PORT_B, DC_B | CS_B | WR_B, 0);
}

/* Write data */
static void write_data(unsigned int data)
{ 
  int portb_set_val = 0;
  int portb_clear_val = 0;
  int j;

  if (data & (1<<7)) portb_set_val |= DB0_B; else portb_clear_val |= DB0_B;
  if (data & (1<<6)) portb_set_val |= DB1_B; else portb_clear_val |= DB1_B;
  if (data & (1<<5)) portb_set_val |= DB2_B; else portb_clear_val |= DB2_B;
  if (data & (1<<4)) portb_set_val |= DB3_B; else portb_clear_val |= DB3_B;
  if (data & (1<<3)) portb_set_val |= DB4_B; else portb_clear_val |= DB4_B;
  if (data & (1<<2)) portb_set_val |= DB5_B; else portb_clear_val |= DB5_B;
  if (data & (1<<1)) portb_set_val |= DB6_B; else portb_clear_val |= DB6_B;
  if (data & (1<<0)) portb_set_val |= DB7_B; else portb_clear_val |= DB7_B;

  set_port(PORT_B, portb_set_val | DC_B, portb_clear_val | CS_B | WR_B);
  DELAY;
  set_port(PORT_B, CS_B | WR_B, 0);
}

int status_read(void)
{
  int r, disp_status;  
  int j; // DELAY

  set_port_dir(PORT_B, INPUT_PINS,
	       DB0_B | DB1_B | DB2_B | DB3_B |
	       DB4_B | DB5_B | DB6_B | DB7_B);
  set_port(PORT_B, 0, DC_B); // reading status is a command
  set_port(PORT_C, 0, RD_C); //read starts  
  DELAY;
  r = read_port(PORT_B);
  set_port(PORT_B, DC_B, 0);
  set_port(PORT_C, RD_C, 0); // read ends
  set_port_dir(PORT_B, OUTPUT_PINS,
	       DB0_B | DB1_B | DB2_B | DB3_B |
	       DB4_B | DB5_B | DB6_B | DB7_B);

  disp_status = 0;
  if (r & DB0_B) disp_status |= (1<<0);
  if (r & DB1_B) disp_status |= (1<<1);
  if (r & DB2_B) disp_status |= (1<<2);
  if (r & DB3_B) disp_status |= (1<<3);
  if (r & DB4_B) disp_status |= (1<<4);
  if (r & DB5_B) disp_status |= (1<<5);
  if (r & DB6_B) disp_status |= (1<<6); 
  if (r & DB7_B) disp_status |= (1<<7);
  return disp_status;
}

static void ssd1815b_init(int cold)
{
  int j;
  static int initial = 1;

  /*** Set the ports to the correct mode. */
  if (cold) {
    set_port_dir(PORT_B, OUTPUT_PINS,
		 DB0_B | DB1_B | DB2_B | DB3_B |
		 DB4_B | DB5_B | DB6_B | DB7_B |
		 CS_B | RST_B | WR_B  | DC_B);
    
    set_port_dir(PORT_C, OUTPUT_PINS, RD_C);
  }
  

  /* reset */
  if (initial || cold) {
    initial = 0;
    // set default config:
    set_port(PORT_C, RD_C, 0);
    set_port(PORT_B, CS_B | WR_B | RST_B | DC_B, 0);
    RST_DELAY;
  }

  if (cold) {
    // perform reset cycle:
    set_port(PORT_B, 0, RST_B);
    RST_DELAY;
    set_port(PORT_B, RST_B, 0);
    for(j=0;j<100000;j++);
  }
  else {
    write_command(0xE2);  // Software reset
    for(j=0;j<1000;j++);
  }

  write_command(SET_DISPLAY_OFF);      // AE
  write_command(SET_SEG_REMAP_OFF);    // A0
  write_command(SET_COM_REMAP_OFF);    // C0
  write_command(SET_INIT_RES_RATIO);   // 23
  write_command(SET_CONT_CTRL_REG);    // 81
  write_command((unsigned char)contrast_val);  
  write_command(SET_EXT_COMMAND);      // A9
  write_command(0xc1);                 // 21 kHz Bias = 1/6
  write_command(SET_ALTERNATE_BIAS); // A3
  write_command(SET_POWER_ON);         // 2F
  write_command(SET_DISPLAY_ON);       // AF
  //write_command(SET_BIAS_CTRL_REG);  // A9
  //write_command(SET_1_4_BIAS_RATIO); // AB
  //write_command(0xd4); // set total frame phases
  //write_command(0x30); // 16 phases

//   if (cold)
//     clear_snom_display();
}

/* 0 - idle, 1 - got x, 2 - got y, 3 - still need to bytes.  */

static int write_state = 0;
static int x = 0, y = 0, len = 0;

int write_snom_display(const unsigned char * buf, int count)
{
  int l;
  //temp = status_read();
  //if(temp != display_status) ssd1815b_init();
  // printk("SNOM Display write %d...", count);

  if(count < 0)
    return -1;

  for(l = 0; l < count; l++) {
    const unsigned char c = *buf++;
    switch(write_state) {
    case 0:
      {
	x = c + 1; // because display has moved one pixel in contrast to snom200
	write_state = 1;
	break;
      }
    case 1:
      {
 	y = 7 - c;
        if (y < 0) y = 0;
        else if (y > 7) y = 7;
	write_state = 2;
	break;
      }
    case 2:
      {
	len = c;
	if (x < 1) x = 1;
	else if (x + len > 129) x = 0; // will produce garbage but thats ok
        write_state = 3;

        // setup correct position.
	
        write_command(SET_PAGE_ADDR |  y);
        write_command(SET_LOWER_COL |  (x & 0xf));
        write_command(SET_HIGH_COL  |  ((x >> 4) & 0xf));
	break;
      }
    case 3:
      {
	write_data(c);
	len--;
	if (len == 0) write_state = 0;
	break;
      }
    default:
      {
	write_state = 0;
	break;
      }
    } // switch 
  }

  // printk("done\n");
  return count;
}

#define GPIO0  (1<<7)
#define GPIO1  (1<<6)
#define GPIO2  (1<<5)
#define GPIO3  (1<<4)

#define PC9    (1<<6)
#define PC15   (1<<0)
//#define EXT_KBD (1<<3)
static void init_leds(void)
{
   set_port_dir(PORT_C, OUTPUT_PINS, PC9 | PC15);
   //set_port_dir(PORT_C, INPUT_PINS, EXT_KBD);
   /*set in snomaud200.c*/
   
   set_port_dir(PORT_SNOM_CSR, OUTPUT_PINS, GPIO0 | GPIO1 | GPIO2 | GPIO3);
   set_port(PORT_C, 0,PC9 | PC15);
   set_port(PORT_SNOM_CSR,0, GPIO0 | GPIO1 | GPIO2 | GPIO3);
}

void show_arrows_snom_display(void);
void hide_arrows_snom_display(void);


void put_led_snom_display(int nVal)
{
   int bOff = 0;

   if (nVal < 0) {
      bOff = 1;
      nVal = -nVal;
   } 

   switch(nVal) {
      case 1:
	 set_port(PORT_C, !bOff ? PC9 : 0, bOff ? PC9 : 0);
         break;
      case 2: /* Ringer LED !!! */
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO0 : 0, bOff ? GPIO0 : 0);
         break;
      case 3:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO1 : 0, bOff ? GPIO1 : 0);
         break;
      case 4:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO2 : 0, bOff ? GPIO2 : 0);
         break;
      case 5:
	 set_port(PORT_SNOM_CSR, !bOff ? GPIO3 : 0, bOff ? GPIO3 : 0);
         break;
      case 6:
	 set_port(PORT_C, !bOff ? PC15 : 0, bOff ? PC15 : 0);
         break;
    }
}

void toggle_leds(void) 
{
   static int count = 0;

   put_led_snom_display(count & 1 ? 1 : -1);
   put_led_snom_display(count & 2 ? 2 : -2);
   put_led_snom_display(count & 4 ? 3 : -3);
   put_led_snom_display(count & 8 ? 4 : -4);
   put_led_snom_display(count & 16 ? 5 : -5);
   put_led_snom_display(count & 32 ? 6 : -6);

   count++;
}

void put_light_snom_display(int bOn)
{
  set_port_dir(PORT_A, OUTPUT_PINS, LIGHT); //I will move this line to init place when it is ready
 if(bOn)
    set_port(PORT_A, LIGHT, 0);
  else
    set_port(PORT_A, 0, LIGHT);
}

 
/* Contrast Control */


void inc_contrast_snom_display(int bUp)
{
  if(bUp) {
    contrast_val += 1;
    if(contrast_val > 63) contrast_val = 63;
  } 
  else {
    contrast_val -= 1;
    if(contrast_val < 0) contrast_val = 0;
  }
  write_command(SET_CONT_CTRL_REG);
  write_command((unsigned char)contrast_val);  
}

void set_contrast_snom_display(unsigned char contrast)
{
  // printk("SNOM Display contrast: %d...", contrast);
  contrast_val = contrast;
  if(contrast_val > 63) contrast_val = 63;
  if(contrast_val < 0) contrast_val = 0;

  write_command(SET_CONT_CTRL_REG);
  write_command((unsigned char)contrast_val);  
  // printk("done\n", contrast);
}

void hide_arrows_snom_display(void)  // taken from lcs2/doc/diplay.drv
{
  unsigned char i;
  
  for (i = 0x10; i <= 0x1D; i++) {
    if (i < 0x17 ) {
      write_command(SET_LOWER_COL | 0);
      write_command(SET_HIGH_COL  | 0);
    } else {
      write_command(SET_LOWER_COL | 1);
      write_command(SET_HIGH_COL  | 8);
    }
            
    write_command(arrow_page[i & 0x0F]);
    write_data(0);
  }
}

void show_arrows_snom_display(void)  // taken from lcs2/doc/diplay.drv
{
  unsigned char i;
  
  for (i = 0x10; i <= 0x1D; i++) {
    if (i < 0x17 ) {
      write_command(SET_LOWER_COL | 0);
      write_command(SET_HIGH_COL  | 0);
    } else {
      write_command(SET_LOWER_COL | 1);
      write_command(SET_HIGH_COL  | 8);
    }
            
    write_command(arrow_page[i & 0x0F]);
    write_data(0xff);
  }
}

int init_snom_display(void)
{
  // printk("SNOM Display Init start\n");
  ssd1815b_init(1);
  init_leds();
  set_contrast_snom_display(contrast_val);
  put_light_snom_display(1); // we want to see something in the begining!
  //toggle_leds();
  hide_arrows_snom_display ();
  write_state = 0;
  // printk("SNOM Display Init done\n");
  // display_status = status_read();
  return 0;
}

#endif
