#include <linux/config.h>
#include "../../kernel/ppc_defs.h"
#include "../../kernel/ppc_asm.tmpl"
#include <asm/processor.h>
#include <asm/cache.h>

	.text

/*
 * $Id: snomhead.S,v 1.7 2003/10/01 10:27:14 fischesv Exp $
 *	
 * This code is loaded by the ROM loader at some arbitrary location.
 * Move it to high memory so that it can load the kernel at 0x0000.
 *
 * This is a three step process that will also work when booting from
 * a Flash PROM normally located in high memory.
 *
 * First, the entire image is loaded into some high memory address.
 * This is usually at or above 0x02000000.  This is done by a network
 * boot function supported by the board or a debugger over BDM port.
 *
 * Second, the start up function here will relocate the decompress
 * function to run at the link address of 0x01000000.
 *
 * Last, the decompression function will reloate the initrd, zImage, and
 * the residual data to locations under 8 Meg.  This is necessary because
 * the embedded kernel start up uses 8 Meg translations to access physical
 * space before the MMU is enabled.  Finally, the zImage is uncompressed
 * to location 0 and we jump to it.
 *
 * On the MBX,
 *              R1 - Stack pointer at a high memory address.
 *              R3 - Pointer to Board Information Block.
 *              R4 - Pointer to argument string.
 *              Interrupts masked, cache and MMU disabled.
 *
 *		...and the first and second functions listed above are
 *			done for us (it knows ELF images).
 *
 * For other embedded boards we build the Board Information Block.
 */

#if defined (CONFIG_SNOM1)

/* The initial software was developed on the 860T/FADS board.
 * Modifications may be necessary for other processors.
 */

#define SIUMCR_VAL	0x01012440

#ifdef CONFIG_WATCH_DOG
#define SYPCR_VAL	0xffffff07	/* enable watch dog */
#else
#define	SYPCR_VAL	0xffffff08	/* disable watch dog */
#endif

#define TBSCR_VAL	0x00c2
#define RTCSC_VAL	0x00c2
#define PISCR_VAL	0x0082
#define SCCR_FAST_VAL	0x03E00300
#define SCCR_SLOW_VAL	0x03E20300

#define MAMR_VAL	0x60a21114	

#define BR0     0x00000100
#define OR0     0x00000104
#define BR1     0x00000108
#define OR1     0x0000010C
#define BR2     0x00000110
#define OR2     0x00000114
#define BR3     0x00000118
#define OR3     0x0000011C
#define BR4     0x00000120
#define OR4     0x00000124
#define BR5     0x00000128
#define OR5     0x0000012C
#define BR6     0x00000130
#define OR6     0x00000134
#define BR7     0x00000138
#define OR7     0x0000013C           

#define SYPCR	0x00000004
#define SWSR	0x0000000e
#define MAMR	0x00000170
#define MBMR	0x00000174
#define MAR	0x00000164
#define MPTPR	0x0000017a
#define MCR	0x00000168
#define MDR	0x0000017c
#define TBSCR	0x00000200
#define RTCSC	0x00000220
#define PISCR	0x00000240
#define SCCR	0x00000280
#define PLPRCR	0x00000284
#define RSR	0x00000288

#define KEY_VAL	0x55ccaa33
	
#define TBSCRK	0x00000300
#define RTCSCK	0x00000320
#define PISCRK	0x00000340
#define SCCRK	0x00000380
#define PLPRCRK	0x00000384

#define ICR     148
#define DER     149
#define ICTRL   158

#define SSR1	27

#ifdef CONFIG_SNOM1
			
/* Processor Chip selects.
*/

#define BR0_VAL 0xfe000801
#define OR0_VAL 0xffe00d34
	
#define BR2_VAL	0x00000081	/* BA=0x00000000, AT=000, PS=00 (32 bit),
				   PARE=0, WP=0 (RW), MS=10 (UPMA), V=1 */ 
#define OR2_VAL	0xff800a00	/* 8MB block size, all types access, initial
				   address according to AMA. */
	
#define BR3_VAL	0x00800081	/* BA=0x00800000, AT=000, PS=00 (32 bit),
				   PARE=0, WP=0 (RW), MS=10 (UPMA), V=1 */ 
#define OR3_VAL	0xff800a00	/* 8MB block size, all types access, initial
				   address according to AMA. */
	
#define BR4_VAL 0xfe400401  	/* BA=0xfe400000, AT=000, PS=01 (8 bit),
                                   PARE=0 (disable), WP=0 (RW),
                                   MS=00 (GPCM), V=1 (Valid) */
 
#define OR4_VAL 0xffe008f4      /* AM=0xffe00000 (2MB), all types access,
                                   CS early negate, 16 w.s., timing relax */

#endif /* CONFIG_SNOM */
	
#define ONE_SEC	1000000		/* One second of TBL ticks @ 40 MHz */
#define CTR_ONESEC	166667

#undef IMMR
#define IMMR 0xff000000

	.long 0xffffffff	/* reserve space for ethernet addr. */
	.long 0xffffffff	
#ifdef CONFIG_SNOM220
	.long 0x00040900	/* Version 49 = 4.9 */
#else
#ifdef CONFIG_SNOM200
	.long 0x00020900	/* Version 29 = 2.9 */
#else
#ifdef CONFIG_SNOM105
	.long 0x00030900	/* Version 39 = 3.9 */
#else
	.long 0x00010961	/* Version 19a = 1.9a */
#endif
#endif
#endif
	.long 0xffffffff	/* CRC to be filled out. */

	. = 0x10

  	lis	r3,print_reset_regs@h
	ori	r3,r3,print_reset_regs@l
	mtlr	r3		/* Easiest way to do an absolute jump */
	blr
		
/* See SIU9 errata */
	. = 0x40	
		
.global _set_plprcr
_set_plprcr:

	/* Crank up the PLL so we run at about 50 Mhz
	 and clear TEXPS signal. */
	
	stw	r3, PLPRCR(r20)
	isync
	addi	r3, r0, r1
	divw	r4, r4, r3
	divw	r4, r4, r3
	addi	r3, r0, r1
	addi	r3, r0, r1
	divw	r4, r4, r3
	divw	r4, r4, r3
	blr
	
	
	. = 0x100

#endif /* CONFIG_SNOM1 */

	
	.globl	start
start:
	/* Disable data cache. */

	mfspr	r3, DC_CST
	ori	r8, r3, 0
	lis	r3, IDC_DISABLE@h
	mtspr	DC_CST, r3
	mtspr	IC_CST, r3

	/* Initialize MSR and SRR1 */

	li	r10, MSR_	/* Set ME, RI flags */
	mtmsr	r10
	mtspr	SSR1, r10
	
	/* Initialize Instruction Support Control Register, ICTRL */

	/* Disable serialized ifetch and show cycles (i.e. set processor
	 * to normal mode).
	 */
	
	li	r21, 0x0007
	mtspr	ICTRL, r21

	/* Disable debug mode entry. */
	
	li	r21, 0
	mtspr	DER, r21

	/* Initialize Interrupt Cause Register */

	li	r21, 0
	mtspr	ICR, r21
	
	/* Move internal memory map. */
	
	lis	r20, IMMR@h
	mtspr	638, r20
	
	mfmsr	r3		/* Turn off interrupts  */
	li	r4,0
	ori	r4,r4,MSR_EE
	andc	r3,r3,r4
	mtmsr	r3
	
	lis	r21, SIUMCR_VAL@h
	ori	r21, r21, SIUMCR_VAL@l
	stw	r21, 0(r20)

#if 0 /* this can only be set once after reset and the boot loader does that ... */
	lis	r21, SYPCR_VAL@h
	ori	r21, r21, SYPCR_VAL@l
	stw	r21, SYPCR@l(r20)
#endif

	lis	r21, KEY_VAL@h
	ori	r21, r21, KEY_VAL@l
	stw	r21, SCCRK(r20)

	lis	r21, SCCR_FAST_VAL@h
	ori	r21, r21, SCCR_FAST_VAL@l
	stw	r21, SCCR(r20)

	lis	r21, KEY_VAL@h
	ori	r21, r21, KEY_VAL@l
	stw	r21, TBSCRK(r20)

	li	r21, TBSCR_VAL@l
	sth	r21, TBSCR(r20)
	
	lis	r21, KEY_VAL@h
	ori	r21, r21, KEY_VAL@l
	stw	r21, RTCSCK(r20)

	li	r21, RTCSC_VAL@l
	sth	r21, RTCSC(r20)
	
	lis	r21, KEY_VAL@h
	ori	r21, r21, KEY_VAL@l
	stw	r21, PISCRK(r20)

	li	r21, PISCR_VAL@l
	sth	r21, PISCR(r20)

/* Clear Reset Status Register */
	
	lis	r21, 0xffff
	ori	r21, r21, 0xffff
	stw	r21, RSR(r20) 
	
#define WANT_DELAY 0
#ifdef WANT_DELAY
/* We need this loop in order to wait until the power supply
   has stabilized before we do the next critical step.  */ 
	
	lis	r21, 20
	mtctr	r21
loop1:	or	r0, r0, r0
	lis	r3, 0
	ori	r3, r3, 0x556c
	sth	r3, SWSR(r20)
	lis	r3, 0
	ori	r3, r3, 0xaa39
	sth	r3, SWSR(r20)
	bdnz	loop1

#endif

#if 0
	lis	r21, IDC_ENABLE@h	/* Enable all */
	mtspr	IC_CST, r21
	isync
#endif

#if 0	/* this tradionally give one a headache and is already done in the boot loader */
	
	lis	r21, KEY_VAL@h
	ori	r21, r21, KEY_VAL@l
	stw	r21, PLPRCRK(r20)

	bl	snom_get_plprcr_val
	bl	_set_plprcr
#endif
	
	/* Reset the caches.
	*/
	lis	r21, IDC_DISABLE@h	/* Disable */
	mtspr	IC_CST, r21
	mtspr	DC_CST, r21
	lis	r21, IDC_UNALL@h	/* Unlock all */
	mtspr	IC_CST, r21
	mtspr	DC_CST, r21
	lis	r21, IDC_INVALL@h	/* Invalidate all */
	mtspr	IC_CST, r21
	mtspr	DC_CST, r21

  	lis	r3,start_ldr@h
	ori	r3,r3,start_ldr@l
	mtlr	r3		/* Easiest way to do an absolute jump */
	blr

start_ldr:
/* Most 8xx boards don't boot up with the I-cache enabled.  Do that
 * now because the decompress runs much faster that way.
 */
	tlbia		/* Gotta do it somewhere.... */
	isync

	/* Because of the way the 860 starts up and assigns CS0 the
	 * entire address space, we have to set the memory controller
	 * differently.  Normally, you write the option register
	 * first, and then enable the chip select by writing the
	 * base register.  For CS0, you must write the base register
	 * first, followed by the option register.
	 */

	/* Enable timebase register.
	*/

	lhz	r21, TBSCR@l(r20)
	ori	r21, r21, 1

	lis	r21, KEY_VAL@h
	ori	r21, r21, KEY_VAL@l
	stw	r21, TBSCRK(r20)
	
	sth	r21, TBSCR@l(r20)


	/* Initialize stack pointer and jump to main functions.
	 * Leave some space at the top of memory for other uses.
	*
	* YOU MUST UPDATE THE END OF FILE IF YOU CHANGE THIS.
	*
	 */

#ifdef CONFIG_SNOM1
	lis	r1, 0x00ff		/* just under 16 MB */
	ori	r1, r1, 0xd000
#endif
	
#if 0
	lis	r3, IDC_INVALL@h
	mtspr	IC_CST, r3
	lis	r3, IDC_ENABLE@h
	mtspr	IC_CST, r3
#endif
	
#ifdef CONFIG_SNOM1
	bl	copy_snom1_data_seg
#endif /* CONFIG_SNOM1 */
	
/* Clear all of BSS */
	lis	r3,edata@h
	ori	r3,r3,edata@l
	lis	r4,end@h
	ori	r4,r4,end@l
	subi	r3,r3,4
	subi	r4,r4,4
	li	r0,0
50:	stwu	r0,4(r3)
	cmp	0,r3,r4
	bne	50b

#if 0
	/* Lets adjust the IMMR to a better value, high in kernel 1:1
	   address space */
	lis	r10,IMMR@h
	mtspr	638,r10 
	stw	r10,8(0)
#endif
	
	/* Perform configuration of the various boards.  This is done
	 * by reading some configuration data from EEPROM and building
	 * the board information structure.
	 */

#ifdef CONFIG_SNOM1
	lis	r11, 0x00ff
	ori	r11, r11, 0xe000
#endif

	mr	r3, r11
        mr      r21, r11
	mr	r22, r8
	mr	r23, r7
	mr	r24, r6

#ifdef CONFIG_SNOM1
	bl	snom1_cfg
#endif
	
	mr	r3, r21

	bl      serial_init	/* Init MBX serial port */
	
	mr      r11, r21
	mr	r8, r22
	mr	r7, r23
	mr	r6, r24

	lis	r3, 0xfe00
	lis	r4, 0
	lis	r5, 0

	lis	r21, IDC_ENABLE@h	/* Enable all */
	mtspr	IC_CST, r21
	isync

#undef WANT_DCACHE_ON
#ifdef WANT_DCACHE_ON
	
	mtspr	DC_CST, r21
	sync
	
#endif
	
	mr	r6,r11			/* Residual data */
	bl	decompress_kernel

/* Reset the caches. */

/*
	lis	r21, IDC_DISABLE@h
	mtspr	IC_CST, r21
	mtspr	DC_CST, r21
*/
	lis	r21, IDC_UNALL@h	
	mtspr	IC_CST, r21
	mtspr	DC_CST, r21
	lis	r21, IDC_INVALL@h	
	mtspr	IC_CST, r21
	mtspr	DC_CST, r21
	
	/* changed to use r3 (as firmware does) for kernel
	   as ptr to residual -- Cort*/
	lis	r6,cmd_line@h
	ori	r6,r6,cmd_line@l
	lwz	r6, 0(r6)
	subi	r7,r6,1
00:	lbzu	r2,1(r7)
	cmpi	0,r2,0
	bne	00b

	/* r4,r5 have initrd_start, size */
	lis	r2,initrd_start@h
	ori	r2,r2,initrd_start@l
	lwz	r4,0(r2)
	lis	r2,initrd_end@h
	ori	r2,r2,initrd_end@l
	lwz	r5,0(r2)
	
	/* The world starts from the beginning.
	*/
	li	r9,0x0
	mtlr	r9

	/* Invalidate the instruction cache because we just copied a
	 * bunch of kernel instructions.
	 */
	lis	r9, IDC_INVALL@h
	mtspr	IC_CST, r9

	blr
hang:
	b	hang	

/*
 * Delay for a number of microseconds
 * -- Use the BUS timer (assumes 66MHz)
 */
	.globl	udelay
udelay:		
	mulli	r4,r3,1000	/* nanoseconds */
	addi	r4,r4,59
	li	r5,60
	divw	r4,r4,r5	/* BUS ticks */
1:	mftbu	r5
	mftb	r6
	mftbu	r7
	cmp	0,r5,r7
	bne	1b		/* Get [synced] base time */
	addc	r9,r6,r4	/* Compute end time */
	addze	r8,r5
2:	mftbu	r5
	cmp	0,r5,r8
	blt	2b
	bgt	3f
	mftb	r6
	cmp	0,r6,r9
	blt	2b
3:	blr		

.globl _get_RSR
_get_RSR:
	lis	r3, IMMR@h
	lwz	r3, RSR@l(r3)
	blr

	
.globl _get_TBSCR
_get_TBSCR:
	lis	r3, IMMR@h
	lhz	r3, TBSCR@l(r3)
	blr

.globl _get_MSR
_get_MSR:		
	mfmsr	r3
	blr
	
.globl _put_MSR
_put_MSR:		
	mtmsr	r3
	blr


#ifdef WANT_MEMCPY_IN_ASSEMBLER
	
.globl	memcpy
memcpy:
	rlwinm.	r7,r5,32-3,3,31		/* r0 = r5 >> 3 */
	addi	r6,r3,-4
	addi	r4,r4,-4
	beq	2f			/* if less than 8 bytes to do */
	andi.	r0,r6,3			/* get dest word aligned */
	mtctr	r7
	bne	5f
1:	lwz	r7,4(r4)
	lwzu	r8,8(r4)
	stw	r7,4(r6)
	stwu	r8,8(r6)
	bdnz	1b
	andi.	r5,r5,7
2:	cmplwi	0,r5,4
	blt	3f
	lwzu	r0,4(r4)
	addi	r5,r5,-4
	stwu	r0,4(r6)
3:	cmpwi	0,r5,0
	beqlr
	mtctr	r5
	addi	r4,r4,3
	addi	r6,r6,3
4:	lbzu	r0,1(r4)
	stbu	r0,1(r6)
	bdnz	4b
	blr
5:	subfic	r0,r0,4
	mtctr	r0
6:	lbz	r7,4(r4)
	addi	r4,r4,1
	stb	r7,4(r6)
	addi	r6,r6,1
	bdnz	6b
	subf	r5,r0,r5
	rlwinm.	r7,r5,32-3,3,31
	beq	2b
	mtctr	r7
	b	1b

#endif
	.comm	.stack,4096*2,4
	
#ifndef CONFIG_MBX
local_bd_info:
	.long	0
#ifdef CONFIG_SNOM1
	.long	0x01000000    /* 16 MB */
#endif
	.long	10	      /* MF = 10 */
	.long	5000000	      /* EXCLK = 5000000 */
	.long	0
	.long	0
	.long	0
#endif





