/*
 *  linux/drivers/char/snomdisp.c
 *
 *  Copyright (C) 2000, snom technology AG
 *
 *  This driver is for the LCD module connected to the SNOM board.
 *  It only implements the write system call and uses the lowlevel 
 *  driver that is also available at boot time.
 */

/***********************************************************************/

#include <linux/config.h>

#ifdef CONFIG_SNOM_DISP

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>

#ifdef  CONFIG_SNOM_DISP
#ifndef  CONFIG_SNOM200
#include "../arch/ppc/8xx_io/progress.h"
#endif
#endif

/* Debug */

#if 0
#define DEB(x)        printk(x);
#define DEB2(x,y)     printk(x,y);
#define DEB3(x,y,z)   printk(x,y,z);
#define DEB4(x,y,z,a) printk(x,y,z,a);
#else
#define DEB(x)
#define DEB2(x,y)
#define DEB3(x,y,z)
#define DEB4(x,y,z,a)
#endif

/****** Generic driver stuff. */

extern void put_light_snom_display(int bOn);
extern void inc_contrast_snom_display(int bUp);
extern void set_contrast_snom_display(unsigned char contrast_val);

static struct file_operations snom_disp_fops;

static ssize_t write_snom_disp(struct file * file, const char * buf,
			  size_t count, loff_t *ppos)
{
  write_snom_display(buf, count);
  return count;
}

static int open_snom_disp(struct inode * inode, struct file * filp)
{
  extern void init_snom_display(void);

	DEB("snomdisp: open being called.\n");

	switch (MINOR(inode->i_rdev)) {
		case 1:
			filp->f_op = &snom_disp_fops;
			init_snom_display();
			break;
		default:
			return -ENXIO;
	}
	
	return 0;
}

static int ioctl_snom_disp(struct inode *inode, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
  int nError;

  DEB("snomdisp: ioctl being called.\n");

  switch(cmd) {
#ifdef CONFIG_SNOM200
  case SNOM_DISP_PUT_LED: 
    {
      int nVal;
      get_user(nVal, (int *)arg);
      put_led_snom_display(nVal);
      return 0;
    }
#endif
#ifdef CONFIG_SNOM220
  case SNOM_DISP_PUT_LED: 
    {
      int nVal;
      get_user(nVal, (int *)arg);
      put_led_snom_display(nVal);
      return 0;
    }
#endif
  case SNOM_DISP_PUT_LIGHT: 
    {
      int bOn;
      get_user(bOn, (int *)arg);
      put_light_snom_display(bOn);
      return 0;
    }
  case SNOM_DISP_INC_CONTRAST: 
    {
      int bUp;
      get_user(bUp, (int *)arg);
      inc_contrast_snom_display(bUp);
      return 0;
    }
#ifdef CONFIG_SNOM200
  case SNOM_DISP_SET_CONTRAST: 
    {
      int bVal;
      get_user(bVal, (int *)arg);
      set_contrast_snom_display((unsigned char)bVal);
      return 0;
    }
#endif
#ifdef CONFIG_SNOM220
  case SNOM_DISP_SET_CONTRAST: 
    {
      int bVal;
      get_user(bVal, (int *)arg);
      set_contrast_snom_display((unsigned char)bVal);
      return 0;
    }
#endif
  default:
    return -EINVAL;
  }
  return nError;
}

static struct file_operations snom_disp_fops = {
  write:    write_snom_disp,
  ioctl:    ioctl_snom_disp,		
  open:     open_snom_disp
};

void snom_disp_init(void)
{
  printk("SNOM Display $Revision: 1.3 $ $Date: 2003/09/23 16:16:10 $.\n");

  if (register_chrdev(SNOM_DISP_MAJOR,"snomdisp",&snom_disp_fops))
    printk("unable to get major %d for memory devs\n", SNOM_DISP_MAJOR);

#ifdef CONFIG_SNOM_DISP
#ifndef CONFIG_SNOM200
        write_snom_display(progress_pic[5], sizeof(progress_pic[0]));
#else
        progress_snom_display();
#endif
#endif
}

#endif






