/*
 * A collection of structures, addresses, and values associated with
 * the SNOM1 board. 
 */

#ifndef __MACH_SNOM_DEFS
#define __MACH_SNOM_DEFS

/*
 * 3.4.00 : The SNOM1 is using the following physical memorymap:
 *
 * fe000000 -> ff3fffff : flash      connected to CS0, CS1
 * 00000000 -> 01000000 : sdram/dram connected to CS2, CS3
 * fe400000 -> ff5fffff : Amplifier base addr  to CS4
 */

/* ------------------------------------------------------------------------- */
/* 8xxrom                                                                    */

#define IMMR_VALUE              0xff000000      /* needed by linux too */

/* ------------------------------------------------------------------------- */
/* linux                                                                     */

#ifndef __ASSEMBLY__

#define IMAP_ADDR			((uint) IMMR_VALUE)
#define IMAP_SIZE			((uint)(64 * 1024))

#define FLASH_BASE_ADDR			((uint) 0xfe000000)
#define FLASH_BASE_PLUS_2MB_ADDR	((uint) 0xfe200000)

#define SNOM_CSR_BASE_ADDR	        ((uint) 0xfe400000)
#define SNOM_CSR_SIZE	                ((uint)(4 * 1024)) 

#define FLASH_2MB_SIZE			((uint)(2 * 1024 * 1024)) 
#define FLASH_4MB_SIZE			((uint)(4 * 1024 * 1024)) 

// Ioctls

#define SNOM_FLASH_ERASE                 1
#define SNOM_FLASH_ERASE_SETTINGS_SECTOR 2
#define SNOM_DISP_PUT_LIGHT              3
#define SNOM_FLASH_ERASE_SERIAL_NUMBER   4
#define SNOM_DISP_INC_CONTRAST           5
#define SNOM_FLASH_ERASE_FIRMWARE_SECTOR 6
#define SNOM_FLASH_ERASE_BOOT_SECTOR     7
#define SNOM_DISP_PUT_LED	         8
#define SNOM_DISP_SET_CONTRAST           9

// Seeks to flash

#define SNOM_FLASH_BEGIN           0
#define SNOM_FLASH_SETTINGS_SECTOR 1
#define SNOM_FLASH_SEEK_CUR        2

/* A Board Information structure that is given to a program when
 * prom starts it up.
 */
typedef struct bd_info {
	unsigned int	bi_memstart;	/* Memory start address */
	unsigned int	bi_memsize;	/* Memory (end) size in bytes */
	unsigned int	bi_mf;		/* Multiplication Factor */
	unsigned int	bi_exclk;	/* External Freq, in Hz */
	unsigned char	bi_enetaddr[6];
	unsigned int	bi_baudrate;
	unsigned int    bi_flash_type; 
} bd_t;

extern bd_t m8xx_board_info;

/* Flash types */

#define FLASH_2MB	1
#define FLASH_2x2MB	2
#define FLASH_4MB	3

/* We don't use the 8259.
*/
#define NR_8259_INTS	0

/* Machine type
*/
#define _MACH_8xx (_MACH_fads)

#endif /* __ASSEMBLY__ */

#endif
